/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <UpTools/UpConf.h>
#include <fstream>
#include <sstream>
#include <cctype>


UpConf::UpConf()
: status(true), statusPhrase("ok, no data"), confFile("no_file"),
  dataLineRegex(  "^[ \t]*([^#]*[^ \t#])?*[ \t]*(#.*)?$" , 3 ),
  eqLineRegex(    "^[ \t]*([a-zA-Z0-9~!@$%*-_+/]+)[ \t]*=[ \t]*([^=#]*[^ \t#])?[\t ]*(#.*)?$" , 4 ),
  labelLineRegex( "^[ \t]*([a-zA-Z0-9~!@$%*-_+/]+)?[ \t]*(#.*)?$" , 3 )
{ }


bool UpConf::parseFile(const char* _confFile) {
	status=true;
	statusPhrase="ok";
	confFile=_confFile;
	currentLabel.clear();

   paramIt = lastIt = params.end();
	keyIt = keys.end();

	std::ifstream cf(_confFile);
	if(!cf.is_open()) {
		statError("error opening file",0);
		return false;
	}
	char line[UPCONF_MAXLINE]; 
	size_t lineNum=0;
	size_t labelModeCounter=0;

	while(cf.good()) {

		lineNum++;

		cf.getline(line,UPCONF_MAXLINE);
		if( cf.bad() ) {
			if( !cf.eof() ) statError("read error",lineNum);
			break;
		}

		if( currentLabel.size()!=0 ) {
			// currently in label mode, continue until empty line
			if( !dataLineRegex.exec(line) ) {
				statError("parse error",lineNum);
				break;
			};
			const std::string& value(dataLineRegex[1]);
			if( value.size()==0 ) {
				// empty line => end of label mode
				if( labelModeCounter==0 ) {
					statError("label not followed by data lines",lineNum-1);
					break;
				}
				currentLabel.clear();
				continue;
			}
			// assign another value to label
		  labelModeCounter++;
		  keys.insert( currentLabel );
		  params.insert( StrStrMap::value_type(currentLabel,value) );
		  continue;
		}

		if( eqLineRegex.exec(line) ) {
			// equal line assignment
			const std::string& value(eqLineRegex[2]);
			if( !value.size() ) {
				statError("label assignment without value",lineNum);
				break;
			}
			currentLabel = eqLineRegex[1];
			keys.insert(currentLabel);
			params.insert(StrStrMap::value_type(currentLabel,value));
			currentLabel.clear();
			continue;
		}

		if( !labelLineRegex.exec(line) ) {
			statError("parse error",lineNum);
			break;
		}
		currentLabel = labelLineRegex[1]; // may be empty
		labelModeCounter = 0;
  }

  paramIt = lastIt = params.end();
  keyIt = keys.begin();
  currentLabel.clear();

  cf.close();
  return status;
}



void UpConf::statError(const char* err,size_t lnum) {
	std::ostringstream x;
	x << confFile;
	if( lnum>0 ) x << " line " << lnum;
	x << ": " << err;
	statusPhrase = x.str();
	status=false;
}


std::string UpConf::emptyString="";

void UpConf::addParam(const std::string& _label,const std::string& _value) {
	params.insert(StrStrMap::value_type(_label,_value));
	keys.insert( _label );
	status=true;
	statusPhrase="ok";
}


const std::string& UpConf::getParam(const std::string& label) {

		if( label.size()==0 ) return emptyString;

		if(label!=currentLabel) {
			currentLabel=label;
			paramIt = params.lower_bound(label);
			lastIt  = params.upper_bound(label);
		}

		if(paramIt==lastIt) return emptyString;

		const std::string& ret = paramIt->second;
		paramIt++;
		return ret;
}

const std::string& UpConf::getKey(bool first) {
		if(first) keyIt = keys.begin();

		if( keyIt==keys.end() ) return emptyString;

		const std::string& ret = *keyIt;
		keyIt++;
		return ret;
}


std::ostream& operator<<(std::ostream& o, UpConf& c) {
	o<<"Configuration:";
	if(c.confFile.size() ) o<<" file="<<c.confFile;
	o<<" paramters="<<c.keyCount()<<"\n";
	std::string last;
	StrStrMultiMap::iterator it;
	for(it = c.params.begin(); it!=c.params.end(); it++ ) {
		if( it->first!=last) {
			last = it->first;
			o<<"  Parameter "<<last;
			unsigned int m=c.paramCount(last); 
			if( m>1 ) { o<<" multiplicity: "<<m; } 
			o<<"\n";
		}
		o<<"    "<<it->second<<"\n";
	}
	return o;
}

