/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Ignacio Kaufmann for his work in this class

#include <UpTools/UpResolver.h>
#include <UpTools/UpOp.h>
#include <UpTools/UpLog.h>

#include <iostream>
#include <cerrno>
#include <cstring>

// globals
UpResourceRecord UpDnsQuery::invalidResourceRecord(false);

/////////////////////////////////////////////////////////////////////////////
//UpDnsQuery Methods


UpDnsQuery::UpDnsQuery( const std::string& _domain, ns_type _type,
								dnsQueryCallbackType _callbackType,
							   dnsQueryBlockingType _blockingType          )
: id(0), status(QUERY_INITIAL), rstatus(QR_IDLE),
  domain(_domain), type(_type), querySize(0),answerSize(0),
  transmitTime(0),count(0), parsed(0), first(0), current(0),
  callbackType(_callbackType), blockingType(_blockingType),
  resolverPtr(0),blockingCondition(-1)
{ }

UpDnsQuery::UpDnsQuery( const UpAddress& upa,
								dnsQueryCallbackType _callbackType,
							   dnsQueryBlockingType _blockingType          )
: id(0), status(QUERY_INITIAL), rstatus(QR_IDLE),
  querySize(0),answerSize(0),
  transmitTime(0),count(0), parsed(0), first(0), current(0),
  resolverPtr(0),blockingCondition(-1)
{ set(upa,_callbackType,_blockingType); }

UpDnsQuery::UpDnsQuery() 
: id(0), status(QUERY_INITIAL), rstatus(QR_IDLE),
  type(ns_t_invalid), querySize(0),answerSize(0),
  transmitTime(0),count(0), parsed(0), first(0), current(0),
  callbackType(QUERY_DONT_USE_CALLBACK), blockingType(QUERY_BLOCKING),
  resolverPtr(0),blockingCondition(-1)
{ }

UpDnsQuery::~UpDnsQuery() {
	UpResolver* rp(resolverPtr);
	if( rp ) rp->terminateQuery(*this);
	// delete allocated resource records
	deleteRecords();
}

void UpDnsQuery::allowWaiting() {
	UpResolver* rp(resolverPtr);
	if( ! rp ) {
		UPLOG(LOG_ERR,"UpDnsQuery::allowWaiting() error: cannot be called "
			"without being executed by an UpResolver");
		return;
	}
	if( rp->checkSelf() ) {
		UPLOG(LOG_ERR,"UpDnsQuery::allowWaiting() error: cannot be called "
			"from within UpResolver's own thread (called from "
			"UpDnsQuery::completed()?)");
		return;
	}
	// prevents a second thread to enter the safe region
	blockingCondition.lock();
		if( blockingCondition.var==-1 )
			blockingCondition.var = 0;
	blockingCondition.unlock();
}

bool UpDnsQuery::waitCompletion() {
	UpResolver* rp( resolverPtr );
	if( ! rp ) {
		return false;
	}
	if( rp->checkSelf() ) {
		UPLOG(LOG_ERR,"UpDnsQuery::waitCompletion() error: cannot be called "
			"from within UpResolver's own thread (called from "
			"UpDnsQuery::completed()?)");
		return false;
	}
	// prevents a second thread to enter the safe region
	blockingCondition.lock();
		if( blockingCondition.var>=0 ) {
			++blockingCondition.var;
			blockingCondition.wait(); // unlocks on calling, and locks on return
		}
	blockingCondition.unlock();
	return status==QUERY_READY;
}

void UpDnsQuery::releaseWaitingThreads() {
	bool shouldBroadcast(false);
	// prevents a second thread to enter the safe region
	blockingCondition.lock();
		if( blockingCondition.var>0 ) {
			blockingCondition.var = -1;
			shouldBroadcast = true;
		}
		if( shouldBroadcast )
			blockingCondition.broadcast(); 
	blockingCondition.unlock();
}


void UpDnsQuery::set(const std::string& _domain, ns_type _type,
					 		dnsQueryCallbackType _callbackType,
							dnsQueryBlockingType _blockingType    )
{
	cancel(); // this sets rstatus as QR_IDLE
	// _domain must be used BEFORE deleteRecords(), because the application
	// may pass a reference to a string contained in one of the resource
	// records to be deleted.
	// This may be the case when requerying the same UpDnsQuery using
	// results from a previous operation.
	domain       = _domain;
	deleteRecords();
	id				 = 0;
	status		 = QUERY_INITIAL;
	querySize	 = answerSize = 0;
	type         = _type;
	callbackType = _callbackType;
	blockingType = _blockingType;
}

void UpDnsQuery::clear() {
	cancel();
	deleteRecords();
	domain.clear();
	id				= 0;
	status		= QUERY_INITIAL;
	type			= ns_t_invalid;
	querySize	= answerSize = 0;
}

// este metodo lo llamamos desde el mutex del resolver porque no
// sabemos si res_mkquery() es thread safe
// si resulta que es thread safe, mudar al constructor y a set()
bool UpDnsQuery::prepare() {
	// erase previous records
	deleteRecords(); 
	// si ya fue preparado y fallo => no intenta de nuevo
	if( status==QUERY_CREATION_ERROR ) return false;
	// sa ya fue preparado de nuevo OK => no hace falta mas
	if( status==QUERY_CREATED )        return true;

	answerSize = 0;
	querySize  = res_mkquery( ns_o_query, domain.c_str(),
							ns_c_in, type, NULL, 0, NULL,
							query_buff, NS_PACKETSZ );
	if(querySize < 0) {
		id=0;
	} else {
		id = ns_get16(query_buff);
		status = QUERY_CREATED;
	}

	return true;
}


bool UpDnsQuery::initialResponseParse() { 
	if( (answerSize<=0) || (ns_initparse(answer_buff, answerSize, &msg)<0) ) {
		status = QUERY_PARSE_ERROR;
		return false;
	}
	count = ns_msg_count(msg, ns_s_an);
	status = QUERY_READY;
	return true;
}


const UpResourceRecord& UpDnsQuery::getFirst() const {
	if(status!=QUERY_READY) return invalidResourceRecord;
    while( !first && (parsed<count) ) {
        ns_rr rr;
        if( ns_parserr((ns_msg*)&msg, ns_s_an, parsed++, &rr) < 0 )
            continue; // error, try next rr
        first = new UpResourceRecord(rr,msg);
    }
    return first ? *(current=first) : invalidResourceRecord;
}


const UpResourceRecord& UpDnsQuery::getNext() const {
	if(status!=QUERY_READY) return invalidResourceRecord;
    if(!current) return getFirst();
    while( !current->next && (parsed<count) ) {
        ns_rr rr;
        if( ns_parserr((ns_msg*)&msg, ns_s_an, parsed++, &rr) < 0 )
            continue; // error, try next rr
        current->next = new UpResourceRecord(rr,msg);
    }
    return current->next ? *(current=current->next) : invalidResourceRecord;
}



std::ostream& operator<<(std::ostream& o,const UpDnsQuery& q) {
	o << "UpDnsQuery id: " << q.id << " status: " << dnsQueryStatus[q.status]
		<< " records: " << q.count << '\n';
	for(;;) {
		const UpResourceRecord& rr( q.getNext() );
		if(rr.getType()==ns_t_invalid ) break;
		std::string s;
		o << rr<<'\n';
	}
	return o;
}


std::string& operator<<(std::string& o,const UpDnsQuery& q) {
	o << "UpDnsQuery id: " << q.id << " status: " << dnsQueryStatus[q.status]
		<< " records: " << q.count << '\n';
	for(;;) {
		const UpResourceRecord& rr( q.getNext() );
		if(rr.getType()==ns_t_invalid ) break;
		std::string s;
		o << rr<<'\n';
	}
	return o;
}


/////////////////////////////////////////////////////////////////////////////
// these method are only executed for nonblocking queries
// and should be overwritten in derived classes
void UpDnsQuery::completed() {
	UPLOG(LOG_INFO,"UpDnsQuery id %u completed, status: %s records: %u",
		id,dnsQueryStatus[status].c_str(),count);
	if( status!=QUERY_READY ) return;
	for(;;) {
		const UpResourceRecord& rr( getNext() );
		if(rr.getType()==ns_t_invalid ) break;
		std::string s;
		s << rr;
		UPLOG(LOG_INFO,"%s",s.c_str());
	}
}

//////////////////////////////////////////////////////////////////////////////
void UpDnsQuery::set(const UpAddress& upa,
					 		dnsQueryCallbackType _callbackType,
							dnsQueryBlockingType _blockingType    )
{
	domain.clear();
	unsigned char* a=(unsigned char*)upa.ptr();
	type = ns_t_ptr;
	blockingType = _blockingType;
	callbackType = _callbackType;
	switch( upa.family() ) {
		case AF_INET:
			domain<<a[3]<<'.'<<a[2]<<'.'<<a[1]<<'.'<<a[0]<<".in-addr.arpa";
			break;
		case AF_INET6:
			for(unsigned char* p(a+15); p>=a; --p ) {
				domain << resolverUchar2hex((*p)&0x0f) << '.'
						 << resolverUchar2hex((*p)>>4)   << '.' ;
			}
			domain << ".IP6.INT";
			break;
		default:
			type=ns_t_invalid;
   }
}

