/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstring>
#include <cctype>
#include <cerrno>
#include <UpTools/UpInet.h>
#include <UpTools/UpOp.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <endian.h>
#include <byteswap.h>
#include <iostream>
#include <cassert>

typedef unsigned int uint;

in_addr zero_in_addr={0};

//////////////////////////////////////////////////////////////////////////
////// UpAdress

#define IPBUFSIZE 40
#define chauBuf()	if(buf!=0) { delete[] buf; buf = NULL; }
#define holaBuf()	if(buf==0) buf = new char[IPBUFSIZE]

// Note: network byte order is big endian
//       UpAddress and UpSockAddr store addresses in network byte order!
//       intel x86 and amd64 are little endian
//       sparc is big endian
// 64 bits constants
const unsigned long long one64ull     = 1;
const unsigned long long zero64ull    = 0;
const unsigned long long allones64ull = ~zero64ull;    // 2^64-1

const unsigned int zero32ui    = 0;
const unsigned int allones32ui = ~zero32ui;            // 2^32-1

const double       allones32d  = (double)allones32ui;  // 2^32-1
const double       twoQ32d     = allones32d+1;         // 2^32
const double       twoQ64d     = twoQ32d*twoQ32d;      // 2^64
const long double  twoQ64ld    = (long double)twoQ64d; // 2^64
const float        twoQ64f     = (float)twoQ64d;       // 2^64

// Network order definitions
#define _NLL 3
#define _NL  2
#define _NH  1
#define _NHH 0


////// destructor
UpAddress::~UpAddress() {
	if(buf!=0) delete[] buf;
}


////// constructors
UpAddress::UpAddress() : addrType(AF_UNSPEC), buf(NULL) {
}

UpAddress::UpAddress(const in_addr& in4) : addrType(AF_INET), buf(NULL) {
	upAddr.a4 = in4;
}

UpAddress::UpAddress(const in6_addr& in6)
: addrType(AF_INET6), buf(NULL) {
	upAddr.a6 = in6;
}

UpAddress::UpAddress(const UpAddress& upa) : addrType(upa.addrType), buf(NULL) {
	switch(addrType) {
		case AF_INET:		upAddr.a4 = upa.upAddr.a4; return;
		case AF_INET6:		upAddr.a6 = upa.upAddr.a6; return;
		case AF_UNSPEC:	return;
		default:				addrType = AF_UNSPEC; return;
	}
}

UpAddress::UpAddress(const char* p) : buf(NULL) {
	if( p==0 or *p==0 )									{ addrType = AF_UNSPEC; return;}
	if( inet_pton(AF_INET,p,&(upAddr.a4)) > 0 )	{ addrType = AF_INET; return;  }
	if( inet_pton(AF_INET6,p,&(upAddr.a6)) > 0 )	{ addrType = AF_INET6; return; }
	addrType = AF_UNSPEC;
}

UpAddress::UpAddress(const std::string& s) : buf(NULL) {
	if( s.size()==0 )
		{ addrType = AF_UNSPEC; return;}
	if( inet_pton(AF_INET,s.c_str(),&(upAddr.a4)) > 0 )
		{ addrType = AF_INET; return;  }
	if( inet_pton(AF_INET6,s.c_str(),&(upAddr.a6)) > 0 )
		{ addrType = AF_INET6; return; }
	addrType = AF_UNSPEC;
}

UpAddress::UpAddress(const char* p,size_t len) : buf(NULL) {
	if( p==0 or len==0 or len>=IPBUFSIZE )
		{ addrType = AF_UNSPEC; return; }
	buf = new char[IPBUFSIZE]; memcpy(buf,p,len); buf[len]=0;
	if( inet_pton(AF_INET,buf,&(upAddr.a4)) > 0 )
		{ addrType = AF_INET; return;  }
	if( inet_pton(AF_INET6,buf,&(upAddr.a6)) > 0 )
		{ addrType = AF_INET6; return; }
	clear();
}

UpAddress::UpAddress(const UpSockAddr& sa) : addrType(sa.family()),buf(NULL) {
	if( addrType==AF_INET )
		{ upAddr.a4 = ((sockaddr_in*)sa.ptr())->sin_addr; return; }
	else
	if( addrType==AF_INET6 )
		{ upAddr.a6 = ((sockaddr_in6*)sa.ptr())->sin6_addr; return; }
	// other family types are not supported
	addrType  = AF_UNSPEC;
}
// when linking with glibc version < 2.9 be64toh is not defined
// we use byteswap.h (next defines are taken from glibc version>=2.9)
#ifndef be64toh
#if __BYTE_ORDER == __LITTLE_ENDIAN
#  define htobe64(x) bswap_64 (x)
#  define htole64(x) (x)
#  define be64toh(x) bswap_64 (x)
#  define le64toh(x) (x)
#else
#  define htobe64(x) (x)
#  define htole64(x) bswap_64 (x)
#  define be64toh(x) (x)
#  define le64toh(x) bswap_64 (x)
#endif
#endif

UpAddress::operator float() {
	if( addrType==AF_INET ) {
		return (float)ntohl(upAddr.a4.s_addr);
	} else if( addrType==AF_INET6 ) {
		float ret = (float) be64toh(upAddr.be128.low);
		if( upAddr.be128.high!=0 )
			ret += twoQ64f * (float) be64toh(upAddr.be128.high);
		return ret;
	}
	return 0.0;
}


UpAddress::operator double() {
	if( addrType==AF_INET ) {
		return (double)ntohl(upAddr.a4.s_addr);
	} else if( addrType==AF_INET6 ) {
		double ret = (double) be64toh(upAddr.be128.low);
		if( upAddr.be128.high!=0 )
			ret += twoQ64d * (double) be64toh(upAddr.be128.high);
		return ret;
	}
	return 0.0;
}

UpAddress::operator long double() {
	if( addrType==AF_INET ) {
		return (long double)ntohl(upAddr.a4.s_addr);
	} else if( addrType==AF_INET6 ) {
		long double ret = (long double) be64toh(upAddr.be128.low);
		if( upAddr.be128.high!=0 )
			ret += twoQ64ld * (long double) be64toh(upAddr.be128.high);
		return ret;
	}
	return 0.0;
}

UpAddress::operator unsigned int() {
	if( addrType==AF_INET ) {
		return (unsigned int)ntohl(upAddr.a4.s_addr);
	} else if( addrType==AF_INET6 ) {
		return (unsigned int)ntohl(upAddr.a6.s6_addr32[_NLL]);
	}
	return 0;
}

void UpAddress::clear() {
	chauBuf();
	addrType=AF_UNSPEC;
	upAddr.be128.high = upAddr.be128.low = 0;
}

////// operator= methods
UpAddress& UpAddress::operator=(const in_addr& in4) {
	chauBuf();
	addrType = AF_INET;
	upAddr.a4 = in4;
	return *this;
}

UpAddress& UpAddress::operator=(const in6_addr& in6) {
	chauBuf();
	addrType = AF_INET6;
	upAddr.a6 = in6;
	return *this;
}


UpAddress& UpAddress::operator=(const UpAddress& upa) {
	addrType = upa.addrType;
	chauBuf();
	if( addrType==AF_INET )  { upAddr.a4 = upa.upAddr.a4; return *this; }
	if( addrType==AF_INET6 ) { upAddr.a6 = upa.upAddr.a6; return *this; }
	return *this;
}

UpAddress& UpAddress::operator=(const UpSockAddr& sa) {
	addrType = sa.family();
	chauBuf();
	if( addrType==AF_INET )
		{ upAddr.a4 = ((sockaddr_in*)sa.ptr())->sin_addr; return *this; }
	else
	if( addrType==AF_INET6 )
		{ upAddr.a6 = ((sockaddr_in6*)sa.ptr())->sin6_addr; return *this; }
	addrType  = AF_UNSPEC;
	return *this;
}


UpAddress& UpAddress::operator=(const char* p) {
	chauBuf();
	if(*p== 0)
		{ addrType = AF_UNSPEC; return *this; }
	if(inet_pton(AF_INET,p,&(upAddr.a4)) > 0 )
		{ addrType = AF_INET; return *this; }
	if(inet_pton(AF_INET6,p,&(upAddr.a6)) > 0 )
		{ addrType = AF_INET6; return *this; }
   addrType = AF_UNSPEC;
	return *this;
}

UpAddress& UpAddress::set(const char* p,size_t len) {
	if( *p==0 or len==0 or len>=IPBUFSIZE ) { clear(); return *this; }
	holaBuf(); memcpy(buf,p,len); buf[len]=0;
	if( inet_pton(AF_INET,buf,&(upAddr.a4)) > 0 )
		{ addrType = AF_INET; return *this; }
	if( inet_pton(AF_INET6,buf,&(upAddr.a6)) > 0 )
		{ addrType = AF_INET6; return *this; }
	clear();
	return *this;
}

UpAddress& UpAddress::setf(const char* p,size_t len,int fmly) {
// this method is faster than the version with no fmly parameter because it
// it does not need to guess addrType type
	if( p==0 or len==0 or len>=IPBUFSIZE ) { clear(); return *this; }
	holaBuf(); memcpy(buf,p,len); buf[len]=0;
	if( fmly==AF_INET ) {
		if( inet_pton(AF_INET,buf,&(upAddr.a4)) > 0 )
			{ addrType = AF_INET; return *this; }
	} else
	if( fmly==AF_INET6 ) {
		if( inet_pton(AF_INET6,buf,&(upAddr.a6)) > 0 )
			{ addrType = AF_INET6; return *this; }
	}
	clear();
	return *this;
}

UpAddress& UpAddress::setf(const char* p,int fmly) {
// this method is faster than the version with no fmly parameter because it
// it does not need to guess addrType type
	chauBuf();
	switch(fmly) {
		case AF_INET:
			if( inet_pton(AF_INET,p,&(upAddr.a4)) > 0 ) addrType = AF_INET;
			else addrType = AF_UNSPEC;
			break;
		case AF_INET6:
			if( inet_pton(AF_INET6,p,&(upAddr.a6)) > 0 ) addrType = AF_INET6;
			else addrType = AF_UNSPEC;
			break;
		default:
			addrType = AF_UNSPEC;
			break;
	}
	return *this;
}




////// get methods
size_t		UpAddress::size()			  const {
	return (addrType==AF_INET) ? sizeof(in_addr) :
			 ( (addrType==AF_INET6) ? sizeof(in6_addr) : 0 ) ;
}
void 			UpAddress::write(void* p) const {
	if(addrType!=AF_UNSPEC) std::memcpy(p,&upAddr,size());
}

const char* UpAddress::c_str() const {
	if(buf) return buf; // this means that buf was already set
	if(addrType!=AF_UNSPEC) {
		holaBuf();
		inet_ntop(addrType,(void*)&upAddr,buf,IPBUFSIZE);
		return buf;
	}
	return "";
}

bool UpAddress::operator==(const UpAddress& upa) const {
	if( addrType!=upa.addrType )
		return false;
	if( addrType==AF_INET )
		return upAddr.a4.s_addr==upa.upAddr.a4.s_addr;

	return	upAddr.be128.high==upa.upAddr.be128.high and
			upAddr.be128.low ==upa.upAddr.be128.low;
}


bool UpAddress::operator<(const UpAddress& upa) const {
	// uso memcmp() porque guardamos el dato en net order,
	// y entonces va a andar para todos los tipos de procesador
	switch(addrType) {
		case AF_INET:
			switch(upa.addrType) {
				case AF_INET:
					return ntohl(upAddr.a4.s_addr)<ntohl(upa.upAddr.a4.s_addr);
				default:
					return true; // IPV4 is smaller than all other families
			}
		case AF_INET6:
			switch(upa.addrType) {
				case AF_INET:
					return false; // IPV6 is bigger than IPV4
				case AF_INET6:
					if( upAddr.be128.high!=upa.upAddr.be128.high )
						return be64toh(upAddr.be128.high)<be64toh(upa.upAddr.be128.high);
					return be64toh(upAddr.be128.low)<be64toh(upa.upAddr.be128.low);
					// return memcmp( &upAddr, &(upa.upAddr),  size() ) < 0 ;
				default:
					return true;  // and is smaller than the rest
		 }
		default: return false;
	}
}


size_t UpAddress::hash() const {
	switch(addrType) {
		case AF_INET:
			return (size_t)upAddr.a4.s_addr;
		case AF_INET6:
			return (size_t)(upAddr.be128.high^upAddr.be128.low);
		default:		return 0;
	}
}

std::ostream& operator<<(std::ostream& o, const UpAddress& a) {
	if(a.addrType!=AF_UNSPEC) o << a.c_str();
	return o;
}

std::string& operator<<(std::string& o, const UpAddress& a) {
	if(a.addrType!=AF_UNSPEC) o << a.c_str();
	return o;
}



void UpAddress::negate() {
	chauBuf();
	switch(addrType) {
		case AF_INET:
			upAddr.a4.s_addr = ~upAddr.a4.s_addr;
			return;
		case AF_INET6:
			upAddr.be128.high = ~upAddr.be128.high;
			upAddr.be128.low  = ~upAddr.be128.low;
			return;
		default: addrType=AF_UNSPEC; // cannot invert any other types
	}
}


UpAddress& UpAddress::operator&=(const UpAddress& upa) {
	if(upa.addrType!=addrType)
		return *this;
	if(addrType==AF_INET) {
		chauBuf();
		upAddr.a4.s_addr &= upa.upAddr.a4.s_addr;
		return *this;
	}
	if(addrType==AF_INET6) {
		chauBuf();
		upAddr.be128.high &= upa.upAddr.be128.high;
		upAddr.be128.low  &= upa.upAddr.be128.low;
	}
	return *this;
}


unsigned int UpAddress::maskBitsLen() const {
	in_addr_t    m;
	unsigned int i;
	if( addrType==AF_INET ) {
		m=ntohl(upAddr.a4.s_addr);
		i=(m!=0);
	} else if( addrType==AF_INET6 ) {
		if( upAddr.a6.s6_addr32[_NHH]==0xffffffff ) {
			if( upAddr.a6.s6_addr32[_NH]==0xffffffff) {
				if( upAddr.a6.s6_addr32[_NL]==0xffffffff) {
					if( upAddr.a6.s6_addr32[_NLL]==0xffffffff) {
						return 128;
					} else {
						m=ntohl(upAddr.a6.s6_addr32[_NLL]);
						i=96+(m!=0);
					}
				} else {
					m=ntohl(upAddr.a6.s6_addr32[_NL]);
					i=64+(m!=0);
				}
			} else {
				m=ntohl(upAddr.a6.s6_addr32[_NH]);
				i=32+(m!=0);
			}
		} else {
			m=ntohl(upAddr.a6.s6_addr32[_NHH]);
			i=(m!=0);
		}
	} else {
		return 0;
	}
	while(m<<=1) ++i;
	return i;
}



void UpAddress::setMask(unsigned int i,int _addrType) {
	addrType=_addrType;
	chauBuf();
	if( addrType==AF_INET ) {
		if( i==0  )      upAddr.a4.s_addr=zero32ui;
		else if( i<=31 ) upAddr.a4.s_addr=htonl(allones32ui<<(32-i));
		else             upAddr.a4.s_addr=allones32ui;
	} else if( addrType==AF_INET6 ) {
		if( i==0 ) {
			upAddr.be128.high=zero64ull;
			upAddr.be128.low=zero64ull;
		} else if( i<=64 ) {
			upAddr.be128.high=htobe64(allones64ull<<(64-i));
			upAddr.be128.low=zero64ull;
		} else if( i<=127 ) {
			upAddr.be128.high=allones64ull;
			upAddr.be128.low=htobe64(allones64ull<<(128-i));
		} else { // i>=128
			upAddr.be128.high=allones64ull;
			upAddr.be128.low=allones64ull;
		}
	} else { // set mask is only supported for AF_INET6 or AF_INET6
		addrType=AF_UNSPEC;
	}
}



//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// UpHost methods

// unchanging globals
std::string       hostOk="ok";
std::string hostNotFound="host not found";
std::string       noData="no data";
std::string   noRecovery="no recovery";
std::string     tryAgain="try again";
std::string       erange="not enough buffer space in object";

std::string						emptyStr;
UpAddress					  	emptyUpAddress;
std::vector<std::string>	emptyStringVector;
std::vector<UpAddress>  	emptyUpAddressesVector;

UpHost::UpHost() : lookupOk(false), lookupErrno(0) {
}


UpHost::UpHost(const std::string& host) : lookupOk(false), lookupErrno(0) {
	init(host.c_str());
}


UpHost& UpHost::operator=(const std::string& host) {
 	lookupOk=false; lookupErrno=0;
	init(host.c_str());
	return *this;
}


UpHost::UpHost(const char* host) : lookupOk(false), lookupErrno(0) {
	init(host);
}


UpHost& UpHost::operator=(const char* host) {
 	lookupOk=false; lookupErrno=0;
	init(host);
	return *this;
}


UpHost::UpHost(const UpHost& uph)
: _hostName(uph._hostName), _aliases(uph._aliases), _addresses(uph._addresses),
  addrType(uph.addrType), addrLen(uph.addrLen), lookupOk(uph.lookupOk),
  lookupErrno(uph.lookupErrno)
{ }


UpHost& UpHost::operator=(const UpHost& uph) {
	_hostName = uph._hostName;		addrLen     = uph.addrLen;
	lookupOk  = uph.lookupOk;		lookupErrno = uph.lookupErrno;
	_aliases  = uph._aliases;		_addresses  = uph._addresses;
	addrType  = uph.addrType;
	return *this;
}



const std::string& UpHost::errPhrase()	const {
	if(lookupOk)			return hostOk;
	switch(lookupErrno) {
		case HOST_NOT_FOUND:	return hostNotFound;
		case NO_DATA:			return noData;
		case NO_RECOVERY:		return noRecovery;
		case TRY_AGAIN:		return tryAgain;
		case NETDB_INTERNAL:	return erange;
		default:					return hostOk;
	}
}

const std::string& UpHost::hostName() const {
	if(lookupOk) return _hostName;
	else return emptyStr;
}

int UpHost::family() const {
	if(lookupOk) return addrType;
	else return AF_UNSPEC;
}

size_t UpHost::addrSize() const {
	if(lookupOk) return addrLen;
	else return 0;
}

const std::vector<std::string>& UpHost::aliases() const {
	if(lookupOk) return _aliases;
	else return emptyStringVector;
}

const std::vector<UpAddress>&	UpHost::addresses() const {
	if(lookupOk) return _addresses;
	else return emptyUpAddressesVector;
}

const UpAddress& UpHost::address() const {
	if(lookupOk) return _addresses[0];
	else return emptyUpAddress;
}



bool UpHost::init(const char* host) {
	hostent* hostEntryPtr;

	#if __GNUC__ >= 2

	#define HBUFSIZE 1024
	char buf[HBUFSIZE];
	hostent  hostEntry;
	// GNU GLIBC2 reentrant version of gethostbyname
	int call_err = gethostbyname_r(	host , &hostEntry , (char*)buf ,
												HBUFSIZE , &hostEntryPtr , &lookupErrno );
	// if call_err==ERANGE, buf is too small to store the lookup data
	if( call_err ) { lookupOk = false; lookupErrno = NETDB_INTERNAL; }
	else lookupOk = ! lookupErrno;

	#else // non reentrant version of gethostbyname

	// may generate problems if server is heavily loaded
	hostEntryPtr = gethostbyname( host );
	if(hostEntryPtr) { lookupOk = true;  lookupErrno = 0; }
	else				  { lookupOk = false; lookupErrno=h_errno; }

	#endif

	if(lookupOk) init(hostEntryPtr);

	return lookupOk;
}


void UpHost::init(const hostent* h) {
	if( !h || ( h->h_addrtype!=AF_INET && h->h_addrtype!=AF_INET6 ) )
		{ lookupOk = false; lookupErrno = NO_RECOVERY; return; }

	_hostName = h->h_name;
	addrType = h->h_addrtype;
	addrLen  = h->h_length;

	// aliases and adresses vector initialization
	// it is faster if vector is resized before setting the elemnents
	char **charList; int n;

	charList = h->h_aliases;
	n = 0; while(*charList) { n++; charList++; }
	_aliases.resize(n,emptyStr);
	while(n) _aliases[--n] = *(--charList);

	charList = h->h_addr_list;
	n = 0; while(*charList) { n++; charList++; }
	_addresses.resize(n,emptyUpAddress);
	if(addrType==AF_INET) {
		while(n) _addresses[--n] = *(in_addr*)  *(--charList);
	} else { // addrType can only be AF_INET or AF_INET6 as we checked above
		while(n) _addresses[--n] = *(in6_addr*) *(--charList);
	}

	lookupOk = true;
	lookupErrno	= 0;
}


bool UpHost::operator==(const UpAddress& upa) const {
	if(addrType!=upa.family()) return false;
	std::vector<UpAddress>::const_iterator it=_addresses.begin();
	while(it!=_addresses.end()) { if(upa==*it) return true; it++; }
	return false;
}

bool UpHost::operator==(const UpSockAddr& usa) const {
	if(addrType!=usa.family()) return false;
	std::vector<UpAddress>::const_iterator it=_addresses.begin();
	while(it!=_addresses.end()) { if(usa==*it) return true; it++; }
	return false;
}

bool UpHost::operator==(const UpHost& uh) const {
	if(addrType!=uh.family()) return false;
	std::vector<UpAddress>::const_iterator it=_addresses.begin();
	while(it!=_addresses.end()) { if(uh==*it) return true; it++; }
	return false;
}

std::ostream& operator<<(std::ostream& o, const UpHost& h) {
	o << "HostName: " << h._hostName;
	if(h._aliases.size()) {
		o << "\nAliases: " << h._aliases[0];
		for(size_t i=1; i<h._aliases.size(); i++) o << ", " << h._aliases[i];
	}
	o << "\nAddress Type: ";
	switch(h.addrType) {
		case AF_UNIX: o << "unix"; break;
		case AF_INET: o << "ipv4"; break;
		case AF_INET6: o << "ipv6"; break;
		default: o << "unknown";
	}
	o << "\nAddress length: " << h.addrSize();

	if(h._addresses.size()) {
		o << "\nAddresses: " << h._addresses[0];
		for(size_t i=1; i < h._addresses.size(); i++)
		o << ", " << h._addresses[i];
	}
	if(!h.status()) o << "\nError: " << h.errPhrase();
	o << '\n';
	return o;
}


std::string& operator<<(std::string& o, const UpHost& h) {
	o << "HostName: " << h._hostName;
	if(h._aliases.size()) {
		o << "\nAliases: " << h._aliases[0];
		for(size_t i=1; i<h._aliases.size(); i++) o << ", " << h._aliases[i];
	}
	o << "\nAddress Type: ";
	switch(h.addrType) {
		case AF_UNIX: o << "unix"; break;
		case AF_INET: o << "ipv4"; break;
		case AF_INET6: o << "ipv6"; break;
		default: o << "unknown";
	}
	o << "\nAddress length: " << h.addrSize();

	if(h._addresses.size()) {
		o << "\nAddresses: " << h._addresses[0];
		for(size_t i=1; i < h._addresses.size(); i++)
		o << ", " << h._addresses[i];
	}
	if(!h.status()) o << "\nError: " << h.errPhrase();
	o << '\n';
	return o;
}



//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
////// UpSockAddr

////// metodos auxiliares
void UpSockAddr::allocAux(int family) {
	if(sa.a) { // some memory was already allocated
		int ownFamily = sa.a->sa_family;
		if(ownFamily == family) return;
		switch(ownFamily) {
			case AF_INET:  delete sa.in;  break;
			case AF_INET6: delete sa.in6; break;
			case AF_UNIX:  delete sa.un;  break;
			default: /* panic */
				assert( "unknown family allocated in UpSockAddr!"==0 );
		}
	}
	switch(family) {
			case 0: sa.a=0; return; // AF_UNSPEC = 0
			case AF_INET:  sa.in  = new sockaddr_in;  break;
			case AF_INET6: sa.in6 = new sockaddr_in6; break;
			case AF_UNIX:  sa.un  = new sockaddr_un;  break;
			default: sa.a=0; return; // error: not supported
	}
	sa.a->sa_family = family;
}


// from text number to number functions
unsigned int digits2uint(const char* p, int len) {
   if(!p||!*p||len<=0) return 0;
   register unsigned int ret=0;
   do {
		if((*p<'0')||(*p>'9')) return 0;
		ret *= 10; ret += (int)((*p++)-'0');
	} while(--len);
   return ret;
}
unsigned int digits2uint(const char* p) {
   if(!p||!*p) return 0;
   register unsigned int ret=0;
   do {
		if((*p<'0')||(*p>'9')) return 0;
		ret*=10; ret+=(int)((*p++)-'0');
	} while(*p);
   return ret;
}
// buffer pointed by p should have at least 11 bytes
char* uint2digits(unsigned int n, char *p) {
   p+=11; *p=0;
   do { *(--p) = '0'+(n%10); } while(n/=10);
   return p;
}

bool UpSockAddr::testAssign4(const char* p, unsigned short port) {
	allocAux(AF_INET);
	if( inet_pton(AF_INET,p,&(sa.in->sin_addr)) <= 0 ) return false;
	sa.in->sin_port = htons(port);
	return true;
}

bool UpSockAddr::testAssign6(const char* p, unsigned short port) {
	allocAux(AF_INET6);
	if( inet_pton(AF_INET6,p,&(sa.in6->sin6_addr)) <= 0 ) return false;
	sa.in->sin_port = htons(port);
	return true;
}

bool UpSockAddr::testAssignU(const char* p) {
	allocAux(AF_UNIX);
	std::strncpy(sa.un->sun_path,p,UNIX_PATH_MAX);
	sa.un->sun_path[UNIX_PATH_MAX-1]='\0';
	return true;
}

bool UpSockAddr::testAssignH(const char* p, unsigned short port) {
	hostent* hp;
	#if __GNUC__ >= 2
		// GNU GLIBC2 reentrant version of gethostbyname
		#define HBUFSIZE 1024
		char buf[HBUFSIZE]; hostent hostEntry; int t_err;
		if( gethostbyname_r(	p, &hostEntry, (char*)buf, HBUFSIZE, &hp, &t_err ) )
			return false;
	#else
		// non reentrant version of gethostbyname
		// may generate problems if server is heavily loaded
		hp=gethostbyname(p);
	#endif

	if( ! hp ) return false;
	switch(hp->h_addrtype) {
		case AF_INET:
			allocAux(AF_INET);
			sa.in->sin_addr = *(in_addr*)(hp->h_addr);
			sa.in->sin_port = htons(port);
			return true;
		case AF_INET6:
			allocAux(AF_INET6);
			sa.in6->sin6_addr = *(in6_addr*)(hp->h_addr);
			sa.in6->sin6_port = htons(port);
			return true;
	}
	return false;
}

#define errReturn { str.clear(); allocAux(0); return; }

void UpSockAddr::assignAux(const UpSockAddr& upa) {
	if(!upa.sa.a) errReturn;
	allocAux(upa.sa.a->sa_family);
	std::memcpy(sa.a, upa.sa.a, size());
	str=upa.str; format=upa.format;
}

void UpSockAddr::assignAux(const char* p, unsigned short port) {
	if(!p || !*p) errReturn;
	if( !testAssign4(p,port) && !testAssign6(p,port) ) errReturn;
	format=SOCKADDR_IP;
	str = p;
}


void UpSockAddr::assignAux(const char* p) {
	if( !p || !*p ) errReturn;
	str.assign(p);
	size_t pos;
	unsigned int port;
	if(str[0]=='[' ) {
		// IPV6REF: "[abc6::0001]" or "[abc6::0001]:80" or "[abc6::0001] 80"
		if((pos=str.rfind(']'))==str.npos) errReturn;
		switch(str[pos+1]) {
			case '\0': format=SOCKADDR_IP;
			           port=0;                              break;
			case ':' : format=SOCKADDR_2DOTS;
			           port=digits2uint(str.c_str()+pos+2); break;
			case ' ' : format=SOCKADDR_SPACE;
			           port=digits2uint(str.c_str()+pos+2); break;
			default  : errReturn;
		}
		if(port>65535) errReturn;
		str[pos]='\0';
		if( ! testAssign6(str.c_str()+1,port) ) errReturn;
		str[pos++]=']';
		if(format==SOCKADDR_SPACE) str.clear(); // accepted, but non-standard
		return;
	}
	if((pos=str.find(' '))!=str.npos)  {
		// IPV6ADDR or IPV4ADDR space format: "abc6:045a::0001 8080"
		format=SOCKADDR_SPACE;
		port = digits2uint(str.c_str()+pos+1);
		if(port==0 || port>65535) errReturn;
		str[pos]='\0';
		if( !testAssign4(str.c_str(),(unsigned short)port) &&
		    !testAssign6(str.c_str(),(unsigned short)port)      ) errReturn;
		str[pos] = ' ';
		return;
	}
	if((pos=str.rfind(':'))!=str.npos) {
		if(str.rfind(':',pos-1)!=str.npos) {
			// IPV6ADDR format: "abcd:1bcd::0001"
			format=SOCKADDR_IP;
			if( testAssign6(str.c_str(),0) ) return;
			errReturn;
		}
		// IPV4ADDR format: "200.55.33.123:8080"
		format=SOCKADDR_2DOTS;
		port = digits2uint(str.c_str()+pos+1);
		if(port==0 || port>65535) errReturn;
		str[pos]='\0';
		if( ! testAssign4(str.c_str(),(unsigned short)port) ) errReturn;
		str[pos] = ':';
		return;
	}
	if( testAssign4(str.c_str(),0) ) {
		// IPV4ADDR format: "200.69.213.4"
      format=SOCKADDR_IP;
		return;
	}
	//  unix socket sockAddress
	format=SOCKADDR_SPACE;
	testAssignU(str.c_str());
	return;
}

/////////////////////////////////////////////////////////////////////////////
///// host port methods (only for inet4 and inet6 sockets

bool UpSockAddr::hostPort(const char* hostAddr,unsigned short port) {
	str.clear();
	if( testAssign4(hostAddr,port) ) return true;
	if( testAssign6(hostAddr,port) ) return true;
	if( testAssignH(hostAddr,port) ) return true;
	allocAux(0); return false;
}

#define errhReturn { str.clear(); allocAux(0); return false; }

bool UpSockAddr::hostPort(const char* p) {
	if( !p || !*p ) errhReturn;
	str.assign(p);
	size_t pos;
	unsigned int port;
	if(str[0]=='[' ) {
		// IPV6REF: "[abc6::0001]" or "[abc6::0001]:80" or "[abc6::0001] 80"
		if((pos=str.rfind(']'))==str.npos) errhReturn;
		switch(str[pos+1]) {
			case '\0': format=SOCKADDR_IP;
			           port=0;                              break;
			case ':' : format=SOCKADDR_2DOTS;
			           port=digits2uint(str.c_str()+pos+2); break;
			case ' ' : format=SOCKADDR_SPACE;
			           port=digits2uint(str.c_str()+pos+2); break;
			default  : errhReturn;
		}
		if(port>65535) errhReturn;
		str[pos]='\0';
		if( ! testAssign6(str.c_str()+1,port) ) errhReturn;
		str[pos++]=']';
		if(format==SOCKADDR_SPACE) str.clear(); // accepted, but non-standard
		return true;
	}
	if((pos=str.find(' '))!=str.npos)  {
		// IPV6ADDR or IPV4ADDR or hostname space format:
		// "abc6:045a::0001 8080" or "www.palermo.edu 80"
		format=SOCKADDR_SPACE;
		port = digits2uint(str.c_str()+pos+1);
		if(port==0 || port>65535) errhReturn;
		str[pos]='\0';
		if( testAssign4(str.c_str(),(unsigned short)port) ||
		    testAssign6(str.c_str(),(unsigned short)port)    )
			{ str[pos] = ' '; return true; }
		if( testAssignH(str.c_str(),(unsigned short)port) )
			{ str.clear(); return true; }
		errhReturn;
	}
	if((pos=str.rfind(':'))!=str.npos) {
		if(str.rfind(':',pos-1)!=str.npos) {
			// IPV6ADDR format: "abcd:1bcd::0001"
			format=SOCKADDR_IP;
			if( ! testAssign6(str.c_str(),0) ) errhReturn;
			return true;
		}
		// IPV4ADDR or hostname: "200.55.33.123:8080" or  "www.palermo.edu:80"
		format=SOCKADDR_2DOTS;
		port = digits2uint(str.c_str()+pos+1);
		if(port==0 || port>65535) errhReturn;
		str[pos]='\0';
		if( testAssign4(str.c_str(),(unsigned short)port) )
			{ str[pos] = ':'; return true; }
		if( testAssignH(str.c_str(),(unsigned short)port)  )
			{ str.clear(); return true; }
		errhReturn;
	}
	if( testAssign4(str.c_str(),0) ) {
		// IPV4ADDR format: "200.69.213.4"
      format=SOCKADDR_IP;
		return true;
	}
	if( testAssignH(str.c_str(),0) ) {
		// hostname format: "www.palermo.edu"
      str.clear();
		return true;
	}
	// hostPort does not accept unix sockets, use operator=() or set() for that
	// format=SOCKADDR_SPACE;
	//	testAssignU(str.c_str());
	// return true;
	errhReturn;
}



/////////////////////////////////////////////////////////////////////////////
///// general methods
UpSockAddr::UpSockAddr() : format(SOCKADDR_SPACE) { sa.a=0; }
UpSockAddr::~UpSockAddr() { allocAux(0); }
int UpSockAddr::family() const { if(sa.a) return sa.a->sa_family; return 0; }
void UpSockAddr::allocSA(int fmly) { allocAux(fmly); str.clear(); }


/////////////////////////////////////////////////////////////////////////////
///// UpSockAddr methods
UpSockAddr::UpSockAddr(const UpSockAddr& upa)
: format(SOCKADDR_SPACE)
{ sa.a=0; assignAux(upa); }

UpSockAddr& UpSockAddr::operator=(const UpSockAddr& upa) {
	assignAux(upa);
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
///// Methods for in_addr and in6_addr types
// the following methods assume network byte order in the ip address
// but hardware byte order in the port
UpSockAddr::UpSockAddr(const in_addr& in4, unsigned short port)
: format(SOCKADDR_SPACE)
{
	sa.in = new sockaddr_in;
	sa.in->sin_family = AF_INET;
	sa.in->sin_addr = in4;
	sa.in->sin_port = htons(port);
}

UpSockAddr::UpSockAddr(const in6_addr& in6, unsigned short port)
: format(SOCKADDR_SPACE)
{
	sa.in6 = new sockaddr_in6;
	sa.in6->sin6_family = AF_INET6;
	sa.in6->sin6_addr = in6;
	sa.in6->sin6_port = htons(port);
}

UpSockAddr& UpSockAddr::operator=(const in_addr& in4) {
	str.clear();
	allocAux(AF_INET);
	sa.in->sin_addr = in4;
	sa.in->sin_port = 0;
	return *this;
}

UpSockAddr& UpSockAddr::operator=(const in6_addr& in6) {
	str.clear();
	allocAux(AF_INET6);
	sa.in6->sin6_addr = in6;
	sa.in6->sin6_port = 0;
	return *this;
}

void UpSockAddr::set(const in_addr& in,unsigned short port) {
	str.clear();
	allocAux(AF_INET);
	sa.in->sin_addr = in;
	sa.in->sin_port = htons(port);
}

void UpSockAddr::set(const in6_addr& in6,unsigned short port) {
	str.clear();
	allocAux(AF_INET6);
	sa.in6->sin6_addr = in6;
	sa.in6->sin6_port = htons(port);
}

void UpSockAddr::set(const UpAddress& upa, unsigned short port) {
	str.clear();
	switch(upa.family()) {
		case AF_INET:
			allocAux(AF_INET);
			sa.in->sin_addr = *((in_addr*)upa.ptr());
			sa.in->sin_port = htons(port);
			return;
		case AF_INET6:
			allocAux(AF_INET6);
			sa.in6->sin6_addr = *((in6_addr*)upa.ptr());
			sa.in6->sin6_port = htons(port);
			return;
		default:
			allocAux(AF_UNSPEC);
			return;
	}
}

UpSockAddr::UpSockAddr(const UpAddress& upa, unsigned short port)
: format(SOCKADDR_SPACE)
{
	sa.a = 0;
	switch(upa.family()) {
		case AF_INET:
			allocAux(AF_INET);
			sa.in->sin_addr = *((in_addr*)upa.ptr());
			sa.in->sin_port = htons(port);
			return;
		case AF_INET6:
			allocAux(AF_INET6);
			sa.in6->sin6_addr = *((in6_addr*)upa.ptr());
			sa.in6->sin6_port = htons(port);
			return;
		default:
			allocAux(AF_UNSPEC);
	}
}

// _addr should be in_addr* or in6_addr* depending on fmly
void UpSockAddr::set(const void* _addr, unsigned short port,int fmly) {
	str.clear();
	allocAux(fmly);
	switch(fmly) {
		case AF_INET:
			sa.in->sin_addr = *(in_addr*)_addr;
			sa.in->sin_port = htons(port);
			break;
		case AF_INET6:
			sa.in6->sin6_addr = *(in6_addr*)_addr;
			sa.in6->sin6_port = htons(port);
			break;
		default: allocAux(0);
	}
}
/////////////////////////////////////////////////////////////////////////////
///// Methods for struct sockaddr_xx types
UpSockAddr::UpSockAddr(const sockaddr* saddr)
: format(SOCKADDR_SPACE)
{
	sa.a=0;
	if(!saddr) return;
	allocAux(saddr->sa_family);
	memcpy(sa.a,saddr,size());
}

UpSockAddr::UpSockAddr(const sockaddr_un* saddr)
: format(SOCKADDR_SPACE)
{
	sa.a=0;
	if(!saddr) return;
	allocAux(AF_UNIX);
	std::memcpy(sa.un,saddr,sizeof(const sockaddr_un));
}

UpSockAddr::UpSockAddr(const sockaddr_in* saddr)
: format(SOCKADDR_SPACE)
{
	sa.a=0;
	if(!saddr) return;
	allocAux(AF_INET);
	std::memcpy(sa.in,saddr,sizeof(const sockaddr_in));
}

UpSockAddr::UpSockAddr(const sockaddr_in6* saddr)
: format(SOCKADDR_SPACE)
{
	sa.a=0;
	if(!saddr) return;
	allocAux(AF_INET6);
	std::memcpy(sa.in6,saddr,sizeof(const sockaddr_in6));
}

UpSockAddr& UpSockAddr::operator=(const sockaddr* saddr) {
	str.clear();
	if(!saddr) { allocAux(0); return *this; }
	allocAux(saddr->sa_family);
	memcpy(sa.a,saddr,size());
	return *this;
}

UpSockAddr& UpSockAddr::operator=(const sockaddr_un* saddr) {
	str.clear();
	if(!saddr) { allocAux(0); return *this; }
	allocAux(AF_UNIX);
	memcpy(sa.un,saddr,sizeof(const sockaddr_un));
	return *this;
}

UpSockAddr& UpSockAddr::operator=(const sockaddr_in* saddr) {
	str.clear();
	if(!saddr) { allocAux(0); return *this; }
	allocAux(AF_INET);
	memcpy(sa.in,saddr,sizeof(const sockaddr_in));
	return *this;
}

UpSockAddr& UpSockAddr::operator=(const sockaddr_in6* saddr) {
	str.clear();
	if(!saddr) { allocAux(0); return *this; }
	allocAux(AF_INET6);
	memcpy(sa.in6,saddr,sizeof(const sockaddr_in6));
	return *this;
}

void UpSockAddr::write(void* p) const {
	if(sa.a) std::memcpy(p,sa.a,size());
}

size_t UpSockAddr::size() const {
	if(!sa.a) return 0;
	switch(sa.a->sa_family) {
		case AF_UNIX: return sizeof(const sockaddr_un);
		case AF_INET: return sizeof(const sockaddr_in);
		case AF_INET6: return sizeof(const sockaddr_in6);
	}
	return 0;
}

/////////////////////////////////////////////////////////////////////////////
///// methods to use with human readable strings
// the string may be of the formats:
// ipv4 address:      "200.44.5.23"
// ipv4 address port: "200.44.5.23 80"
// ipv6 address:      "3ffe:401a:3000:0003"
// ipv6 address port: "3ffe:401a:3000:0003 443"
// unix address:      "/tmp/usock"
// port is in hardware byte order
// set is faster than operator= because the family is not guessed from data
UpSockAddr::UpSockAddr(const char* p)
: format(SOCKADDR_SPACE)
{ sa.a=0; assignAux(p); }

UpSockAddr::UpSockAddr(const char* p, unsigned short n)
: format(SOCKADDR_SPACE)
{ sa.a=0; assignAux(p,n); }

UpSockAddr& UpSockAddr::operator=(const char* p) {
	assignAux(p);
	return *this;
}

bool UpSockAddr::set(const char* p,unsigned short port,int fmly) {
// this method is faster than operator=(char*) because it
// it does not need to guess family type
	str=p;
	switch(fmly) {
		case AF_INET:
			format=SOCKADDR_IP;
			if( testAssign4(p,port) ) return true;
			break;
		case AF_INET6:
			format=SOCKADDR_IP;
			if( testAssign6(p,port) ) return true;
			break;
		case AF_UNIX:
			format=SOCKADDR_SPACE;
			allocAux(AF_UNIX);
			testAssignU(p);
			return true;
	}
	str.clear();
	return false;
}


/////////////////////////////////////////////////////////////////////////////
///// methods to use with human readable std::strings
// the string may be of the formats:
// ipv4 address:      "200.44.5.23"
// ipv4 address port: "200.44.5.23 80"
// ipv6 address:      "3ffe:401a:3000:0003"
// ipv6 address port: "3ffe:401a:3000:0003 443"
// unix address:      "/tmp/usock"
// port is in hardware byte order
// set is faster than operator= because the family is not guessed from data
UpSockAddr::UpSockAddr(const std::string& s)
: format(SOCKADDR_SPACE)
{ sa.a=0; assignAux(s.c_str()); }

UpSockAddr::UpSockAddr(const std::string& s, unsigned short n)
: format(SOCKADDR_SPACE)
{ sa.a=0; assignAux(s.c_str(),n); }

UpSockAddr& UpSockAddr::operator=(const std::string& s) {
	assignAux(s.c_str());
	return *this;
}


const std::string& UpSockAddr::string() const {
	if(!sa.a) return str;	// nothing to return
	if( str.size() && format==SOCKADDR_SPACE ) return str; // str already created
	format=SOCKADDR_SPACE;
	char buf[64], bufport[16], *p(buf), *q;
	switch(sa.a->sa_family) {
		case AF_INET:
			inet_ntop(AF_INET,&(sa.in->sin_addr),p,32);
			q=uint2digits(htons(sa.in->sin_port),bufport);
			while(*p) p++; *p++=' '; do { *p++=*q++; } while(*q); *p=0;
			return str = buf;
		case AF_INET6:
			inet_ntop(AF_INET6,&(sa.in6->sin6_addr),p,64);
			q=uint2digits(htons(sa.in6->sin6_port),bufport);
			while(*p) p++; *p++=' '; do { *p++=*q++; } while(*q); *p=0;
			return str = buf;
		case AF_UNIX:
			return str = sa.un->sun_path;
		default:
			return str = "error in family type"; /* should'nt get here! */
	}
}


const std::string& UpSockAddr::string_ip() const {
	if(!sa.a) return str;	// nothing to return
	if( str.size() && format==SOCKADDR_IP ) return str; // str already created
	format=SOCKADDR_IP;
	char buf[64];
	switch(sa.a->sa_family) {
		case AF_INET:
			inet_ntop(AF_INET,&(sa.in->sin_addr),buf,32);
			return str = buf;
		case AF_INET6:
			inet_ntop(AF_INET6,&(sa.in6->sin6_addr),buf,64);
			return str = buf;
		case AF_UNIX:
			return str = sa.un->sun_path;
		default:
			return str = "error in family type"; /* should'nt get here! */
	}
}

const std::string& UpSockAddr::string_2dots() const {
	if(!sa.a) return str;	// nothing to return
	if( str.size() && format==SOCKADDR_2DOTS ) return str; // str already created
	format=SOCKADDR_2DOTS;
	char buf[64], bufport[16], *p(buf), *q;
	switch(sa.a->sa_family) {
		case AF_INET:
			inet_ntop(AF_INET,&(sa.in->sin_addr),p,32);
			q=uint2digits(htons(sa.in->sin_port),bufport);
			while(*p) p++; *p++=':'; do { *p++=*q++; } while(*q); *p=0;
			return str = buf;
		case AF_INET6:
			*p++='[';
			inet_ntop(AF_INET6,&(sa.in6->sin6_addr),p,64);
			q=uint2digits(htons(sa.in6->sin6_port),bufport);
			while(*p) p++; *p++=']'; *p++=':'; do { *p++=*q++; } while(*q); *p=0;
			return str = buf;
		case AF_UNIX:
			return str = sa.un->sun_path;
		default:
			return str = "error in family type"; /* should'nt get here! */
	}
}

bool UpSockAddr::operator==(const UpSockAddr& usa) const {
	if( family() != usa.family() ) return false;
	switch(sa.a->sa_family) {
		case AF_INET: {
			register sockaddr_in *p((sockaddr_in*)usa.ptr());
			return (sa.in->sin_port == p->sin_port) &&
					 (sa.in->sin_addr.s_addr == p->sin_addr.s_addr) ;
			}
		case AF_INET6: {
			register sockaddr_in6 *p((sockaddr_in6*)usa.ptr());
			return (sa.in6->sin6_port == p->sin6_port)       &&
					 (sa.in6->sin6_addr.s6_addr32[0] == p->sin6_addr.s6_addr32[0]) &&
					 (sa.in6->sin6_addr.s6_addr32[1] == p->sin6_addr.s6_addr32[1]) &&
					 (sa.in6->sin6_addr.s6_addr32[2] == p->sin6_addr.s6_addr32[2]) &&
					 (sa.in6->sin6_addr.s6_addr32[3] == p->sin6_addr.s6_addr32[3]) ;
			}
		case AF_UNIX: {
			register char *px(sa.un->sun_path);
			register char *py(((sockaddr_un*)usa.ptr())->sun_path);
			while( *px == *py ) { if(*px==0) return true; ++px; ++py; }
			return false;
			}
		default:
			return false;
	}
}

bool UpSockAddr::operator==(const UpAddress& upa) const {
	if( family() != upa.family() ) return false;
	switch(sa.a->sa_family) {
		case AF_INET: {
			register in_addr* p((in_addr*)upa.ptr());
			return (sa.in->sin_addr.s_addr == p->s_addr) ;
			}
		case AF_INET6: {
			register in6_addr* p((in6_addr*)upa.ptr());
			return (sa.in6->sin6_addr.s6_addr32[0] == p->s6_addr32[0]) &&
					 (sa.in6->sin6_addr.s6_addr32[1] == p->s6_addr32[1]) &&
					 (sa.in6->sin6_addr.s6_addr32[2] == p->s6_addr32[2]) &&
					 (sa.in6->sin6_addr.s6_addr32[3] == p->s6_addr32[3]) ;
			}
		default:
			return false;
	}
}

bool UpSockAddr::operator<(const UpSockAddr& usa) const {
	switch(family()) {
		case AF_INET:
			switch(usa.family()) {
				case AF_INET: {
					// uso memcmp() porque guardamos el dato en net order,
					// y entonces va a andar para todos los tipos de procesador
					int x = memcmp( &(    sa.in->sin_addr),
										 &(usa.sa.in->sin_addr),
										 sizeof(in_addr) );
					if( x<0 ) return true;
					if( x>0 ) return false;
					return ntohs(sa.in->sin_port) < ntohs(usa.sa.in->sin_port);
					}
				case AF_INET6:
				case AF_UNIX:
				case AF_UNSPEC:
				default:
					return true;
			}
		case AF_INET6:
			switch(usa.family()) {
				case AF_INET:
					return false;
				case AF_INET6: {
					// uso memcmp() porque guardamos el dato en net order,
					// y entonces va a andar para todos los tipos de procesador
					int x = memcmp( &(    sa.in6->sin6_addr),
										 &(usa.sa.in6->sin6_addr),
										 sizeof(in6_addr) );
					if( x<0 ) return true;
					if( x>0 ) return false;
					return ntohs(sa.in6->sin6_port) < ntohs(usa.sa.in6->sin6_port);
				}
				case AF_UNIX:
				case AF_UNSPEC:
				default:
					return true;
		 }
		case AF_UNIX:
			switch(usa.family()) {
				case AF_INET:
				case AF_INET6: return false;
				case AF_UNIX: {
					register char *px(sa.un->sun_path);
					register char *py(((sockaddr_un*)usa.ptr())->sun_path);
					while( *px && (*px==*py) ) { ++px; ++py; }
					return *px < *py;
					}
				case AF_UNSPEC:
				default:       return true;
			}
		case AF_UNSPEC:
		default:             return false;
	}
}

size_t UpSockAddr::hash() const {
	switch(family()) {
		case AF_INET:
			return ((size_t)(sa.in->sin_port))^(~(size_t)(sa.in->sin_addr.s_addr));
		case AF_INET6: {
			register size_t h((size_t)sa.in6->sin6_port);
			h<<=1; h^=(size_t)sa.in6->sin6_addr.s6_addr32[0];
			h<<=1; h^=(size_t)sa.in6->sin6_addr.s6_addr32[1];
			h<<=1; h^=(size_t)sa.in6->sin6_addr.s6_addr32[2];
			h<<=1; h^=(size_t)sa.in6->sin6_addr.s6_addr32[3];
			return h;
			}
		case AF_UNIX: {
			register char* p((char*)(sa.un->sun_path));
			register size_t h(0);
			while( *p ) h = 5*h + *(p++);
			return h;
			}
		default:
			return 0;
	}
}

unsigned short UpSockAddr::port() const {
	switch(family()) {
		case AF_INET: return ntohs(sa.in->sin_port);
		case AF_INET6: return ntohs(sa.in6->sin6_port);
		default: return 0;
	}
}


std::ostream& operator<<(std::ostream& o, const UpSockAddr& upa) {
	if(upa.sa.a) {
		switch(upa.sa.a->sa_family) {
			case AF_INET:  o << "IPv4 " << upa.string(); break;
			case AF_INET6: o << "IPv6 " << upa.string(); break;
			case AF_UNIX:  o << "UNIX " << upa.string(); break;
			default: o << "(error in address)";
		}
	} else o << "(empty address)";
	return o;
}

std::string& operator<<(std::string& o, const UpSockAddr& upa) {
	if(upa.sa.a) {
		switch(upa.sa.a->sa_family) {
			case AF_INET:  o << "IPv4 " << upa.string(); break;
			case AF_INET6: o << "IPv6 " << upa.string(); break;
			case AF_UNIX:  o << "UNIX " << upa.string(); break;
			default: o << "(error in address)";
		}
	} else o << "(empty address)";
	return o;
}



//////////////////////////////////////////////////////////////////////////
// UpSocket methods

UpSocket::UpSocket() :
	sd(-1), family(AF_UNSPEC), nonBlocking(false), direct(false), async(false),
	sockType(SOCK_STREAM) { }

UpSocket::~UpSocket() { if(sd>=0) ::close(sd); }

void UpSocket::setSendTimeout(const UpTimeVal& tv)
	{ sendTimeout=tv; if(sd>=0) setFlags(); }

void UpSocket::setRecvTimeout(const UpTimeVal& tv)
	{ recvTimeout=tv; if(sd>=0) setFlags(); }

void UpSocket::setDirect(bool t)			{ direct = t; if(sd>=0) setFlags(); }
void UpSocket::setNonBlocking(bool t)	{ nonBlocking = t; if(sd>=0) setFlags(); }
void UpSocket::setAsync(bool t)			{ async = t; if(sd>=0) setFlags(); }

bool UpSocket::setFlags() {
	// que pasa si era no nulo y se quiere bajar a cero?
	if( (sendTimeout!=(time_t)0) &&
		 setsockopt(sd,SOL_SOCKET,SO_SNDTIMEO,(timeval*)&sendTimeout,sizeof(timeval))
     ) return false;
	// que pasa si era no nulo y se quiere bajar a cero?
	if( (recvTimeout!=(time_t)0) &&
		 setsockopt(sd,SOL_SOCKET,SO_RCVTIMEO,(timeval*)&recvTimeout,sizeof(timeval))
     ) return false;
	int flags = fcntl( sd , F_GETFL , 0 );
	int newflags = flags;
	if( nonBlocking ) newflags |=   O_NONBLOCK   ;
	else					newflags &= (~O_NONBLOCK ) ;
	if( direct )		newflags |=   O_DIRECT     ;
	else					newflags &= (~O_DIRECT   ) ;
	if( async )			newflags |=   O_ASYNC      ;
	else					newflags &= (~O_ASYNC    ) ;
	if( newflags == flags ) return true;
	return ! fcntl( sd , F_SETFL , newflags );
}

void UpSocket::showFcntlFlags() {
	if( sd<0 ) {
		std::cout << "Socket is not opened. Can't get flags." << std::endl;
	} else {
		int flags = fcntl( sd , F_GETFL , 0 );
		std::cout << "O_ASYNC: " << (bool)(flags&O_ASYNC) << std::endl;
		flags = flags & (~O_ASYNC);
		std::cout << "O_NONBLOCK: " << (bool)(flags&O_NONBLOCK) << std::endl;
		flags = flags & (~O_NONBLOCK);
		std::cout << "O_DIRECT: " << (bool)(flags&O_DIRECT) << std::endl;
		flags = flags & (~O_DIRECT);
		#ifdef DEBUG
		if(flags) std::cout << "there are still some unknown flags: "
								  << flags << std::endl;
		#endif
	}
}

bool UpSocket::setDatagram()	{
	if(sd<0) { sockType = SOCK_DGRAM; return true; }
	return sockType==SOCK_DGRAM;
}

bool UpSocket::setStream() {
	if(sd<0) { sockType = SOCK_STREAM; return true; }
	return sockType==SOCK_STREAM;
}

/////////////////////////////////////////////////////////////////////////////
int UpSocket::fd() {
	if(sd>=0) return sd;
	family = localSockAddr.family();
	if(family==AF_UNSPEC) family = remoteSockAddr.family();
	if(family==AF_UNSPEC) return sd = -1;
	sd = socket(family, sockType, 0);
	if(sd<0) return sd;
	// to be able to reuse address and port immediately after closing
   int opt = 1; void *pOpt = (void *)&opt;
   setsockopt(sd,SOL_SOCKET,SO_REUSEADDR,pOpt,sizeof(int));
	setFlags();
	return sd;
}


/////////////////////////////////////////////////////////////////////////////
// bind methods

// principal bind method, all others call this one
bool UpSocket::bind() {
	if( ! fd() ) return false;
	if(family==AF_UNIX) ::unlink(localSockAddr.c_str());
	return ! ::bind( sd , localSockAddr.ptr() , localSockAddr.size() );
	// TODO: unlink an already existing socket
}

// bind a partir de un UpSockAddr
bool UpSocket::bind(const UpSockAddr& localSA) {
	localSockAddr = localSA;
	return bind();
}
// a partir de un struct sockaddr
bool UpSocket::bind(const sockaddr* localSA) {
	localSockAddr = localSA;
	return bind();
}

bool UpSocket::bind(const char* localHostAddr,unsigned short localPort) {
	if( localSockAddr.hostPort(localHostAddr,localPort) ) return bind();
	return false;
}

bool UpSocket::bind(const std::string& localHostAddr,unsigned short localPort) {
	if( localSockAddr.hostPort(localHostAddr,localPort) ) return bind();
	return false;
}

bool UpSocket::bind(const std::string& localHostAddr) {
	if( localSockAddr.hostPort(localHostAddr) ) return bind();
	return false;
}

bool UpSocket::bind(const char* localHostAddr) {
	if( localSockAddr.hostPort(localHostAddr) ) return bind();
	return false;
}




bool UpSocket::bindInterface(const char* interf, unsigned short localPort,int family) {
	localSockAddr.allocSA(family);
	if( fd()<0 ) return false;
	struct ifreq ifr;
   std::strncpy(ifr.ifr_name,interf,IF_NAMESIZE);
	ifr.ifr_name[IF_NAMESIZE-1]=0;
	if( ioctl(sd,SIOCGIFADDR,&ifr) < 0 ) { close(); return false; }
	switch(family) {
		case AF_INET:
			((sockaddr_in*)(&ifr.ifr_addr))->sin_port = htons(localPort);
			localSockAddr = (sockaddr_in*)&ifr.ifr_addr;
			return bind();
		case AF_INET6:
			((sockaddr_in6*)(&ifr.ifr_addr))->sin6_port = htons(localPort);
			localSockAddr = (sockaddr_in6*)&ifr.ifr_addr;
			return bind();
		default:
			close(); return false;
	}
	return ! ::bind( sd , localSockAddr.ptr() , localSockAddr.size() );
}



/////////////////////////////////////////////////////////////////////////////
bool UpSocket::listen() {
	// bind should be called before listen in order to get a socket descriptor
	if( sd<0 || ::listen(sd,16) ) return false;
	return true;
}

bool UpSocket::listen(int queueSize) {
	// queueSize: Now it specifies the queue  length  for  completely  estab-
	// lished  sockets waiting to be accepted, instead of the number of incom-
	// plete connection requests.

	// bind should be called before listen in order to get a socket descriptor
	if( sd<0 || ::listen(sd,queueSize) ) return false;
	return true;
}

/////////////////////////////////////////////////////////////////////////////
bool UpSocket::accept(UpSocket& connectedSocket) {
	// bind and listen should be called before accept !
	if( sd<0 ) return false;
	connectedSocket.family			= family;
	connectedSocket.localSockAddr	= localSockAddr;
	UpSockAddr& remoteSA			= connectedSocket.remoteSockAddr;
	remoteSA.allocSA(family);
	socklen_t saddrLen		= remoteSA.size();
	int newSd = ::accept(sd,remoteSA.ptr(),&saddrLen);
	if( newSd <= 0 ) return false;
	connectedSocket.sd				= newSd;
	connectedSocket.setFlags(); // it needs to apply its previous settings
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// connect methods

// principal connect method, all others call this one
bool UpSocket::connect() {
	if( ! fd() ) return false;
	return ! ::connect(sd,remoteSockAddr.ptr(),remoteSockAddr.size());
}

bool UpSocket::connect(const sockaddr* remoteSA) {
	remoteSockAddr = remoteSA;
	return connect();
}

bool UpSocket::connect(const UpSockAddr& remoteSA) {
	remoteSockAddr = remoteSA;
	return connect();
}

bool UpSocket::connect(const std::string& remoteHostAddr, unsigned short remotePort) {
	if( remoteSockAddr.hostPort(remoteHostAddr,remotePort) ) return connect();
	return false;
}

bool UpSocket::connect(const char* remoteHostAddr, unsigned short remotePort) {
	if( remoteSockAddr.hostPort(remoteHostAddr,remotePort) ) return connect();
	return false;
}

bool UpSocket::connect(const std::string& remoteHostAddr) {
	if( remoteSockAddr.hostPort(remoteHostAddr) ) return connect();
	return false;
}

bool UpSocket::connect(const char* remoteHostAddr) {
	if( remoteSockAddr.hostPort(remoteHostAddr) ) return connect();
	return false;
}

/////////////////////////////////////////////////////////////////////////////
void UpSocket::close() {
		if(sd>=0) { ::close(sd); sd = -1; }
}

/////////////////////////////////////////////////////////////////////////////
int UpSocket::write(const std::string& s) {
	if( sd<0 ) return -1;
	return ::write(sd,s.c_str(),s.size());
}

int UpSocket::write(const char* buf, size_t len) {
	if( sd<0 ) return -1;
   return ::write(sd,buf,len);
}

int UpSocket::writeUrg(const std::string& s) {
	if( sd<0 ) return -1;
	return ::send(sd,s.c_str(),s.size(),MSG_OOB);
}

int UpSocket::writeUrg(const char* buf, size_t len) {
	if( sd<0 ) return -1;
   return ::send(sd,buf,len,MSG_OOB);
}

int UpSocket::sendto(const char* buf, size_t len, const UpSockAddr& remoteSA) {
	if( sd<0 ) return -1;
	return ::sendto(sd,buf,len,0,remoteSA.ptr(),remoteSA.size());
}


int UpSocket::sendtoUrg(const char* buf, size_t len, const UpSockAddr& remoteSA) {
	if( sd<0 ) return -1;
	return ::sendto(sd,buf,len,MSG_OOB,
									remoteSA.ptr(),remoteSA.size());
}

int UpSocket::sendto(const std::string& s, const UpSockAddr& remoteSA) {
	if( sd<0 ) return -1;
   return ::sendto(sd,s.c_str(),s.size(),0,remoteSA.ptr(),remoteSA.size());
}


int UpSocket::sendtoUrg(const std::string& s, const UpSockAddr& remoteSA) {
	if( sd<0 ) return -1;
   return ::sendto(sd,s.c_str(),s.size(),MSG_OOB,
									remoteSA.ptr(),remoteSA.size());
}

int UpSocket::read(char* buf,size_t len) {
	if( sd<0 ) return -1;
	return ::read(sd,buf,len);
}

int UpSocket::readUrg(char* buf,size_t len) {
	if( sd<0 ) return -1;
	return ::recv(sd,buf,len,MSG_OOB);
}


int UpSocket::recvFrom(char* buf, size_t len, UpSockAddr& remoteSA){
	if( sd<0 ) return -1;
	remoteSA.allocSA(family);
	socklen_t saddrLen = remoteSA.size();
	return ::recvfrom(sd,buf,len,0,remoteSA.ptr(),&saddrLen);
}


int UpSocket::recvFromUrg(char* buf, size_t len, UpSockAddr& remoteSA){
	if( sd<0 ) return -1;
	remoteSA.allocSA(family);
	socklen_t saddrLen = remoteSA.size();
	return ::recvfrom(sd,buf,len,MSG_OOB,
											remoteSA.ptr(),&saddrLen);
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

void UpNet::checkPrivateData() {
	if(  (ip.addrType!=m.addrType) or
	     (ip.addrType!=AF_INET and ip.addrType!=AF_INET6)
     )
	  { ip.clear(); m.clear(); }
}



UpNet::UpNet()
: ip(zero_in_addr),m(zero_in_addr)
{}

UpNet::UpNet(const UpNet& net)
: ip(net.ip),m(net.m)
{}

UpNet::UpNet(const UpAddress& baseAddr,const UpAddress& mask)
: ip(baseAddr),m(mask)
{ checkPrivateData(); }

UpNet::UpNet(const UpAddress& baseAddr,unsigned int maskLen)
: ip(baseAddr)
{
	m.setMask(maskLen,ip.addrType); // works only for INET or INET6
	if( m.addrType==AF_UNSPEC ) ip.clear();
}

UpNet::UpNet(in_addr baseAddr,in_addr mask)
:ip(baseAddr),m(mask)
{ }

UpNet::UpNet(const in6_addr& baseAddr,const in6_addr& mask)
: ip(baseAddr),m(mask)
{ }

UpNet::UpNet(const std::string& net)
{ set(net); }

void UpNet::set(const UpNet& n) {
	ip=n.ip; m=n.m;
}


void UpNet::set(const UpAddress& baseAddr,const UpAddress& mask) {
	ip=baseAddr; m=mask;
	checkPrivateData();
}


void UpNet::set(const UpAddress& baseAddr,unsigned int maskLen) {
	m.setMask(maskLen,baseAddr.addrType); // works only for INET or INET6
	if( m.addrType!=AF_UNSPEC ) ip=baseAddr; else ip.clear();
}

void UpNet::set(in_addr baseAddr,in_addr mask) {
	ip=baseAddr; m = mask;
}


void UpNet::set(const in6_addr& baseAddr,const in6_addr& mask) {
	ip=baseAddr; m = mask;
}

bool UpNet::set(const std::string& net) {
	size_t pos=net.find('/');
	if( pos!=std::string::npos ) { // slash found
		ip=net.substr(0,pos++);
		if( net.find_first_of(".:",pos)!=std::string::npos ) {
			m = net.c_str()+pos;
		} else {
			int len = std::atoi(net.c_str()+pos);
			if( len>=0 ) m.setMask((unsigned int)len,ip.addrType);
			else { ip.clear(); m.clear(); return false; }
		}
	} else { // slash not found: one-host net
		ip=net;
		m.setMask(128,ip.addrType);
	}
	if( (ip.addrType!=m.addrType) or !(ip.addrType==AF_INET or ip.addrType==AF_INET6) )
	 { ip.clear(); m.clear(); return false; }
	return true;
}



/*
UpNetRelation UpNet::check(const UpAddress& addr) const {
	if(addr.family()!=family()) return UPNR_INVALID_RELATION;
	if(addr<ip)      return UPNR_BEFORE; // addr esta antes que *this
	if((addr&m)==ip) return UPNR_SUBNET; // addr es una subnet de *this
	else             return UPNR_AFTER;  // addr red esta atras de *this
}
*/

UpNetRelation UpNet::check(in_addr addr) const {
	if(family()!=AF_INET)  return UPNR_INVALID_RELATION;
	if(ntohl(addr.s_addr)<ntohl(ip.upAddr.a4.s_addr)) return UPNR_BEFORE;
	if((addr.s_addr&m.upAddr.a4.s_addr)==ip.upAddr.a4.s_addr) return UPNR_SUBNET;
	return UPNR_AFTER;
}


UpNetRelation UpNet::check(const in6_addr& _addr) const {
	if(family()!=AF_INET6)  return UPNR_INVALID_RELATION;
	const UpAddress::addrUnion& addr( *(UpAddress::addrUnion*)&_addr );
	if( addr.be128.high!=ip.upAddr.be128.high ) {
		if( be64toh(addr.be128.high)<be64toh(ip.upAddr.be128.high) ) return UPNR_BEFORE;
	} else {
		if( be64toh(addr.be128.low)<be64toh(ip.upAddr.be128.low)   ) return UPNR_BEFORE;
	}
	if( (addr.be128.high&m.upAddr.be128.high)==ip.upAddr.be128.high and
		(addr.be128.low &m.upAddr.be128.low )==ip.upAddr.be128.low ) return UPNR_SUBNET;
	return UPNR_AFTER;
}


UpNetRelation UpNet::check(const UpNet& r) const {
	if(r.family()!=family()) return UPNR_INVALID_RELATION;
   if( m <= r.m ) { // r tine menor o igual tamano que *this
      if(r.ip < ip) return UPNR_BEFORE;    // r esta antes que *this
      if((r.ip & m) == ip) {
         if( m < r.m ) return UPNR_SUBNET; // r es una subnet de *this
         else return UPNR_EQUAL;           // r es igual a *this
      } else return UPNR_AFTER;            // r esta despues de *this
   } else {                // r tiene mayor tamano que *this
      if(ip < r.ip) return UPNR_AFTER;     // r esta despues que *this
      if((ip & r.m) == r.ip) return UPNR_SUPERNET;// r incluye a esta red
      return UPNR_BEFORE;                  // r esta antes que *this
   }
}


std::string UpNet::relString[] = {
	std::string("BEFORE"),
	std::string("AFTER"),
	std::string("SUPERNET"),
	std::string("SUBNET"),
	std::string("EQUAL"),
	std::string("INVALID_RELATION")
};



// preincrement operators
UpAddress& UpAddress::operator++() {
	if( addrType==AF_INET ) {
		chauBuf();
		register unsigned int aux = ntohl(upAddr.a4.s_addr);
		upAddr.a4.s_addr = htonl( ++aux );
		return *this;
	}
	if( addrType==AF_INET6 ) {
		chauBuf();
		if( upAddr.be128.low!=allones64ull ) {
			register unsigned long long aux = be64toh(upAddr.be128.low);
			upAddr.be128.low = htobe64( ++aux );
		} else {
			upAddr.be128.low  = 0;
			register unsigned long long aux = be64toh(upAddr.be128.high);
			upAddr.be128.high = htobe64( ++aux );
		}
	}
	return *this;
}


UpAddress& UpAddress::operator--() {
	if( addrType==AF_INET ) {
		chauBuf();
		register unsigned int aux = ntohl(upAddr.a4.s_addr);
		upAddr.a4.s_addr = htonl( --aux );
		return *this;
	}
	if( addrType==AF_INET6 ) {
		chauBuf();
		if( upAddr.be128.low!=zero64ull ) {
			register unsigned long long aux = be64toh(upAddr.be128.low);
			upAddr.be128.low = htobe64( --aux );
		} else {
			upAddr.be128.low  = allones64ull;
			register unsigned long long aux = be64toh(upAddr.be128.high);
			upAddr.be128.high = htobe64( --aux );
		}
	}
	return *this;
}


UpAddress& UpAddress::operator-=(const UpAddress& upa) {
	if( addrType!=upa.addrType )
		return *this;
	if( addrType==AF_INET ) {
		chauBuf();
		upAddr.a4.s_addr = htonl( ntohl(upAddr.a4.s_addr) - ntohl(upa.upAddr.a4.s_addr) );
		return *this;
	}
	if( addrType==AF_INET6 ) {
		chauBuf();
		unsigned long long ah = be64toh(upAddr.be128.high);
		unsigned long long al = be64toh(upAddr.be128.low);
		unsigned long long bh = be64toh(upa.upAddr.be128.high);
		unsigned long long bl = be64toh(upa.upAddr.be128.low);
		if( al<bl ) --ah;
		al -= bl;
		ah -= bh;

		upAddr.be128.high = htobe64(ah);
		upAddr.be128.low  = htobe64(al);
	}
	return *this;
}


UpAddress& UpAddress::operator+=(const UpAddress& upa) {
	if( addrType!=upa.addrType )
		return *this;
	if( addrType==AF_INET ) {
		chauBuf();
		upAddr.a4.s_addr = htonl( ntohl(upAddr.a4.s_addr) + ntohl(upa.upAddr.a4.s_addr) );
		return *this;
	}
	if( addrType==AF_INET6 ) {
		chauBuf();
		unsigned long long ah = be64toh(upAddr.be128.high);
		unsigned long long al = be64toh(upAddr.be128.low);
		unsigned long long bh = be64toh(upa.upAddr.be128.high);
		unsigned long long bl = be64toh(upa.upAddr.be128.low);
		al += bl;
		ah += bh;
		if( al<bl ) ++ah;

		upAddr.be128.high = htobe64(ah);
		upAddr.be128.low  = htobe64(al);
	}
	return *this;
}

UpNet::~UpNet()
{ }


void UpNet::clear() {
	ip.clear(); m.clear();
}

