/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpNaptrQuery.h>
#include <iostream>

UpNaptrQuery::UpNaptrQuery(const std::string& _domain, const std::string& _uri,
                            const std::string& _service)
: UpDnsQuery(_domain,ns_t_naptr,QUERY_USE_CALLBACK,QUERY_NONBLOCKING),
  nstate(SNAPTR_NAPTR), uri(_uri), service(_service)
{}

void UpNaptrQuery::completed() {

	if( getStatus()!=QUERY_READY ) {
		nstate = SNAPTR_COMPLETE_FAILED;
		msg << "Failed " << dnsQueryType(getType()) << " query for "
				<< getDomain() << '\n';
		endOfAlgorithm();
		return;
	}
	if( getCount()==0 ) { // current query failed
		nstate = SNAPTR_COMPLETE_FAILED;
		msg << "No records for " << dnsQueryType(getType()) <<getDomain() << '\n';
		endOfAlgorithm();
		return;
	}
	
	const UpResourceRecord *aux(&(getFirst())), *rr(0);
	unsigned short order(65535), pref(65535), weight(65535);

	switch(nstate) {

		case SNAPTR_NAPTR:

			for( ; aux->getType()!=ns_t_invalid; aux=&(getNext()) ) {
        		if( aux->getService()==service ) {
           		if( aux->getOrder()<order ) {
						rr    = aux;
						order = rr->getOrder();
						pref  = rr->getPreference();
					} else if( (aux->getOrder()==order) &&
								  (aux->getPreference()<pref) ) {
						rr    = aux;
						pref  = rr->getPreference();
           		}
				}
     		}
			if( rr==0 ) {
				nstate = SNAPTR_COMPLETE_FAILED;
				msg << "No matching service ("<< service <<") for "
					 << dnsQueryType(getType()) << getDomain() << '\n';
				endOfAlgorithm();
				return;
			}
			
			if( rr->getFlags()=="s") { // next step is a querytype=SRV
				msg << "Record found: " << *rr << "\nNext step: SRV query\n" ;
				srv    = rr->getReplacement();
           	nstate = SNAPTR_SRV;
				set( srv , ns_t_srv , QUERY_USE_CALLBACK , QUERY_NONBLOCKING );
				getResolverPtr()->query(*this);
				return;
			}
			if( rr->getFlags()=="a" ) { // next step is a querytype=A
				msg << "Record found: " << *rr << "\nNext step: A query\n" ;
				a      = rr->getReplacement();	
				nstate = SNAPTR_A;
				set( a , ns_t_a , QUERY_USE_CALLBACK , QUERY_NONBLOCKING );
				getResolverPtr()->query(*this);
				return;
			}
			if(rr->getFlags() == "u") { // next step is uri substitution
				msg << "Record found: " << *rr <<"\nNext step: URI substitution\n" ;
				UpSubstitute sub( rr->getRegex() );
				if( ! sub.exec( uri ) ) {
					msg << "Substitution failed: " << sub.getError() << '\n';
					nstate = SNAPTR_COMPLETE_FAILED;
					endOfAlgorithm();
					return;
				}
				newuri = sub.getResult();
				nstate = SNAPTR_COMPLETE_OK;
				msg << "Substitution Ok\n";
				endOfAlgorithm();
				return;
			}
			msg << "Record found, unknown flags: " << *rr << '\n';
			nstate = SNAPTR_COMPLETE_FAILED;
			endOfAlgorithm();
			return;

		case SNAPTR_SRV:

			for( ; aux->getType()!=ns_t_invalid; aux=&(getNext()) ) {
				if( aux->getPreference() < pref ) {
					pref   = aux->getPreference();
					weight = aux->getWeight();
					rr     = aux;
				} else if( ( aux->getPreference()==pref ) &&
							  ( aux->getWeight() > weight )      ) {
					weight = aux->getWeight();
					rr =     aux;
				}
        	}
			if( rr==0 ) {
				nstate = SNAPTR_COMPLETE_FAILED;
				msg << "No valid record for SRV " << getDomain() << '\n';
				endOfAlgorithm();
				return;
			}
			msg << "Record found: " << *rr << "\nNext step: A query\n" ;
			port = rr->getPort();
			nstate = SNAPTR_A;
			set( rr->getTarget() , ns_t_a , QUERY_USE_CALLBACK,QUERY_NONBLOCKING );
			getResolverPtr()->query(*this);
			return;			
	
		case SNAPTR_A:

			for( ; aux->getType()!=ns_t_invalid; aux=&(getNext()) )
				{ rr = aux; break; }
			if( rr==0 ) {
				nstate = SNAPTR_COMPLETE_FAILED;
				msg << "No valid record for A " << getDomain() << '\n';
				endOfAlgorithm();
				return;
			}
			msg << "Record found: " << *rr << '\n';
			upa = rr->getAddr();
			nstate = SNAPTR_COMPLETE_OK;
			endOfAlgorithm();
			return;
		
		default:
			endOfAlgorithm();
			break; // do nothing
	}
}

