/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpRegex.h>
#include <iostream>

std::string UpRegex::emptyString = "";

// constructor por copia
UpRegex::UpRegex( const UpRegex& upr )
: nmatch(upr.nmatch) , pmatch(0), cflags(upr.cflags), matched(upr.nmatch),
  regexString(upr.regexString), init(false)
{
	if(!upr.init) return; // copiando desde una clase con errores
	// si se supiera exactamente el tamaño de la memoria alocada desde regcomp,
	// se puede reemplazar este regcomp por memcpy's que es mas rapido
	regexerror = regcomp( &regex , regexString.c_str() , cflags );
	if(regexerror) return; // initialization failed
	if(nmatch>0) pmatch = new regmatch_t[nmatch];
	init=true;
}

// constructor normal
UpRegex::UpRegex()
: nmatch(0) , pmatch(0), cflags(0), init(false)
{}


UpRegex::UpRegex( const char* _regexString , size_t atomCount, bool icase )
: nmatch(atomCount) , pmatch(0), cflags(REG_EXTENDED|REG_NEWLINE), matched(atomCount),
  regexString(_regexString), init(false)
{
	if(icase)     cflags |= REG_ICASE;
	if(nmatch==0) cflags |= REG_NOSUB;
	regexerror = regcomp( &regex , _regexString , cflags );
	if(regexerror) return; // initialization failed
	if(nmatch>0) pmatch = new regmatch_t[nmatch];
	init=true;
}

UpRegex::UpRegex( const std::string& _regexString , size_t atomCount, bool icase )
: nmatch(atomCount) , pmatch(0), cflags(REG_EXTENDED|REG_NEWLINE), matched(atomCount),
  regexString(_regexString), init(false)
{
	if(icase)     cflags |= REG_ICASE;
	if(nmatch==0) cflags |= REG_NOSUB;
	regexerror = regcomp( &regex , _regexString.c_str() , cflags );
	if(regexerror) return; // initialization failed
	if(nmatch>0) pmatch = new regmatch_t[nmatch];
	init=true;
}

bool UpRegex::compile( const char* _regexString , size_t atomCount, bool icase ) {
	if( pmatch ) { delete[] pmatch; pmatch = 0; }
	if(  init  ) { regfree( &regex ); init = false; }
	regexString=_regexString;
	matched.resize(atomCount);
	nmatch = atomCount;
	cflags = REG_EXTENDED|REG_NEWLINE;
	if(icase)     cflags |= REG_ICASE;
	if(nmatch==0) cflags |= REG_NOSUB;
	regexerror = regcomp( &regex , _regexString , cflags );
	if(regexerror) return false; // initialization failed
	if(nmatch>0) pmatch = new regmatch_t[nmatch];
	init=true;
	return true;
}


UpRegex::~UpRegex()
{
	if( pmatch ) delete[] pmatch;
	if(  init  ) regfree( &regex );
}

bool UpRegex::exec(const char* data){
	if( !init ) return false; // can't exec, not initialized
	regexerror = regexec( &regex , data ,  nmatch , pmatch, 0);
	if( regexerror == REG_NOMATCH ) return false;

	for(unsigned int i=0; i<nmatch;i++){
		if(pmatch[i].rm_so!=-1) {
			const char *start = data + pmatch[i].rm_so;
			const char *end   = data + pmatch[i].rm_eo;
			matched[i].assign(start,end-start);
		} else {
			matched[i].clear();
		}
	}
	return true;
}

const std::string& UpRegex::getMatch(unsigned int index){
	return ( !init || regexerror || index>=nmatch ) ?
					emptyString : matched[index];
}

int UpRegex::getMatchStart(unsigned int index) {
	return ( !init || regexerror || index>=nmatch ) ?
					(-1) : pmatch[index].rm_so;
}

int UpRegex::getMatchEnd(unsigned int index) {
	return ( !init || regexerror || index>=nmatch ) ?
					(-1) : pmatch[index].rm_eo;
}

const std::string& UpRegex::getError(){
	if(regexerror) {
		char errbuf[256];
		regerror(regexerror , &regex , errbuf , 255 );
		errorString = errbuf;
		return errorString;
	}
	return emptyString;
}

