/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Ignacio Kaufmann for his work in this class

#include <UpTools/UpResolver.h>
#include <UpTools/UpLog.h>

#include <cerrno>
#include <cstring>
#include <list>
#include <cassert>

UpResolver::UpResolver()
{
	// get data from libresolv (see /usr/include/resolver.h)
	res_init();
	retrans			= (time_t)_res.retrans; // retransmition time interval
	retry				= _res.retry;				// number of times to retransmit
	dnsServerAddr	= &(_res.nsaddr); // we use only first nameserver for now
	// initialize UpResolver
	init();
}

UpResolver::UpResolver( const std::string& _nameserver,time_t _retrans,
		                  unsigned int _retry )
: dnsServerAddr( _nameserver, 53), retrans(_retrans), retry(_retry)
{
	init();
}


bool UpResolver::init() {
	// timeouts administration	
	tickInterval = retrans/2;  // in order not to retramsit too late after tick

	// bind to all interfaces, let OS decide what port to assign us
	sock.setDatagram(); // UDP oriented socket
	sock.setNonBlocking(false); // must block for incoming responses
	sock.setRecvTimeout(tickInterval/2); // to check ticks if there is no activity
	if( !sock.bind("0.0.0.0",0) ) {
		UPLOG(LOG_ERR,"UpResolver::UpResolver() "
					"error binding to all interfaces, any port" );
		return false;
	}
	// start listener thread
	create();
	UPLOG(LOG_DEBUG,"UpResolver::init() nameServer:%s retransmitInterval:%usecs"
			 " retryTimes:%u", dnsServerAddr.c_str_2dots(), (unsigned int)retrans,
			retry );
	return true;
}



UpResolver::~UpResolver() {
	cancel(); // stops thread
	sock.close();
	std::map<unsigned int,UpDnsQuery*>::iterator qIt;
	std::list<UpDnsQuery*> byeQueries;
	UpDnsQuery* qptr;
	mx.lock();
		for( qIt=queryMap.begin(); qIt!=queryMap.end(); ++qIt ) 
			byeQueries.push_back(qIt->second);
	mx.unlock();
	while( byeQueries.begin()!=byeQueries.end() ) {
		qptr = byeQueries.front();
		terminateQuery(*qptr);
		byeQueries.pop_front();
	}
}




bool UpResolver::retransmitQuery(UpDnsQuery& q , time_t now) {
	if( q.status!=QUERY_SEARCHING ) {
		UPLOG(LOG_ERR,"UpResolver::retransmitQuery() cannot retransmit "
				" a query id %u because it is not in state SEARCHING",q.id);
		return false;
	}
	unsigned int n(sock.sendto((char*)(q.query_buff),q.querySize,dnsServerAddr));
	if( n!=q.querySize ) {
		q.status = QUERY_SOCK_ERROR;
		UPLOG(LOG_ERR,"UpResolver::retransmitQuery() error in socket while "
				"sending dns query id %u",q.id);
		return false;
	}
	--q.retra;
	q.transmitTime = now;
	return true;
}


bool UpResolver::selfQuery(UpDnsQuery& q) {
	// UpDnsQuery is being re-posted from UpDnsQuery::completed()
	// from within UpResolver::run()
	assert( q.rstatus==QR_MARKED_FOR_COMPLETION );
	unsigned int n;
	bool ok = false;

	mx.lock();
		if( ! q.prepare() ) {
			UPLOG(LOG_ERR,"UpResolver::selfQuery() error preparing query");
		} else {
			n = sock.sendto((char*)(q.query_buff),q.querySize,dnsServerAddr);
			// envio lo que tenia que enviar? si no, hubo un socket error
			if( n != q.querySize ) {
				q.status = QUERY_SOCK_ERROR;
				UPLOG(LOG_ERR,"UpResolver::selfQuery() error in socket while "
					"sending dns query id %u",q.id);
			} else {
				q.rstatus		= QR_REPOSTED;
				q.status		= QUERY_SEARCHING;
				q.transmitTime	= time(0);
				q.retra			= retry-1;
				queryMap[q.id]	= &q;
				ok             = true;
				UPDEBUG("UpResolver::selfQuery() query id %u inserted, "
									"query count: %u",q.id,(unsigned int)queryMap.size() );
			}
		}
	mx.unlock();
	return ok;
}



bool UpResolver::query(UpDnsQuery& q) {
	// if posting from UpDnsQuery::completed() we need a special treatment
	if( checkSelf() ) return selfQuery(q);

	unsigned int n;
	bool ok = false;

	mx.lock();
		if( q.rstatus!=QR_IDLE ) {
			UPLOG(LOG_ERR,"UpResolver::query() cannot query a non-idle query");
		} else if( ! q.prepare() ) {
			UPLOG(LOG_ERR,"UpResolver::query() error preparing query");
		} else {
			n = sock.sendto((char*)(q.query_buff),q.querySize,dnsServerAddr);
			// envio lo que tenia que enviar? si no, hubo un socket error
			if( n != q.querySize ) {
				q.status = QUERY_SOCK_ERROR;
				UPLOG(LOG_ERR,"UpResolver::query() error in socket while "
					"sending dns query id %u",q.id);
			} else {
				q.rstatus		= QR_AWAITING_RESPONSE;
				q.status       = QUERY_SEARCHING;
				q.transmitTime = time(0);
				q.retra			= retry-1;
				q.resolverPtr  = this;
				queryMap[q.id] = &q;
				q.allowWaiting();
				ok             = true;
				UPDEBUG("UpResolver::query() query id %u inserted, "
									"query count: %u",q.id,(unsigned int)queryMap.size() );
			}
		}
	mx.unlock();
	if( ok ) {
		return ( q.blockingType==QUERY_BLOCKING ) ? q.waitCompletion() : true;
	}
	return false;
}



/////////////////////////////////////////////////////////////////////////////

// este es el codigo del thread que escucha en el socket
void* UpResolver::run() { 

	UpSockAddr remoteSA;
	int n;
	unsigned int rid;
	time_t now;
	UpDnsQuery* qptr;
	std::list<UpDnsQuery*> expiredQueries;
	std::list<UpDnsQuery*>::iterator expIt;
	std::map<unsigned int,UpDnsQuery*>::iterator qIt,auxIt;

	time(&lastTick);

	while(true) {


		n = sock.recvFrom((char*)(buff),NS_PACKETSZ,remoteSA);

		if( n<0 ) {
			if( errno==EAGAIN ) {
				//	UPDEBUG("UpResolver::run() socket timer timeout"); //DO NOTHING
			} else {
				UPLOG(LOG_ERR,"UpResolver::run() read error in socket(%d): %s",
									errno,strerror(errno));
			}
		} else if( n==0 ) {
			UPLOG(LOG_ERR,"UpResolver::run() error: EOF received in udp socket?");
		} else if( n>0 ) {

			rid  = ns_get16(buff);
			qptr = 0;

			mx.lock();
				qIt = queryMap.find(rid);
				if( qIt!=queryMap.end() ) {
					qptr = qIt->second;
					if( qptr->rstatus==QR_AWAITING_RESPONSE ) {
						// query was waiting for response or or timeout
						qptr->rstatus = QR_MARKED_FOR_COMPLETION;
						queryMap.erase(qIt);
					} else {
						qptr = 0;
					}
				}
				#ifdef DEBUG
				unsigned int qsz = queryMap.size();
				#endif
			mx.unlock();

			if( qptr==0 ) {
				UPLOG(LOG_NOTICE,"UpReslover::run() unsolicited query id %u "
						"response from %s",rid,remoteSA.c_str());
			} else {
				memcpy((char*)qptr->answer_buff,(char*)buff,n);
				qptr->answerSize  = n; 
				qptr->status      = QUERY_RECEIVED;
				qptr->initialResponseParse(); // sets qptr->status to READY or PARSE_ERROR
				// we run UpDnsQuery::completed() outside the lock
				if( qptr->callbackType==QUERY_USE_CALLBACK )
					qptr->completed();	 // may reQuery itself
				UPDEBUG("UpResolver::run() query id %u completed, remaining "
							"queries %u",rid,qsz);
				mx.lock(); // we lock again here to protect qptr->rstatus
					// if query re-posts itself from within completed()
					// we need a special treatment
					if( qptr->rstatus != QR_REPOSTED ) {
						qptr->resolverPtr = 0;
						qptr->rstatus = QR_IDLE;
						qptr->releaseWaitingThreads();
					} else {
						qptr->rstatus = QR_AWAITING_RESPONSE;
					}
				mx.unlock();
			}
		}

		
		// check for timeouts
		time(&now);
		if( (now-lastTick)>=tickInterval ) {
			//	UPDEBUG("UpResolver::checkTimeouts() checking timeouts...");
			lastTick += tickInterval;
			mx.lock();
				qIt = auxIt = queryMap.begin();
				while( qIt!=queryMap.end() ) {
					++auxIt;
					qptr = qIt->second;
					if( ( qptr->rstatus==QR_AWAITING_RESPONSE ) &&
						 ( (now-qptr->transmitTime) > retrans )    )
					{
						// last query datagram timed out
						if( qptr->retra>0 ) {
							// we can still try retransmition of query datagram
							if( retransmitQuery(*qptr,now) ) {
								// retransmitted ok, do nothing for the moment
								continue;
							}
						// retransmission failed! socket error?
						} else {
							// query timed out, cannot retransmit anymore
							qptr->status = QUERY_TIME_OUT;
							UPDEBUG("UpResolver::checkTimeouts() Query id %u expired, "
										"remaining queries: %u",qptr->id,
										(unsigned int)queryMap.size()-1);
						}
						// select and schedule for completion
						qptr->rstatus = QR_MARKED_FOR_COMPLETION;
						expiredQueries.push_front(qptr);
						// erase query ptr from container
						queryMap.erase(qIt);
					}
					qIt = auxIt;
				}
			mx.unlock();

			if( expiredQueries.begin()==expiredQueries.end() ) continue;

			// query completion is done from outside mx.lock() to allow
			// new queries from queryCompletion() avoiding deadlocks
			for( expIt=expiredQueries.begin() ; expIt!=expiredQueries.end(); ++expIt ) {
				qptr = *expIt;
				// we run UpDnsQuery::completed() outside the lock
				if( qptr->callbackType==QUERY_USE_CALLBACK )
					qptr->completed(); // may reQuery itself
			}

			mx.lock(); // we lock again here to protect qptr->rstatus
				while( expiredQueries.begin() != expiredQueries.end() ) {
					qptr = expiredQueries.front();
					if( qptr->rstatus != QR_REPOSTED )
					     { qptr->resolverPtr = 0; qptr->rstatus = QR_IDLE; }
					else { qptr->rstatus = QR_AWAITING_RESPONSE; qptr->allowWaiting();  }
					qptr->releaseWaitingThreads();
					expiredQueries.pop_front();
				}
			mx.unlock();

		} // end of check Timeouts


	} // end of UpResolver::run() main loop

	return 0;
}



/////////////////////////////////////////////////////////////////////////////


void UpResolver::terminateQuery(UpDnsQuery& q) {
	// if called from UpResolver's own thread (that is from
	// UpDnsQuery::completed() ) we cannot terminate it
	// and the application programmer must be fired
	assert( ! checkSelf() /* cannot be called from UpDnsQuery */ );
	#ifdef DEBUG
	unsigned int sz;
	#endif
	bool shouldWait=(false), shouldComplete(false);
	do {
		mx.lock();
			if( q.rstatus==QR_MARKED_FOR_COMPLETION || q.rstatus==QR_REPOSTED ) {
				shouldWait = true;
			} else {
				// ok, can go on with termination
				if( q.rstatus==QR_AWAITING_RESPONSE ) {
					shouldComplete = true;
				}
				q.rstatus = QR_MARKED_FOR_DELETION;
				q.status = QUERY_TERMINATED;
				shouldWait = false;
				queryMap.erase(q.id);
				#ifdef DEBUG
				sz = queryMap.size();
				#endif
			}
		mx.unlock();
		if( shouldWait ) q.waitCompletion(); // q.blockCond.var is -1 on return
	} while( shouldWait );

	// When it gets here, query is not in the queryMap, it is marked for
	// deletion, and no one can change its rstatus.
	if( shouldComplete && q.callbackType==QUERY_USE_CALLBACK )
		q.completed();

	// no need for protection here because q.state==QR_MARKED_FOR_DELETION
	q.resolverPtr = 0;
	q.releaseWaitingThreads();

	UPDEBUG("UpResolver::terminateQuery() id %u, remaining queries %u" , q.id ,sz);
}


/////////////////////////////////////////////////////////////////////////////

void UpResolver::cancelQuery(UpDnsQuery& q) {
	// if called from UpResolver's own thread (that is from
	// UpDnsQuery::completed() ) we cannot cancel it.
	if( checkSelf() ) return; // do nothing

	bool shouldWait(false), shouldComplete(false), doNothing(false);
	#ifdef DEBUG
	unsigned int sz;
	#endif
	do {
		mx.lock();
			if( q.rstatus==QR_MARKED_FOR_COMPLETION || q.rstatus==QR_REPOSTED ) {
				shouldWait     = true;
			} else
			if( q.rstatus==QR_AWAITING_RESPONSE ) {
				q.rstatus      = QR_MARKED_FOR_COMPLETION;
				q.status       = QUERY_CANCELLED;
				shouldWait     = false;
				shouldComplete = true;
				queryMap.erase(q.id);
				#ifdef DEBUG
				sz = queryMap.size();
				#endif
			} else {
				// q.rstatus is QR_IDLE or QR_MARKED_FOR_DELETION
				doNothing      = true;
			}
		mx.unlock();

		if( doNothing  ) return;
		if( shouldWait ) q.waitCompletion(); // q.blockCond.var is -1 on return

	} while( shouldWait );

	// When it gets here, query is not in the queryMap and is marked for
	//  completion, and no one can change its rstatus.
	// If query tries to query itself from completed(), it will not be allowed,
	//  because UpResolver::query() fails when called from outside its thread
	//  and rstatus is not QR_IDLE
	if( shouldComplete && q.callbackType==QUERY_USE_CALLBACK )
		q.completed();

	mx.lock(); // we lock again here to protect q.rstatus
		q.resolverPtr = 0;
		q.rstatus = QR_IDLE;
		q.releaseWaitingThreads();
	mx.unlock();
	UPDEBUG("UpResolver::cancelQuery() id %u, remaining queries %u" , q.id ,sz);
}

