/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Ignacio Kaufmann for his work in this class

#include <UpTools/UpResolverSimple.h>
#include <UpTools/UpLog.h>

#include <cerrno>
#include <cstring>

UpResolverSimple::UpResolverSimple()
{
	// get data from libresolv (see /usr/include/resolver.h)
	res_init();
	retrans			= (time_t)_res.retrans; // retransmition time interval
	retry				= _res.retry;				// number of times to retransmit
	dnsServerAddr	= &(_res.nsaddr); // we use only first nameserver for now
	// initialize UpResolverSimple
	init();
}

UpResolverSimple::UpResolverSimple( const std::string& _nameserver,time_t _retrans,
		                  unsigned int _retry )
: dnsServerAddr( _nameserver, 53), retrans(_retrans), retry(_retry)
{ init(); }


bool UpResolverSimple::init() {
	// bind to all interfaces, let OS decide what port to assign us
	sock.setDatagram(); // UDP oriented socket
	sock.setNonBlocking(false); // must block for incoming responses
	sock.setRecvTimeout(retrans); // retransmit time interval
	if( !sock.bind("0.0.0.0",0) ) {
		UPLOG(LOG_ERR,"UpResolverSimple::UpResolverSimple() "
					"error binding to all interfaces, any port" );
		return false;
	}
	UPLOG(LOG_DEBUG,"UpResolverSimple::init() nameServer:%s retransmitInterval:%usecs"
			 " retryTimes:%u", dnsServerAddr.c_str_2dots(), (unsigned int)retrans,
			retry );
	return true;
}



UpResolverSimple::~UpResolverSimple() {
	sock.close();
}



bool UpResolverSimple::query(UpDnsQuerySimple& q) {
	if( ! q.prepare() ) {
		UPLOG(LOG_ERR,"UpResolverSimple::query() error preparing query");
		return false;
	}
	
	UpSockAddr remoteSA;
	int n;
	unsigned int rid;
	unsigned int attempts = retry;

	while( attempts > 0 ) {

		q.status = QUERY_SEARCHING;
		n = sock.sendto((char*)(q.query_buff),q.querySize,dnsServerAddr);
		// envio lo que tenia que enviar? si no, hubo un socket error
		if( n != (int)q.querySize ) {
			q.status = QUERY_SOCK_ERROR;
			if( n>=0 ) { UPLOG(LOG_ERR,"UpResolverSimple::query() error in socket while "
									"sending dns query id %u",q.id); }
			else       UPLOG(LOG_ERR,"UpResolverSimple::query() error(%u) in "
									"socket while sending dns query id %u: %s",errno,
									q.id,strerror(errno));
			return false;
		}

		n = sock.recvFrom((char*)(q.answer_buff),NS_PACKETSZ,remoteSA);

		if( n<0 ) {
			if( errno==EAGAIN ) {
				// UPDEBUG("UpResolverSimple::query() socket timer timeout");
				--attempts;
				continue; // retransimt
			} else {
				UPLOG(LOG_ERR,"UpResolverSimple::query() read error in socket(%d): %s",
									errno,strerror(errno));
				return false;
			}
		}
		if( n==0 ) {
			UPLOG(LOG_ERR,"UpResolverSimple::query() error: EOF received in udp socket?");
			return false;
		}
		if( ! (dnsServerAddr==remoteSA ) ) {
			UPLOG(LOG_ALERT,"UpResolverSimple::query() unsolicited datagram "
					"from %s",remoteSA.c_str_2dots());
			continue;
		}
		
		rid=ns_get16(q.answer_buff);

		if( q.id==rid ) {
			q.answerSize = n; 
			q.status     = QUERY_RECEIVED;
			return q.initialResponseParse();
		}

		UPLOG(LOG_ALERT,"UpResolverSimple::query() unsolicited response "
				"id %u from server",rid );
	}

	q.status = QUERY_TIME_OUT;
	return false;
}

