/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Horacio J. Peña for his work in this class

#include <UpTools/UpRoute.h>


UpRoute::UpRoute()
: metric(0),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{}

UpRoute::UpRoute(const UpRoute& route)
: UpNet(route.ip,route.m),metric(route.metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{}

UpRoute::UpRoute(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{}

UpRoute::UpRoute(const UpAddress& baseAddr,unsigned int maskLen,unsigned int _metric)
: UpNet(baseAddr,maskLen),metric(_metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{}

UpRoute::UpRoute(in_addr baseAddr,in_addr mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{}

UpRoute::UpRoute(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric)
: UpNet(baseAddr,mask),metric(_metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{ }

UpRoute::UpRoute(const std::string& route)
: metric(0),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{ set(route); }

UpRoute::UpRoute(const std::string& route,unsigned int _metric)
: metric(_metric),next(0),previous(0),firstSon(0),lastSon(0),auxNet(false)
{ set(route,_metric); }

UpRoute::~UpRoute() {
	if( firstSon ) {
		UpRoute *q,*r=firstSon;
		while(r) { q=r ; r=r->next; delete q; }
	}
}

bool UpRoute::set(const std::string& net,unsigned int _metric) {
	size_t pos=net.find('/');
	if( pos!=std::string::npos ) { // slash found
		ip=net.substr(0,pos++);
		if( net.find_first_of(".:",pos)!=std::string::npos ) {
			m = net.c_str()+pos;
		} else {
			int len = std::atoi(net.c_str()+pos);
			if( len>=0 ) m.setMask((unsigned int)len,ip.addrType);
			else m.clear();
		}
	} else { // slash not found: one-host net
		ip=net;
		m.setMask(128,ip.addrType);
	}
	metric = _metric;
	if( ip.family()!=m.family() or ! (m.family()==AF_INET or m.family()==AF_INET6) )
	{ ip.clear(); m.clear(); return false; }
	return true;
}

bool UpRoute::set(const std::string& net) {
	size_t pos = net.find('/'), spc;
	if( pos!=std::string::npos ) { // slash found
		ip = net.substr(0,pos++);
		spc = net.find(' ',pos);
		if( net.find_first_of(".:",pos)!=std::string::npos ) { // mask as addr
			if( spc!=std::string::npos ) { // metric found?
				m = net.substr(pos,spc-pos);
				metric = (spc<net.size()) ? std::atoi(net.c_str()+spc+1) : 0;
			} else { // no metric found
				m = net.c_str()+pos;
				metric = 0;
			}
		} else { // mask as number
			int len = std::atoi(net.c_str()+pos);
			if( len>=0 ) m.setMask((unsigned int)len,ip.addrType);
			else m.clear();
			metric = (spc!=std::string::npos) ? std::atoi(net.c_str()+spc+1) : 0;
		}
	} else { // slash not found: one-host net
		spc =  net.find(' ');
		if( spc!=std::string::npos ) { // metric found?
			ip = net.substr(0,spc++);
			metric = std::atoi(net.c_str()+spc);
		} else {
			ip = net;
			metric = 0;
		}
		m.setMask(128,ip.addrType);
	}
	if( ip.family()!=m.family() or ! (m.family()==AF_INET or m.family()==AF_INET6) )
	{ ip.clear(); m.clear(); return false; }
	return true;
}


void UpRoute::insertHead(UpRoute* newRoute) {
	if( firstSon ) firstSon->previous = newRoute;
	else lastSon = newRoute;
	newRoute->next = firstSon;
	newRoute->previous = 0;
	firstSon = newRoute;
}

void UpRoute::insertTail(UpRoute *newRoute) {
	if( lastSon ) lastSon->next = newRoute;
	else firstSon = newRoute;
	newRoute->previous = lastSon;
	newRoute->next = 0;
	lastSon = newRoute;
}

void UpRoute::insertBefore(UpRoute* currentNet, UpRoute* newRoute) {
	if( currentNet==firstSon ) firstSon = newRoute;
	else currentNet->previous->next = newRoute;

	newRoute->next = currentNet;
	newRoute->previous = currentNet->previous;

	currentNet->previous = newRoute;
}

void UpRoute::replace(UpRoute* currentNet, UpRoute* newRoute) {
	if( currentNet==firstSon ) firstSon = newRoute;
	else currentNet->previous->next = newRoute;

	if( currentNet==lastSon ) lastSon = newRoute;
	else currentNet->next->previous = newRoute;

	newRoute->next = currentNet->next;
	newRoute->previous = currentNet->previous;

	currentNet->next = currentNet->previous = 0;
}

void UpRoute::unlist(UpRoute* oldRoute) {
	if(oldRoute!=lastSon) oldRoute->next->previous = oldRoute->previous;
	else										  lastSon = oldRoute->previous;

	if(oldRoute!=firstSon) oldRoute->previous->next = oldRoute->next;
	else										  firstSon = oldRoute->next;

	oldRoute->next = oldRoute->previous = 0;
}

void UpRoute::becomeAncestor(UpRoute* r1,UpRoute* r2) {
	m = MIN(r1->m, r2->m);
	while( (r1->ip & m) != (r2->ip & m) ) m += m;
	// it found a common parent
	ip = (r1->ip & m);
}

UpNetRelation UpRoute::addCheck(UpRoute* r) {
	UpNetRelation incl = check(*r);
	if( incl!=UPNR_SUBNET ) return incl;
	if( firstSon==0 ) { insertHead(r); return incl; }
	bool inserted = false;
	UpRoute* q = firstSon;
	for(;;) {
		switch( q->addCheck(r) ) {
			case UPNR_SUPERNET:
				if( !inserted ) {
					replace(q,r);
					inserted = true;
				} else unlist(q);
				r->addCheck(q);
				q = r->next;
				break;
			case UPNR_AFTER:
				q = q->next;
				break;
			case UPNR_EQUAL:
				return UPNR_SUBNET_COLLIDES;
			case UPNR_BEFORE:
				if( !inserted ) insertBefore(q,r);
			case UPNR_SUBNET:
			default: // this default here is just to turn off compile warnings
				return UPNR_SUBNET;
		}
		if( q==0 ) {
			if( !inserted ) insertTail(r);
			break;
		}
	}
	return UPNR_SUBNET;
}


// removes a net and possibly its subnets
bool UpRoute::remove(const UpAddress& _ip,const UpAddress& _m,bool includeSubnets) {
	UpNetRelation incl;
	UpRoute *p=this,*parent=0,*toDel=0;
	do {
		incl = p->check(_ip);
		if( incl==UPNR_SUBNET ) {
			if( p->m >= _m ) { // p is a smaller or equal net than r
				if( p->m == _m ) toDel = p; // equal net
				break;
			} else { // p is a bigger net than r
				parent = p;
				p = p->firstSon;
			}
		} else if( incl==UPNR_AFTER ) {
			p = p->next;
		} else break;
	} while(p);

	if( ! (toDel and parent) ) return false;

	if( toDel->firstSon and !includeSubnets ) {
		if( parent->firstSon==toDel ) parent->firstSon = toDel->firstSon;
		else {                   toDel->previous->next = toDel->firstSon;
		                     toDel->firstSon->previous = toDel->previous;
		}
		if( parent->lastSon ==toDel ) parent->lastSon  = toDel->lastSon;
		else {                   toDel->next->previous = toDel->lastSon;
		                          toDel->lastSon->next = toDel->next;
		}
		toDel->firstSon = toDel->lastSon = 0;
	} else {
		if( parent->firstSon==toDel ) parent->firstSon = toDel->next;
		else                     toDel->previous->next = toDel->next;
		if( parent->lastSon ==toDel ) parent->lastSon  = toDel->previous;
		else                     toDel->next->previous = toDel->previous;
	}
	delete toDel;
	return true;
}

void UpRoute::unoptimize() {
	UpRoute *toDel=firstSon,*aux;
	while( toDel ) {
		if( toDel->auxNet ) {
			if( toDel->firstSon ) {
				aux = toDel->firstSon;
				if( firstSon==toDel ) firstSon = toDel->firstSon;
				else {   toDel->previous->next = toDel->firstSon;
		             toDel->firstSon->previous = toDel->previous;
				}
				if( lastSon ==toDel ) lastSon  = toDel->lastSon;
				else {   toDel->next->previous = toDel->lastSon;
		              	  toDel->lastSon->next = toDel->next;
				}
				toDel->firstSon = toDel->lastSon = 0;
			} else {
				aux = toDel->next;
				if( firstSon==toDel ) firstSon = toDel->next;
				else     toDel->previous->next = toDel->next;
				if( lastSon ==toDel ) lastSon  = toDel->previous;
				else     toDel->next->previous = toDel->previous;
			}
			delete toDel;
			toDel = aux;
		} else {
			toDel->unoptimize();
			toDel = toDel->next;
		}
	}
}



void UpRoute::optimize() {
 if(firstSon!=lastSon) {
	// ultima ip de la red a optimizar
	UpAddress last=getLastIp();
	// red propuesta para insercion
	UpRoute* anc = new UpRoute; anc->auxNet = true;
	// debe insertarse anc?
	bool shouldInsert;
	// ri,rf redes predefinidas que deben ser incluidas por anc
	UpRoute *ri,*rf=lastSon;
	// Las siguientes variables deben deben poder representar numeros que
	// puedan ser mayores a la cantidad de hosts en la internet y con signo
	typedef double addrCountType;
	// cantidad de redes predefinidas que quedan adentro de anc
	//  (solo tiene el valor correcto justo antes de que se inserte anc)
	addrCountType netsToInclude;
	// cantidad de hosts que empeoran en una unidad (iteracion) el tiempo de
	// busqueda en caso de agregar anc
	addrCountType hostsWorse;
	// cantidad de hosts que mejoran en (netsToInclude-1) unidades por incluir
	// la red anc
	addrCountType hostsBetter;
	// eficiencia de insercion de anc
	addrCountType figure,priorFigure;
	while( rf && (ri=rf->previous) ) {
	
		figure = priorFigure = 0;
		netsToInclude = 2;
		shouldInsert = false;
		for(;;) {
			anc->becomeAncestor(ri,rf);
			hostsBetter = (addrCountType)( last - anc->getLastIp() );
			////////////////////////////////////////////////
			// las siguientes lineas solo mejoran el tiempo de optimizacion
			// pero no alteran el arbol optimizado final
			if( hostsBetter==0.0 ) { if( !shouldInsert ) break; }
			else while( ri->previous && (anc->check(ri->previous->ip)==UPNR_SUBNET) ) {
					ri = ri->previous;
					++netsToInclude;
			} //////////////////////////////////////////////
			hostsWorse = (addrCountType) anc->getSize();
			figure = hostsBetter * (netsToInclude-1.0) - hostsWorse;
			if( figure>priorFigure ) shouldInsert=true;
			else if( (figure<priorFigure)&&shouldInsert ) {
				anc->becomeAncestor(ri->next,rf);
				addCheck(anc); rf=lastSon; anc=new UpRoute; anc->auxNet = true;
				break;
			}
			if( ri==firstSon ) {
				if(shouldInsert) { addCheck(anc); rf=lastSon; anc=new UpRoute; anc->auxNet = true; }
				break;
			}
			ri = ri->previous;
			++netsToInclude;
			priorFigure = MAX(figure,priorFigure);
		}
		rf = rf->previous;
	}
	delete anc;
 }
 for(UpRoute* r=firstSon; r!=0; r=r->next) r->optimize();
}

void UpRoute::clear() {
	ip.clear(); m.clear(); auxNet = false; metric = 0;
	if( firstSon ) {
		UpRoute *q,*r=firstSon;
		while(r) { q=r ; r=r->next; delete q; }
		firstSon = lastSon = 0;
	}
	next = previous = 0;
}

void UpRoute::tree(std::ostream& o, int indent) const
{
  for(int i=0;i<indent;++i) o.write("  ",2);
  append(o);
  if( auxNet ) o << "(aux)\n"; else o << '\n';;
  if(firstSon==0) return;
  ++indent;
  for(UpRoute* r=firstSon; r; r=r->next)
	r->tree(o,indent);
}


const UpRoute* UpRoute::lookup(in_addr ipv4,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute *p=this,*ret=0;
	do {
		incl = p->check(ipv4);
		if( incl==UPNR_SUBNET ) {
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->firstSon;
		} else if( incl==UPNR_AFTER ) {
			p = p->next;
		} else break;
	} while(p);
	return ret;
}


const UpRoute* UpRoute::lookup(const in6_addr& ipv6,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute *p=this,*ret=0;
	do {
		incl = p->check(ipv6);
		if( incl==UPNR_SUBNET ) {
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->firstSon;
		} else if( incl==UPNR_AFTER ) {
			p = p->next;
		} else break;
	} while(p);
	return ret;
}


const UpRoute* UpRoute::lookup(const UpNet& n,bool useMetric) const {
	UpNetRelation incl;
	const UpRoute *p=this,*ret=0;
	do {
		incl = p->check(n.ip);
		if( incl==UPNR_SUBNET ) {
			if( p->m > n.m ) break; // n is a bigger
			if( !p->auxNet and (!useMetric or !ret or p->metric<=ret->metric) )
				ret = p;
			p = p->firstSon;
		} else if( incl==UPNR_AFTER ) {
			p = p->next;
		} else break;
	} while(p);
	return ret;
}


std::list<const UpRoute*> UpRoute::lookupAll(const UpAddress& ipAddr) const {
	UpNetRelation incl;
	const UpRoute *p=this;
	std::list<const UpRoute*> ret;
	do {
		incl = p->check(ipAddr);
		if( incl==UPNR_SUBNET ) {
			if( ! p->auxNet ) ret.push_back(p);
			p = p->firstSon;
		} else if( incl==UPNR_AFTER ) {
			p = p->next;
		} else break;
	} while(p);
	return ret;
}
