/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpRegex.h>
#include <UpTools/UpLog.h>
#include <UpTools/UpArch.h>
#include <cstdlib>
#include <cstring>

#define UPSQL_LIBRARY_COMPILE_TIME 1
#include <UpTools/UpSql.h>

void strScr(std::string& s, const std::string& m);
void strScr(std::string& s);
void strScrCpy(std::string& dst, const std::string& src);
void strScrScrCpy(std::string& dst, const std::string& src);
void strToLower(std::string& s);

const std::string UpSqlConn::dbTypeStr[] = { std::string("undefined"),
			std::string("mysql"), std::string("postgresql"),std::string("max") };

QueryElem UpSqlQuery::emptyQueryElem(0);

UpSqlConnUndef upSqlConnUndef;

//////////////////////////////////////////////////////////////////////////////
// UpSqlConn methods

UpSqlConn::UpSqlConn(upDbType t)
: type((t<DBTYPE_MAX)?t:DBTYPE_UNDEF), need_close(false), statusOk(false),
  statusPhrase(NOT_CONNECTED), port(0), retryTime(0)
{ currentHostIt = hosts.end(); }

UpSqlConn::UpSqlConn(const UpSqlConn& c)
: type(c.type), need_close(false), statusOk(false), statusPhrase(NOT_CONNECTED),
  user(c.user), dbname(c.dbname), port(c.port), hosts(c.hosts),
  retryTime(c.retryTime)
{ currentHostIt = hosts.begin(); strScrScrCpy(pass,c.pass); }



bool UpSqlConn::copyParams(const UpSqlConn& c) {
  if( c.type!=type ) return false;
  disconnect();
  user		 = c.user;
  dbname		 = c.dbname;
  port		 = c.port;
  hosts		 = c.hosts;
  retryTime	 = c.retryTime;
  connParams = c.connParams; 
  currentHostIt = hosts.begin();
  strScrScrCpy(pass,c.pass);
  return true;
}



void UpSqlConn::addHost(const std::string& host,bool pushFront) {
	if( host.empty() ) return;
	for(std::list<std::string>::iterator it=hosts.begin(); it!=hosts.end(); ++it)
		if( host==*it ) return; // will not add twice the same host
	if( pushFront ) hosts.push_front(host); else hosts.push_back(host);
	if( pushFront or currentHostIt==hosts.end() ) currentHostIt = hosts.begin();
}


bool UpSqlConn::set( const std::string& sqlUri ) {
	UpRegex dbu(
		"([a-z0-9]+)://([^:]+)(:[^@]+)?@([[a-z0-9.-]+)(:[0-9]+)?/([^;]+)(;.+)?",
		8,true);
	if( !dbu.exec(sqlUri) )
		{ statusPhrase = "error parsing sqlUri"; return false; }

	std::string dbt(dbu[1]);	 strToLower(dbt);
	std::string pass(dbu[3]);   pass.erase(0,1);
	std::string port(dbu[5]);   port.erase(0,1);
	std::string params(dbu[7]); params.erase(0,1);

	unsigned int t;
	for( t=1; t<DBTYPE_MAX; ++t )
		if( dbTypeStr[t]==dbt ) break;
	if( (upDbType)t!=type )
		{ statusPhrase = "wrong dbtype in sql URI"; return false; }

	// convert params string to a space separated list of atrib=value elements
	for( char* p=(char*)params.c_str(); *p!=0 ; ++p )
		if( *p==';' ) *p=' ';
	
	return set(dbu[4],dbu[2],pass,dbu[6],std::atoi(port.c_str()),params);
}


bool UpSqlConn::set( const std::string& _host, const std::string& _user,
							const std::string& _pass, const std::string& _dbname,
							unsigned short _port, const std::string& _params)
{
   if( _user.empty() or _host.empty() )
		{ statusPhrase = "error setting user or host parameters"; return false; }
	addHost(_host,true); // this host will be connected first
	strScrCpy(pass,_pass);
	user			= _user;
	dbname		= _dbname;
	port			= _port;
	connParams	= _params;
   return true;
}


bool UpSqlConn::connect( const std::string& sqlUri ) {
	if( ! set(sqlUri) )
	{ UPLOG(LOG_ERR,"UpSqlConn::connect(...) %s",statusPhrase); return false; }
	return connect();
}

bool UpSqlConn::connect( const std::string& _host, const std::string& _user,
					 const std::string& _pass, const std::string& _dbname,
					 unsigned short     _port, const std::string& _params)
{
	if( ! set(_host,_user,_pass,_dbname,_port, _params) )
	{ UPLOG(LOG_ERR,"UpSqlConn::connect(...) %s",statusPhrase); return false; }
	return connect();
}



bool UpSqlConn::connect() {
	statusOk = false;
	if( hosts.empty() or user.empty() ) {
		statusPhrase = "no db connection parameters set";
		UPLOG(LOG_ERR,"UpSqlConn::connect() cannot connect before setting "
						  "conn parameters");
		return false;
	}
	std::list<std::string>::iterator actualHostIt(currentHostIt);
	if( need_close and retryTime>0 ) sleep( retryTime );
	do {
		disconnect(); // disconnect if necessary, before reconnecting
		if( privateConnect(*currentHostIt) ) return true;
		++currentHostIt; // try next host
		if( currentHostIt==hosts.end() )
			currentHostIt = hosts.begin();
	} while( currentHostIt!=actualHostIt );
	if( hosts.size()>1) {
		statusPhrase = "cannot connect to any server";
		UPLOG(LOG_ERR,"UpSqlConn::connect() cannot connect to any server");
	}
	return false;
}



void UpSqlConn::scrPass()
{ strScr(pass); }

//////////////////////////////////////////////////////////////////////////////
// UpSqlQuery methods

UpSqlQuery::UpSqlQuery()
: elem(&emptyQueryElem), fCount(0), errNumber(0)
{ errPhrase[0]= errPhrase[255]='\0'; }

UpSqlQuery::~UpSqlQuery()
{ clearResultList(); }

void UpSqlQuery::clearResultList() {
	if( elem!=&emptyQueryElem ) {
		elem->deleteNext();
		delete elem;
		elem = &emptyQueryElem;
	}
}

void UpSqlQuery::clear() {
	clearResultList();
	sql.clear();
	fCount       = 0;
	errNumber    = 0;
	errPhrase[0] = '\0';
} 

void UpSqlQuery::addQueryElem( QueryElem* e )
{
	if( elem == &emptyQueryElem ) {
		elem         = e;
		fCount       = e->fCount;
		errNumber    = 0;
		errPhrase[0] = '\0';
	} else {
		elem->pushBack(e);
	}
} 


bool UpSqlQuery::nextResult()
{
	if( elem==&emptyQueryElem ) return false;
	QueryElem *elemToDel = elem;
	elem = elem->next;
	delete elemToDel;
	if( elem==0 ) {
		elem=&emptyQueryElem;
		fCount = 0;
		return false;
	}
	fCount = elem->fCount;
	return true;
}



void UpSqlQuery::setError(
						unsigned int _errorNumber,
						const char* _errorPhrase
								)
{
	clearResultList();
	fCount    = 0;
	errNumber = _errorNumber;
	std::strncpy(errPhrase,_errorPhrase,255);// errPhrase[255] is already zero
}

//////////////////////////////////////////////////////////////////////////////
// useful functions

void strScr(std::string& s, const std::string& m) {
	union { const std::string* p; int i[INTS_IN_PTR]; } w;
	w.p = &m;
	int x = w.i[0];
	#if __WORDSIZE==64
		for( unsigned int h=1; h<INTS_IN_PTR; ++h ) x ^= w.i[h];
	#endif
	int*        i((int*)s.c_str());
	const int*  e(i+(s.size()>>2));
	char*       j((char*)e);
	const char* k(s.c_str()+s.size());
	char        z(x);
	while( i<e ) { *i++ ^= x; }
	while( j<k ) { *j++ ^= z; }
}


void strScr(std::string& s) {
	strScr( s , s );
}

void strScrCpy(std::string& dst, const std::string& src) {
	dst=src;
	strScr(dst);
}

void strScrScrCpy(std::string& dst, const std::string& src) {
	dst=src;
	strScr(dst,src);
	strScr(dst);
}


void strToLower(std::string& s) {
	char* p((char*)s.c_str());
	while(*p) {
		*p=std::tolower(*p); ++p;
	}
}


