/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpSsl.h>
#include <UpTools/UpOp.h>

#include <cassert>
#include <fstream>

#include <openssl/pem.h>
#include <openssl/evp.h>



//////////////////////////////////////////////////////////////////////////////
// useful functions

void setStringFromBIO(BIO* bk,std::string& s) {
	char* p; long len(BIO_get_mem_data(bk,&p));
	s.assign(p,len);
}

void setStringFromX509_NAME(X509_NAME* n,std::string& s) {
	BIO* bk(BIO_new(BIO_s_mem()));
	if( !bk ) { s.clear(); return; }
	// flag parameter options: XN_FLAG_ONELINE, XN_FLAG_MULTILINE,
	//						   XN_FLAG_RFC2253, or XN_FLAG_COMPAT
	// indent parameter: indent spaces
	X509_NAME_print_ex(bk,n,0,XN_FLAG_ONELINE);
	setStringFromBIO(bk,s);
	BIO_free(bk);
//	alternative version
/*	char buf[1024];
	X509_NAME_oneline(xn,buf,1024);
	s.assign(buf); */
}

	



//////////////////////////////////////////////////////////////////////////////

UpRsa::UpRsa()
: x(0),privKey(false)
{}

UpRsa::~UpRsa() {
	freeX();
}

void UpRsa::freeX() {
	if( x!=0 ) { RSA_free(x); x=0; }
}


void UpRsa::statusPadError() {
	char buf[128]; buf[127]=0;
	ERR_error_string_n(ERR_get_error(),buf,127);
	status<<buf;
}


bool UpRsa::readKey(BIO* bk, const std::string& passphrase)
{
	freeX();
	status.clear();
	void* u = 0;
	if( passphrase.size() ) {
		u = (void*)passphrase.c_str();
		// see "man pem", "man evp" and "man OpenSSL_add_all_algorithms"
		OpenSSL_add_all_ciphers();
	}
	x = PEM_read_bio_RSAPrivateKey(bk,0,0,u);
	if( x!=0 ) { privKey=true; return true; }
	status = "UpRsa::readKey(bio,passphrase). PEM_read_bio_RSAPrivateKey(...) ";
	statusPadError();
	if( 1==BIO_reset(bk) ) {
		x = PEM_read_bio_RSA_PUBKEY(bk,0,0,u);
	//	x = PEM_read_bio_RSAPublicKey(bk,0,0,u); // muy similar
		if( x!=0 ) {
			privKey=false; status.clear();
			return true;
		}
	}
	status << ". PEM_read_bio_RSA_PUBKEY(...) ";
	statusPadError();
	EVP_cleanup();
	// TODO: esto podria llegar a perjudicar otras instancias de estas clases... Revisar!
	return false;
}

bool UpRsa::saveKey( BIO* bk,
		bool savePublicOnly, bool encryptKey, const std::string& passphrase )
{
	if( x==0 ) {
		status = "UpRsa::saveKey(bk,savePublicOnly,encryptKey,passphrase) "
				 "error: no key present";
		return false;
	}
	const EVP_CIPHER* enc = encryptKey ? UP_KEY_ENC : 0;
	void* u = passphrase.size() ? (void*)passphrase.c_str() : 0;
	int wret;
	// see "man EVP_CIPHER_type" ans "man pem"
	if( privKey && !savePublicOnly )
		wret = PEM_write_bio_RSAPrivateKey(bk,x,enc,0,0,0,u);
	else
		wret = PEM_write_bio_RSA_PUBKEY(bk,x);
//		wret = PEM_write_bio_RSAPublicKey(bk,x); // muy similar

	if( wret!=0 ) return true;
	status="UpRsa::saveKey(bk,savePublicOnly,passphrase) ";
	statusPadError();
	return false;
}

/////////////////////////////////////////////////////////////////////////////
// the follwing read and save methods use the previous BIO save and read methods

bool UpRsa::readKeyFile(const std::string& filename,
						 const std::string& passphrase)
{
	// see "man BIO" and "man BIO_s_mem" and "man BIO_s_file"
	BIO* bk = BIO_new_file(filename.c_str(),"r");
	if( bk==0 ) {
		status.clear();
		status<<"UpRsa::readKeyFile(filename,passphrase) error opening file "
			  <<filename;
		return false;
	}
	bool retVal = readKey(bk,passphrase);
	BIO_free(bk);
	return retVal;
}

bool UpRsa::readKey(const char* keyBuf,size_t keyBufLen,
						 const std::string& passphrase)
{
	// see "man BIO" and "man BIO_s_mem" and "man BIO_s_file"
	BIO* bk = BIO_new_mem_buf((char*)keyBuf,keyBufLen);
	if( bk==0 ) {
		status = "UpRsa::readKey(keyBuf,keyBufLen,passphrase) error creating bio";
		return false; 
	}
	bool retVal = readKey(bk,passphrase);
	BIO_free(bk);
	return retVal;
}

bool UpRsa::readKey(const std::string& pemKey, const std::string& passphrase)
{ return readKey(pemKey.c_str(),pemKey.size(),passphrase); }



bool UpRsa::saveKey( std::string& pemKey,
		bool savePublicOnly, bool encryptKey, const std::string& passphrase )
{
	BIO* bk = BIO_new(BIO_s_mem());
	if( bk==0 ) {
		status = "UpRsa::saveKey(pemKey,savePublicOnly,encryptKey,passphrase) "
				"error creating bio";
		return false; 
	}
	bool retVal = saveKey(bk,savePublicOnly,encryptKey,passphrase);
	if( retVal ) setStringFromBIO(bk,pemKey); else pemKey.clear();
	BIO_free(bk);
	return retVal;
}


bool UpRsa::saveKeyFile(const std::string& filename,
		bool savePublicOnly, bool encryptKey, const std::string& passphrase )
{
	BIO* bk = BIO_new_file(filename.c_str(),"w");
	if( bk==0 ) {
		status.clear();
		status<<"UpRsa::saveKeyFile(filename,savePublicOnly,encryptKey,"
				"passphrase) error opening file "<<filename;
		return false; 
	}
	bool retVal = saveKey(bk,savePublicOnly,encryptKey,passphrase);
	BIO_free(bk);
	return retVal;
}


//////////////////////////////////////////////////////////////////////////////

std::string UpRsa::privEncrypt(const std::string& cleartext)
{
	std::string cipher;
	privEncrypt(cleartext,cipher);
	return cipher;
}

std::string UpRsa::pubEncrypt(const std::string& cleartext)
{
	std::string cipher;
	pubEncrypt(cleartext,cipher);
	return cipher;
}

int UpRsa::privEncrypt( const std::string& cleartext, std::string& cipher )
{
	if( x==0 || !privKey ) {
		status = "UpRsa::privEncrypt(cleartext,cipher) error: "
					"no private key present";
		return -1;
	}
	int preSize(RSA_size(x)),postSize;
	cipher.resize(preSize,0);
	postSize = RSA_private_encrypt(	cleartext.size(),
												(unsigned char*)cleartext.c_str(),
												(unsigned char*)cipher.c_str(),
												x , RSA_PKCS1_PADDING );
	assert( postSize <= preSize /* memory overrun ? */ );
	if( postSize<0 ) {
		status = "UpRsa::privEncrypt(cleartext,cipher) ";
		statusPadError();
		cipher.resize(0);
		return -1;
	}
	status.clear();
	return postSize;
}

int UpRsa::pubEncrypt( const std::string& cleartext, std::string& cipher )
{
	if( x==0 ) {
		status = "UpRsa::pubEncrypt(cleartext,cipher) error: "
					"no key present";
		return -1;
	}
	int preSize(RSA_size(x)),postSize;
	cipher.resize(preSize,0);
	postSize = RSA_public_encrypt(	cleartext.size(),
												(unsigned char*)cleartext.c_str(),
												(unsigned char*)cipher.c_str(),
												x , RSA_PKCS1_PADDING );
	assert( postSize <= preSize /* memory overrun ? */ );
	if( postSize<0 ) {
		status = "UpRsa::pubEncrypt(cleartext,cipher) ";
		statusPadError();
		cipher.resize(0);
		return -1;
	}
	status.clear();
	return postSize;
}

//////////////////////////////////////////////////////////////////////////////

std::string UpRsa::privDecrypt(const std::string& cipher )
{
	std::string cleartext;
	privDecrypt(cipher,cleartext);
	return cleartext;
}

std::string UpRsa::pubDecrypt(const std::string& cipher )
{
	std::string cleartext;
	pubDecrypt(cipher,cleartext);
	return cleartext;
}


int UpRsa::privDecrypt(	const std::string& cipher, std::string& cleartext )
{
	if( x==0 || !privKey ) {
		status="UpRsa::privDecrypt(cipher,cleartext) error: "
					"no private key present";
		return -1;
	}
	int preSize(RSA_size(x)),postSize;
	cleartext.resize(preSize,0);
	postSize = RSA_private_decrypt(	cipher.size(),
												(unsigned char*)cipher.c_str(),
												(unsigned char*)cleartext.c_str(),
												x , RSA_PKCS1_PADDING );
	assert( postSize <= preSize /* memory overrun ? */ );
	if( postSize<0 ) {
		status="UpRsa::privDecrypt(cipher,cleartext) ";
		statusPadError();
		cleartext.resize(0);
		return -1;
	}
	status.clear();
	return postSize;
}


int UpRsa::pubDecrypt(	const std::string& cipher, std::string& cleartext )
{
	if( x==0 ) {
		status="UpRsa::pubDecrypt(cipher,cleartext) error: "
					"no key present";
		return -1;
	}
	int preSize(RSA_size(x)),postSize;
	cleartext.resize(preSize,0);
	postSize = RSA_public_decrypt(	cipher.size(),
												(unsigned char*)cipher.c_str(),
												(unsigned char*)cleartext.c_str(),
												x , RSA_PKCS1_PADDING );
	assert( postSize <= preSize /* memory overrun ? */ );
	if( postSize<0 ) {
		status="UpRsa::pubDecrypt(cipher,cleartext) ";
		statusPadError();
		cleartext.resize(0);
		return -1;
	}
	status.clear();
	return postSize;
}

//////////////////////////////////////////////////////////////////////////////
bool UpRsa::generateKey(int bits,unsigned long exponent)
{
	freeX();
	status.clear();
	std::ifstream f(RND_SOURCE);
	if( !f.good() ) {
		status<<"couldn't open random source: "<<RND_SOURCE;
		return false;
	}
	int seed;
	f.read((char*)&seed,sizeof(seed));
	int n = f.gcount();
	f.close();
	if( n!=sizeof(seed) ) {
		status<<"UpRsa::generateKey(bits,exponent) "
					"error reading random source "<<RND_SOURCE;
		return false;
	}
	// man "RSA_generate_key":
	// RSA_generate_key() generates a key pair and returns it in a newly
	// allocated RSA structure. The pseudo-random number generator must
	// be seeded prior to calling RSA_generate_key().
	// Comment: es posible que la libreria openssl sea tan trucha?!? como es
	// posible que necesite usar rand()?!?
	// Quiere decir que el espacio de las claves que genera es tan solo del
	// tama#o del seed del pseudo-random generator: 2^32  !!!!!!?!
	srand(seed);
	x = RSA_generate_key(bits,exponent,0,0);
	// segun las pruebas realizadas luego de la compilacion, aun fijando el
	// seed siemre en el mismo valor, la libreria genera claves diferentes.
	// Por lo que quizas la afirmacion del parrafo anterior no era cierta.
	if( x==0 ) {
		status = "UpRsa::generateKey(bits,exponent) ";
		statusPadError();
		return false;
	}
	privKey = true;
	return true;

}


//////////////////////////////////////////////////////////////////////////////
// UpX509 methods

UpX509::UpX509()
: x(0)
{}


UpX509::~UpX509() {
	freeX();
}

void UpX509::freeX() {
	if( x!=0 ) { X509_free(x); x=0; }
}


void UpX509::statusPadError() {
	char buf[128]; buf[127]=0;
	ERR_error_string_n(ERR_get_error(),buf,127);
	status<<buf;
}

bool UpX509::readCert(BIO* bk) {
	freeX();
	status.clear();
	// si fuera DER en lugar de PEM podriamos hacer x=d2i_X509_bio(bp,0);
	x = PEM_read_bio_X509(bk,0,0,0);
	if( x!=0 ) return true;
	status = "UpX509::readCert(bio). PEM_read_bio_X509(...) ";
	statusPadError();
	return false;
}

bool UpX509::saveCert(BIO* bk) {
	if( x==0 ) {
		status = "UpX509::saveCert(bk) error: no cert present";
		return false;
	}
	int wret = PEM_write_bio_X509(bk,x);
	if( wret!=0 ) return true;
	status="UpX509::saveCert(bio) ";
	statusPadError();
	return false;
}

std::string UpX509::getSubject() {
	std::string ret;
	setStringFromX509_NAME(X509_get_subject_name(x),ret);
	return ret;
}

std::string UpX509::getIssuer() {
	std::string ret;
	setStringFromX509_NAME(X509_get_issuer_name(x),ret);
	return ret;
}

bool UpX509::checkTimeValidity() {
	time_t now(time(0));
	return ( X509_cmp_time(X509_get_notBefore(x),&now) < 0 ) and
		   ( X509_cmp_time(X509_get_notAfter(x),&now)  > 0 )     ;
	// alternative version (this one is worse because it calls time(0) twice)
/*	return (X509_cmp_current_time(X509_get_notAfter())<0) and
		   (X509_cmp_current_time(X509_get_notBefore()>0)     ; */
}

bool UpX509::showCert(BIO* bk) {
	return 1==X509_print(bk, x);	
	// alternative versions:
	// first flag parameter options: XN_FLAG_ONELINE, XN_FLAG_MULTILINE,
	//								 XN_FLAG_RFC2253, or XN_FLAG_COMPAT
	// second flag parameter: must go X509_FLAG_COMPAT in order to show
	//						   all fields.
/*	return 1==X509_print_ex(bk,x,XN_FLAG_COMPAT,X509_FLAG_COMPAT); */
/*	return 1==X509_print_ex(bk,x,XN_FLAG_MULTILINE,X509_FLAG_COMPAT); */
}

//////////////////////////////////////////////////////////////////////////////
// the follwing read and save methods use the previous BIO save and read methods

bool UpX509::readCertFile(const std::string& pemCertFilename) {
	// see "man BIO" and "man BIO_s_mem" and "man BIO_s_file"
	BIO* bk = BIO_new_file(pemCertFilename.c_str(),"r");
	if( bk==0 ) {
		status.clear();
		status<<"UpX509::readCertFile(pemCertFilename) error opening file "
			  <<pemCertFilename;
		return false; 
	}
	bool retVal = readCert(bk);
	BIO_free(bk);
	return retVal;
}


bool UpX509::readCert(const char* pemCertBuf,size_t pemCertLen) {
	// see "man BIO" and "man BIO_s_mem" and "man BIO_s_file"
	BIO* bk = BIO_new_mem_buf((char*)pemCertBuf,pemCertLen);
	if( bk==0 ) {
		status = "UpX509::readCert(pemBuf,pemBufLen) error creating bio";
		return false; 
	}
	bool retVal = readCert(bk);
	BIO_free(bk);
	return retVal;
}

bool UpX509::readCert(const std::string& pemCert)
{ return readCert(pemCert.c_str(),pemCert.size()); }



bool UpX509::saveCert(std::string& pemCert) {
	BIO* bk = BIO_new(BIO_s_mem());
	if( bk==0 ) {
		status = "UpX509::saveCert(pemCert) error creating bio";
		return false; 
	}
	bool retVal = saveCert(bk);
	if( retVal!=0 ) setStringFromBIO(bk,pemCert); else pemCert.clear();
	BIO_free(bk);
	return retVal;
}


bool UpX509::showCert(std::string& textCert) {
	BIO* bk = BIO_new(BIO_s_mem());
	if( bk==0 ) {
		status = "UpX509::showCert(textCert) error creating bio";
		return false; 
	}
	bool retVal = showCert(bk);
	if( retVal!=0 ) setStringFromBIO(bk,textCert); else textCert.clear();
	BIO_free(bk);
	return retVal;
}


bool UpX509::saveCertFile(const std::string& pemCertFilename) {
	BIO* bk = BIO_new_file(pemCertFilename.c_str(),"w");
	if( bk==0 ) {
		status.clear();
		status<<"UpX509::saveCertFile(pemCertFilename) error opening file "
			  <<pemCertFilename;
		return false; 
	}
	bool retVal = saveCert(bk);
	BIO_free(bk);
	return retVal;
}


bool UpX509::showCertFile(const std::string& textCertFilename) {
	BIO* bk = BIO_new_file(textCertFilename.c_str(),"w");
	if( bk==0 ) {
		status.clear();
		status<<"UpX509::showCertFile(textCertFilename) error opening file "
			  <<textCertFilename;
		return false; 
	}
	bool retVal = saveCert(bk);
	BIO_free(bk);
	return retVal;
}


