/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpSsl.h>
#include <UpTools/UpLog.h>
#include <UpTools/UpArch.h>
#include <cassert>
#include <openssl/crypto.h>
#include <unistd.h>

const std::string ssl_empty_string;
void* upSslLibInstance=0;

//////////////////////////////////////////////////////////////////////////////
// class UpCryptoSslLib methods

//constructor
UpCryptoSslLib::UpCryptoSslLib() {
	// only one instance
	assert( upSslLibInstance==0 );
	upSslLibInstance = this;
	// readable error messages
	ERR_load_crypto_strings();
	SSL_load_error_strings();
	// registers the available ciphers and digests
	SSL_library_init();
	// TODO: actions to seed PRNG
	// ...
}

// destructor
UpCryptoSslLib::~UpCryptoSslLib() {
	ERR_free_strings();
	upSslLibInstance = 0;
}


//////////////////////////////////////////////////////////////////////////////
// UpSslContext methods


UpSslContext::UpSslContext(upCtxMethodtype mt)
: ctx(0)
{
	memset(errBuf,0,128);
	// manejar SSLv2, SSLv3 y TLSv1
	switch(mt) {
		case UPCTXMT_SERVERONLY:
			ctx = SSL_CTX_new(SSLv23_server_method()); break;
		case UPCTXMT_CLIENTONLY:
			ctx = SSL_CTX_new(SSLv23_client_method()); break;
		default:
			ctx = SSL_CTX_new(SSLv23_method());
	}
	if(!ctx) { checkErr(-1); return; } // severe error!
		
 	// desactivar SSLv2 por inseguro, y activar los workaround de bugs en
	//   otras implemtaciones
	SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2|SSL_OP_ALL);
	// habilitar la reutilizacion de sesiones server (si lo solicita un
	//   cliente), y las sesiones cliente (si lo solicita el usuario)
	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_BOTH);
	// habilitar la renegociacion en caso de problemas para evitar
	// que deba ser tratada manualmente por el usuario
	int ctxMode = SSL_CTX_get_mode(ctx)|SSL_MODE_AUTO_RETRY;
	SSL_CTX_set_mode(ctx,ctxMode);
	// timeout de sesion (default:300)
	// SSL_CTX_set_timeout(ctx,500);
	// set ciphers (default: ?)
	// SSL_CTX_set_cipher_list(ctx, "???");
}


UpSslContext::~UpSslContext() {
	if(ctx) SSL_CTX_free(ctx);
}


bool UpSslContext::useCertAndRsaPrivateKey(
								UpX509& upx,UpRsa& upr )
{
	if(!ctx) { /* error is already set in this case*/ return false; }
	if( !upx.x or !upr.x) {
		setErr("useCertAndRsaPrivateKey(params): error in params");
		return false;
	}
	if( !checkErr(SSL_CTX_use_certificate(ctx,upx.x))   ) return false;
	if( !checkErr(SSL_CTX_use_RSAPrivateKey(ctx,upr.x)) ) return false;
	return checkErr(SSL_CTX_check_private_key(ctx));
}

bool UpSslContext::addCAChainCert(UpX509& upx) {
	if(!ctx) { /* error is already set in this case*/ return false; }
	if( !upx.x ) {
		setErr("addCAChainCert(params): error in params");
		return false;
	}
	return checkErr(SSL_CTX_add_extra_chain_cert(ctx,upx.x));
}
	
bool UpSslContext::addCAChainFile(const char* CAChainFile) {
	if(!ctx) { /* error is already set in this case*/ return false; }
	if( !CAChainFile ) {
		setErr("addCAChainFile(params): error in params");
		return false;
	}
	return checkErr(SSL_CTX_use_certificate_chain_file(ctx,CAChainFile));
}


bool UpSslContext::addCAChainFilePath(
		const char* CAChainFile, const char* hashedCANamesDir )
{
	if(!ctx) { /* error is already set in this case*/ return false; }
	if( !hashedCANamesDir and !CAChainFile ) {
		setErr("addCAChainFilePath(params): error in params");
		return false;
	}
	return checkErr(SSL_CTX_load_verify_locations(ctx,CAChainFile,hashedCANamesDir));
}
//////////////////////////////////////////////////////////////////////////////
// UpSsl methods

UpSsl::UpSsl(UpSslContext& upctx)
: fd(0), upsock(0), ssl(SSL_new(upctx.ctx)), bidirectionalShutdown(false), sslRet(1)
{
	memset(errBuf,0,128);
	if( ssl==0 ) ctxErr();
}

UpSsl::~UpSsl() {
	if(ssl) SSL_free(ssl);
}

bool UpSsl::setChannel(UpSocket& us) {
	if( !checkErr(SSL_set_fd(ssl,us.fd())) ) return false;
	fd = us.fd(); upsock = &us;
	return true;
}

bool UpSsl::setChannel(int _fd) {
	if( !checkErr(SSL_set_fd(ssl,_fd)) ) return false;
	fd = _fd; upsock = 0;
	return true;
}

bool UpSsl::useCertAndRsaPrivateKey( UpX509& upx,UpRsa& upr )
{
	if( !upx.x or !upr.x ) {
		sslErr=-1; strcpy(errBuf,"error in parameter data");
		return false;
	}
	if( !checkErr(SSL_use_certificate(ssl,upx.x))   ) return false;
	if( !checkErr(SSL_use_RSAPrivateKey(ssl,upr.x)) ) return false;
	return checkErr(SSL_check_private_key(ssl));
}



int UpSsl::reInit() {
	close(); // final handshake and socket close
	SSL_CTX* ctx = SSL_get_SSL_CTX(ssl);
	SSL_free(ssl);
	ssl = SSL_new(ctx);
	if( ! ssl ) { ctxErr(); return false; }
	return true;
}

int UpSsl::shutdown() {
	int sslRet=SSL_shutdown(ssl);
/* 1   The shutdown was successfully completed. The "close notify" alert was
       sent and the peer's "close notify" alert was received.
   0   The shutdown is not yet finished. Call SSL_shutdown() for a second time,
       if a bidirectional shutdown shall be performed.  The output of
       SSL_get_error(3) may be misleading, as an erroneous SSL_ERROR_SYSCALL
       may be flagged even though no error occurred.
  -1   The shutdown was not successful because a fatal error occurred either
       at the protocol level or a connection failure occurred. It can also
       occur if action is need to continue the operation for non-blocking BIOs.
       Call SSL_get_error(3) with the return value ret to find out the reason.*/
	if( sslRet==0 and bidirectionalShutdown ) {
		// cerro un solo lado
		sslRet = SSL_shutdown(ssl);
	}
	return checkErr(sslRet);
}

void UpSsl::close() {
	shutdown();
	if( fd ) {
		if( upsock ) upsock->close(); else ::close(fd);
		fd = 0;
		upsock = 0;
	}
}

int UpSsl::connect() {
	if( !checkErr(SSL_connect(ssl)) ) return false;
#ifdef DEBUG
	if( SSL_session_reused(ssl)!=0 ) {
		UPLOG(LOG_DEBUG,"UpSsl::connect() ssl session reused");
	}
#endif
	return true;
}

int UpSsl::accept() {
	if( !checkErr(SSL_accept(ssl)) ) return false;
#ifdef DEBUG
	if( SSL_session_reused(ssl)!=0 ) {
		UPLOG(LOG_DEBUG,"UpSsl::accept() ssl session reused");
	}
#endif
	return true;
}

int UpSsl::write(const char* buf, size_t len) {
	int retVal(SSL_write(ssl,(void*)buf,len));
	checkErr(retVal);
	return retVal;
}

int UpSsl::read(char* buf, size_t len) {
	int retVal(SSL_read(ssl,buf,len));
	checkErr(retVal);
	return retVal;
}



int UpSsl::forceRenegotiation() {
	// http://h71000.www7.hp.com/doc/83final/ba554_90007/ch04s03.html
   int retVal(SSL_renegotiate(ssl));
   if( !checkErr(retVal) ) return retVal;
   retVal = SSL_do_handshake(ssl);
   if( !checkErr(retVal) ) return retVal;
	if( ssl->server ) {
		// when on server side, we need to call twice to handshake
		// (first time the client accepts to renegotiates, and second time
		// it actually does)
	   ssl->state = SSL_ST_ACCEPT;
   	retVal = SSL_do_handshake(ssl);
   	checkErr(retVal);
	}
   return retVal;
}

