/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpSsl.h>
#include <UpTools/UpThread.h>
#include <UpTools/UpLog.h>
#include <UpTools/UpArch.h>
#include <cassert>
#include <openssl/crypto.h>

void* upSslThreadSafePtr = 0;

//////////////////////////////////////////////////////////////////////////////
// Class for multithreaded ssl usage.

// Autmatically instantiated once as a global object:
//		USER MUST NOT INSTANTIATE THIS CLASS
// (this is why it is not even included in the .h file)

// All UpSslThreadSafe functionality could be included in class
// UpCryptoSslLibMultithreading, but it would expose is internals
// in the header file.

class UpSslThreadSafe {
private:
	UpMutex* mx; // table of mutexes
	int mutexCount;
public:
	UpSslThreadSafe();
	~UpSslThreadSafe();
	UpMutex& mutex(int i) { assert( i<mutexCount ); return mx[i]; }
};

//////////////////////////////////////////////////////////////////////////////
// class UpCryptoSslLib methods


//constructor
UpCryptoSslLibMultithreading::UpCryptoSslLibMultithreading() {
	assert( upSslThreadSafePtr == 0 );
	upSslThreadSafePtr = new UpSslThreadSafe;
}

// destructor
UpCryptoSslLibMultithreading::~UpCryptoSslLibMultithreading() {
	delete (UpSslThreadSafe*)upSslThreadSafePtr;
	upSslThreadSafePtr = 0;
}


//////////////////////////////////////////////////////////////////////////////
// la libreria necesita callbacks (es una libreria en C despues de todo)
// estos callbacks seran usados internamente por la libreria
// (ver man CRYPTO_lock)

unsigned long upcrypto_thread_id_callback() {
	return (unsigned long)pthread_self();
}

void upcrypto_mutexop_callback(int mode,int n, const char *file, int line)
{
	#ifdef UPSSL_CALLBACK_DEBUG
	UPLOG(LOG_DEBUG,"upcrypto_mutexop_callback(%d,%d,%s,%d)",mode,n,file,line);
	#endif
	if( mode & CRYPTO_LOCK )
		((UpSslThreadSafe*)upSslThreadSafePtr)->mutex(n).lock();
	else
		((UpSslThreadSafe*)upSslThreadSafePtr)->mutex(n).unlock();
}

struct CRYPTO_dynlock_value {
	UpMutex mx;
};

struct CRYPTO_dynlock_value*
	upcrypto_dyn_create_mutex_callback(const char *file, int line )
{
	struct CRYPTO_dynlock_value* l = new CRYPTO_dynlock_value;
	// this debug line compiles to nothing unless DEBUG is defined
	#ifdef UPSSL_CALLBACK_DEBUG
	UPLOG(LOG_DEBUG,"upcrypto_dyn_create_mutex_callback(%s,%d) returns " PFX,
			file,line,PTR2PRINTFCAST(l) );
	#endif
	return l;
}

void upcrypto_dyn_mutexop_callback(	int mode, struct CRYPTO_dynlock_value *l,
									const char *file, int line				 )
{
	#ifdef UPSSL_CALLBACK_DEBUG
	UPLOG(LOG_DEBUG,"upcrypto_dyn_mutexop_callback(%d," PFX ",%s,%d)",mode,
				PTR2PRINTFCAST(l),file,line );
	#endif
	if( mode & CRYPTO_LOCK )	l->mx.lock();
	else						l->mx.unlock();
	// TODO: revisar para que se usaran en mode los valores CRYPTO_READ y
	//		CRYPTO_WRITE (por ejemplo cuando viene mode=CRYPTO_READ|CRYPTO_LOCK)
	//      donde esta CRYPTO_add?

}

void upcrypto_dyn_destroy_mutex_callback(	struct CRYPTO_dynlock_value *l,
											const char *file, int line		)
{
	#ifdef UPSSL_CALLBACK_DEBUG
	UPLOG(LOG_DEBUG,"upcrypto_dyn_destroy_mutex_callback(" PFX ",%s,%d)",
			PTR2PRINTFCAST(l),file,line);
	#endif
	delete l;
}

//////////////////////////////////////////////////////////////////////////////
// class UpSslThreadSafe methods

//constructor
UpSslThreadSafe::UpSslThreadSafe() {
	// initialize locks
	mutexCount = CRYPTO_num_locks();
	mx = new UpMutex[mutexCount];
	// mutex callbacks
	CRYPTO_set_locking_callback(upcrypto_mutexop_callback);
	CRYPTO_set_id_callback(upcrypto_thread_id_callback);
	// dynamic mutex callbacks (will be used in future versions of openssl)
	CRYPTO_set_dynlock_create_callback(upcrypto_dyn_create_mutex_callback);
	CRYPTO_set_dynlock_lock_callback(upcrypto_dyn_mutexop_callback);
	CRYPTO_set_dynlock_destroy_callback(upcrypto_dyn_destroy_mutex_callback);
}

// destructor
UpSslThreadSafe::~UpSslThreadSafe() {
	delete[] mx;
}


