/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpSubstitute.h>
#include <UpTools/UpOp.h>
#include <iostream>

char UpSubstitute::substituteIndicator('\\');

/////////////////////////////////////////////////////////////////////////////
// private methods

bool UpSubstitute::regexScanner( const std::string& _regexExpression )
{
	if( _regexExpression.size()==0 ) return false;
	scanState  state(SC_REGEX), auxState(SC_REGEX);
	nmatch=1;

	for( const char *p=_regexExpression.c_str(); *p ; ++p ) {
		switch(state) {
			case SC_REGEX:
				if( *p == '(' )	{
					state=SC_ATOM; ++nmatch; break;
				}
				if( *p == '[' )	{
					auxState=state; state=SC_BRACKET_FIRST; break;
				}
				if( *p == '\\' ) {
					auxState=state; state=SC_ESCAPED; break;
				}
				break;
			case SC_ATOM:
				switch(*p) {
					case ')':  state=SC_REGEX; break;
					case '[':  auxState=state; state=SC_BRACKET_FIRST; break;
					case '\\': auxState=state; state=SC_ESCAPED; break;
				}
				break;
			case SC_BRACKET_FIRST:
				state=SC_BRACKET;
				break;
			case SC_BRACKET:
				if( *p==']' ) state=auxState;
				break;
			case SC_ESCAPED:
				state=auxState;
				break;
			default: break;
		}
	}
	return (state==SC_REGEX);
}


bool UpSubstitute::combinedExpressionScanner(
		const std::string& _regexAndSubstituteExpression )
{
	if( _regexAndSubstituteExpression.size()==0 ) return false;
	scanState   state(SC_REGEX), auxState(SC_REGEX);
	const char *p(_regexAndSubstituteExpression.c_str());
	char        separator(*p++);
	const char *exp(p);
	nmatch=1;

	for( ; *p ; ++p ) {
		switch(state) {
			case SC_REGEX:
				if( *p == separator ) {
					regexString.assign(exp,(p-exp));
					state = SC_SUBS; // a parsear la substitutionExpression
					exp = p+1;
					break;
				}
				if( *p == '(' )	{
					state=SC_ATOM; ++nmatch; break;
				}
				if( *p == '[' )	{
					auxState=state; state=SC_BRACKET_FIRST; break;
				}
				if( *p == '\\' ) {
					auxState=state; state=SC_ESCAPED; break;
				}
				break;
			case SC_SUBS:
				if( *p == separator ) {
					substituteExpression.assign(exp,(p-exp));
					state = SC_FLAGS; // a parsear la substitutionExpression
				}
				break;
			case SC_ATOM:
				switch(*p) {
					case ')':  state=SC_REGEX; break;
					case '[':  auxState=state; state=SC_BRACKET_FIRST; break;
					case '\\': auxState=state; state=SC_ESCAPED; break;
				}
				break;
			case SC_BRACKET_FIRST:
				state=SC_BRACKET;
				break;
			case SC_BRACKET:
				if( *p==']' ) state=auxState;
				break;
			case SC_ESCAPED:
				state=auxState;
				break;
			case SC_FLAGS:
				if( *p=='i' ) { icase=true; break; }
				// unknown flag
				return false;
				break;
		}
	}
	return (state==SC_FLAGS);
}

/////////////////////////////////////////////////////////////////////////////
// general methods


bool UpSubstitute::exec(const std::string& _text)
{
	if( ! UpRegex::exec(_text) ) return false;
	result.clear();
	const char* p = substituteExpression.c_str();
	const char* q = p;
	int      cont = 0;

	while(*p) {
		if( *p != substituteIndicator ) {
            ++p; ++cont;
        } else {
      		result.append(q,cont);
            ++p;
				unsigned int i = (*p - '0');
				if( i>9 ) return false;
            result.append(getMatch(i));
            ++p;
            q = p; cont = 0;
        }
    }
	if( cont>0 ) {
   		result.append(q,cont);
	}
	return true;
}

////////////////////////////////////////////////////////////////////////
// Constructor por copia

UpSubstitute::UpSubstitute(const UpSubstitute& s)
: UpRegex(s.regexString,s.nmatch,(s.cflags&REG_ICASE)!=0),
  substituteExpression(s.substituteExpression)
{ } 

////////////////////////////////////////////////////////////////////////
// A) compilacion a partir de datos individuales ingresados por
//    el usuario

UpSubstitute::UpSubstitute(const std::string& _regex,
						   const std::string& _substituteExpression,
						   size_t _atomCount, bool _icase )
: UpRegex(_regex,_atomCount,_icase), substituteExpression(_substituteExpression)
{ }

bool UpSubstitute::compile(const std::string& _regex,
						   const std::string& _substituteExpression,
						   size_t _atomCount, bool _icase )
{
	substituteExpression = _substituteExpression;
	return UpRegex::compile(_regex,_atomCount,_icase);
}


//////////////////////////////////////////////////////////////////////////////////////
// B) calcula solo los parentesis
// ( si la expresion regular esta disponible en el momento de la
//   programacion, entonces es mejor usar los metodos A porque son mas
// rapidos)
UpSubstitute::UpSubstitute(const std::string& _regex,
                     	   const std::string& _substituteExpression,
                       	   bool _icase)
{
    substituteExpression = _substituteExpression;
	if( regexScanner(_regex) ) { // calcula nmatch
		UpRegex::compile(_regex,nmatch,_icase);
	} else {
		regexerror=1;
		errorString.assign("error in regex",14);
	}
}

bool UpSubstitute::compile(const std::string& _regex,
                     	   const std::string& _substituteExpression,
                     	   bool _icase) {
    substituteExpression = _substituteExpression;
    if( regexScanner(_regex) ) { // calcula nmatch
		return UpRegex::compile(_regex,nmatch,_icase);
	}
	regexerror=1;
	errorString.assign("error in regex",14);
	return false;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////
// C) compilacion a partir de una unica expresion como ser:
//        /regex/substituteExpression/i
//     o  !regex!substituteExpression!

UpSubstitute::UpSubstitute(const std::string& _regexAndSubstituteExpression) 
 : icase(false), combinedExpression(_regexAndSubstituteExpression)
{
	// calcula nmatch, obtiene regexString y substituteExpression
	if( combinedExpressionScanner(_regexAndSubstituteExpression) ) {
		UpRegex::compile(regexString,nmatch,icase);
	} else {
		regexerror=1;
		errorString.assign("error in combined expression",28);
	}
}



bool UpSubstitute::compile(const std::string& _regexAndSubstituteExpression) {
	icase=false;
	// calcula nmatch, obtiene regexString y substituteExpression
    if( combinedExpressionScanner(_regexAndSubstituteExpression) ) {
		combinedExpression = _regexAndSubstituteExpression;
		return UpRegex::compile(regexString,nmatch,icase);
    }
	regexerror=1;
	errorString.assign("error in combined expression",28);
	return false;
}
