/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <UpTools/UpThread.h>

// must compile with -lpthread


///////////////////////////////////////////////////////////////////////////////
// important global function (needed to run thread!)///////////////////////////
///////////////////////////////////////////////////////////////////////////////
void* auxUpThreadFunction(void* arg) { 
	pthread_exit( ((UpThread*)arg)->run() );
}


///////////////////////////////////////////////////////////////////////////////
// UpThread methods ///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// constructor && destructor
UpThread::UpThread() : thread(0), started(false) {
	pthread_attr_init(&attr);
}

UpThread::~UpThread() {
	pthread_attr_destroy(&attr);
	if(started) {
		void* dummy;
		pthread_cancel(thread);
		pthread_join(thread,&dummy);
	}
}



//////////////////////////////////////////////////////////////////////////////
// thread creation
bool UpThread::create() {
	if( started ) return false;
	if( pthread_create( &thread , &attr , auxUpThreadFunction , (void*)this ) )
		return false;
	started = true;
	return true;
}

bool UpThread::create( void* (*thFunc)(void*) , void* arg) {
	if( started ) return false;
	if( pthread_create( &thread , &attr , thFunc , arg ) )
		return false;
	started = true;
	return true;
}

// should be replaced by a derived object method
void* UpThread::run() {
	return (void*)"empty thread";
}

///////////////////////////////////////////////////////////////////////////////
// methods callable only from methods in the derived classes 

// checks if the thread that executes this method is the same thread that was
// started by this object
bool UpThread::checkSelf() {
	// pthread_self() devuelve el threadId del llamador, que se compara con el
	// threadId del propio objeto
	return started && ( pthread_equal(thread,pthread_self()) != 0 ) ;
}

// exits a thread (if detached, finished immediately and calls cleanup handlers,
//  if not detached, it waits a join on another thread)
void UpThread::exit(void* retval) { pthread_exit(retval); }

// test for pending cancellation requests
void UpThread::testCancel() { pthread_testcancel(); }

// enable honoring of cancellation requests
bool UpThread::setCancelEnable()
{ return ! pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL); }

// ignore cancellation requests
bool UpThread::setCancelDisable()
{ return ! pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL); }

// set asynchronous (immediate) honoring of cancellation requests
bool UpThread::setCancelTypeAsync()
{return !pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);}

// set deferred (when reaching a cancellation point) honoring of cancellation
//   requests
// cancellation points are: pthread_join,  pthread_cond_wait,
//   pthread_cond_timedwait,  pthread_testcancel,  sem_wait,  sigwait.
bool UpThread::setCancelTypeDeferred()
{ return ! pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL); }


//////////////////////////////////////////////////////////////////////////////
// Atttributes testing (before creating the thread) 

bool UpThread::testAttrDetached() {
	int state;
	pthread_attr_getdetachstate(&attr,&state);
	return (state==PTHREAD_CREATE_DETACHED);
}

bool UpThread::testAttrJoinable() {
	int state;
	pthread_attr_getdetachstate(&attr,&state);
	return (state==PTHREAD_CREATE_JOINABLE);
}

bool UpThread::testAttrSchedOther() {
	int policy;
	pthread_attr_getschedpolicy(&attr,&policy);
	return ( policy == SCHED_OTHER);
}

bool UpThread::testAttrSchedRoundRobin() {
    int policy;
    pthread_attr_getschedpolicy(&attr,&policy);
    return ( policy == SCHED_RR );
}

bool UpThread::testAttrSchedFifo() {
    int policy;
    pthread_attr_getschedpolicy(&attr,&policy);
    return ( policy == SCHED_FIFO );
}

bool UpThread::testAttrExplicitSched() {
    int inherit;
	pthread_attr_getinheritsched(&attr,&inherit);
    return (inherit==PTHREAD_EXPLICIT_SCHED);
}

bool UpThread::testAttrInheritSched() {
    int inherit;
	pthread_attr_getinheritsched(&attr,&inherit);
    return (inherit==PTHREAD_INHERIT_SCHED);
}

bool UpThread::testAttrScopeSystem() {
    int scope;
	pthread_attr_getinheritsched(&attr,&scope);
    return (scope==PTHREAD_SCOPE_SYSTEM);
}

bool UpThread::testAttrScopeProcess() {
    int scope;
	pthread_attr_getinheritsched(&attr,&scope);
    return (scope==PTHREAD_SCOPE_PROCESS);
}

int UpThread::getAttrSchedPriority() {
	sched_param sparam;
	pthread_attr_getschedparam(&attr,&sparam);
	return sparam.sched_priority;
}

//////////////////////////////////////////////////////////////////////////////
// Attributes setting (before creating the thread)

bool UpThread::setAttrDetached() {
	if(started) return false;
	return ! pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
}

bool UpThread::setAttrJoinable() {
	if(started) return false;
	return ! pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
}

bool UpThread::setAttrSchedOther() {
	if(started) return false;
	return ! pthread_attr_setschedpolicy(&attr, SCHED_OTHER);
}

bool UpThread::setAttrSchedRoundRobin() {
	if(started) return false;
    return ! pthread_attr_setschedpolicy(&attr, SCHED_RR);
}

bool UpThread::setAttrSchedFifo() {
	if(started) return false;
    return ! pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
}

bool UpThread::setAttrExplicitSched() {
	if(started) return false;
	return ! pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
}

bool UpThread::setAttrInheritSched() {
	if(started) return false;
	return ! pthread_attr_setinheritsched(&attr, PTHREAD_INHERIT_SCHED);
}

bool UpThread::setAttrScopeSystem() {
	if(started) return false;
	return ! pthread_attr_setinheritsched(&attr, PTHREAD_SCOPE_SYSTEM);
}

bool UpThread::setAttrScopeProcess() {
	if(started) return false;
	return ! pthread_attr_setinheritsched(&attr, PTHREAD_SCOPE_PROCESS);
}

bool UpThread::setAttrSchedPriority(int priority) {
	if(started) return false;
	sched_param sparam;
	sparam.sched_priority = priority;
	return ! pthread_attr_setschedparam(&attr,&sparam);
}

///////////////////////////////////////////////////////////////////////////////
// Operation on running threads

bool UpThread::detach() {
	return ! pthread_detach(thread);
}

void* UpThread::join() {
	if(started) {
		void* retval;
		if( pthread_join(thread,&retval)==0 ) {
			started=false;
			return retval ;
		}
	}
	return 0;
}

bool UpThread::cancel() {
	if( started==false ) return false;
	if( pthread_cancel(thread)==0 ) {
		// We really do not know if the thread catched or ignored the cancel
		// also even if it catched the cancel, we still need to join in order
		// the thread memory, so we leave 'started' as it was.
		return true;
	}
	return false;
}

bool UpThread::kill(int signum) {
	return started ? (!pthread_kill(thread,signum)) : false;
}

bool UpThread::isRunning() {
	if(started) {
		bool err = pthread_kill(thread,SIGCONT);
		if(err==0) return true; // thread is running and ignored the signal
	}
	// thread is not running
	return false;
}

pthread_t UpThread::getThreadId() {
	return thread;
}

///////////////////////////////////////////////////////////////////////////////
// Sheduling policy and priority testing on running test
bool UpThread::testSchedOther() {
	sched_param sparam;
	int policy;
	if( pthread_getschedparam(thread,&policy,&sparam) ) return false;
	return (policy==SCHED_OTHER);
}

bool UpThread::testSchedRoundRobin(int& priority) {
	sched_param sparam;
	int policy;
	if( pthread_getschedparam(thread,&policy,&sparam) ) return false;
	priority = sparam.sched_priority;
	return (policy==SCHED_RR);
}

bool UpThread::testSchedFifo(int& priority) {
	sched_param sparam;
	int policy;
	if( pthread_getschedparam(thread,&policy,&sparam) ) return false;
	priority = sparam.sched_priority;
	return (policy==SCHED_FIFO);
}

///////////////////////////////////////////////////////////////////////////////
// Sheduling policy and priority setting on running threads
bool UpThread::setSchedOther() {
	sched_param sparam;
	sparam.sched_priority=0;
	return ! pthread_setschedparam(thread,SCHED_OTHER,&sparam);
}

bool UpThread::setSchedRoundRobin(int priority) {
	sched_param sparam;
	sparam.sched_priority=priority;
	return ! pthread_setschedparam(thread,SCHED_RR,&sparam);
}

bool UpThread::setSchedFifo(int priority) {
	sched_param sparam;
	sparam.sched_priority=priority;
	return ! pthread_setschedparam(thread,SCHED_FIFO,&sparam);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// The following code is to overwrite at execution startup, some of the
// UpLog code in order to log the thread id whenever UPLOG(...) is called.

#include <UpTools/UpLog.h>


#define upThid ((unsigned int)pthread_self())
#define upPid  ((unsigned int)getpid())

extern "C" {

int upLogHdrThreads(char* buf) {
	char *p=buf+24;
	const char *q=upLogIdent;
	time_t t; time(&t); ctime_r(&t,buf); /* 24 bytes */
	*p++=' '; while( *q ) *p++=*q++; 
	if( upLogPid ) {
		return upLogThid ? (p-buf+sprintf(p,"[%u:%u] ",upPid,upThid))
							  : (p-buf+sprintf(p,"[%u] ",upPid));
	} else {
		return upLogThid ? (p-buf+sprintf(p,"[:%u] ",upThid))
							  : (p-buf);
	}
}

} // end of extern "C"
