/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



#include <UpTools/UpThreadWorkDistributor.h>
#include <UpTools/UpLog.h>
#include <UpTools/UpArch.h>
#include <cerrno>
#include <cstring>

#define MIN(A,B) ( ( (B)<(A) ) ? (B) : (A) )

time_t UpThreadWorkDistributor::maxFinishWait = 3; // seconds
//////////////////////////////////////////////////////////////////////////////
// UpWork methods

UpWork::~UpWork()  { }

void UpWork::action() {
	UPLOG(LOG_DEBUG,
			"UpWork::action() is being executed. this=" PFX " thread=%u now=%f"
			" Replace this method in derived classes.",
			PTR2PRINTFCAST(this),(unsigned int)pthread_self(),startTime.getDouble());
}

//////////////////////////////////////////////////////////////////////////////
// UpWorkSplittable methods

std::string UpWorkSplittable::stateStr[] = {
	std::string("NOTFINISHED"), std::string("OWNFINISHED"),
	std::string("JOINEDCHILDS"), std::string("DONE"), std::string("DONEBASE")
};

UpWorkSplittable::~UpWorkSplittable() {
	// std::cout<<"UpWorkSplittable "<<this<<" destructor"<<std::endl; // DEBUG
	if( mx ) delete mx;
}
UpWorkSplittable* UpWorkSplittable::split() { return 0; }
void UpWorkSplittable::join(UpWorkSplittable* workToJoin) { }
char UpWorkSplittable::joinType() { return 'n'; }
bool UpWorkSplittable::exceedHwThreads() { return true; }




// this method is to be executed only by base works
// devuelve true si ya se puede borrar el baseWork
bool UpWorkSplittable::baseEndWork(UpWorkSplittable* w) {
	UpScopedLock scopedLock( *mx ); // Will unlock on return
//	std::cout<<"**********************************************************************"<<std::endl;
//	std::cout<<"baseEndWork esta siendo llamado por: "<<w<<*w<<std::endl;
	w->state = WS_OWNFINISHED; // w termino su trabajo propio
	// hay que hacer ahora join de algun work?
	UpWorkSplittable *currentWork( w ), *nextWork( w->nextWork ), *auxWork;
	char jt( joinType() );
	for(;;) {
//		std::cout<<"---------------------------------------------------------"<<std::endl;
//		std::cout<<"currentWork "<<currentWork<<*currentWork<<' '<<stateStr[currentWork->state]<<std::endl;
//		if(nextWork) {
//			std::cout<<"nextWork "<<nextWork<<*nextWork<<' '<<stateStr[nextWork->state]<<"    "
//						<<"(nextWork->prevWork==currentwork): "
//						<<( nextWork->prevWork == currentWork ) <<std::endl;
//		} else std::cout<<"nextWork 0"<<std::endl;

		// el siguiente if() sirve tanto para tree join como para serial join
		if( nextWork==0 or nextWork->prevWork!=currentWork ) {
			// nothing else to join for currentWork
			if( currentWork->prevWork==0 ) {
				// no other work needs to join currentWork
				currentWork->state = WS_DONE;
				break;
			} else {
				// some other work needs to join this work
				currentWork->state = WS_JOINEDCHILDS;
			}
			currentWork = currentWork->prevWork;
			if( currentWork->state==WS_NOTFINISHED ) {
				// unfinished works cannot join other works
				break;
			}
			nextWork = currentWork->nextWork;
//			std::cout<<" repointing towards prevWork"<<std::endl;
			continue;
		} else if( nextWork->state<WS_JOINEDCHILDS ) {
			// cannot join nextWork yet
			break;
		}
		if( jt!='n' ) {
//			std::cout<<*currentWork<<" is about to join "<<*nextWork<<std::endl;
			currentWork->join( nextWork );
		}
		nextWork->state = WS_DONE;
		auxWork = nextWork ;
		currentWork->nextWork = nextWork = nextWork->nextWork;
//		std::cout<<"about to delete "<<auxWork<<*auxWork<<std::endl;
		delete auxWork;
	}
	// listo los joins posibles hasta aqui.
	// Veamos ahora si currentWork es el trabajo base y si esta terminado.
	if( this!=currentWork or state<WS_DONE ) {
		// baseWork no esta terminado todavia
		return false; // no se puede borrar el baseWork desde este thread
	}
	// el trabajo base termino de joinear todo!!!
	state = WS_DONEBASE;
	// En caso que se haya solicitado un waitCompletion, y que el metodo
	// actual (baseEndWork(w)) este siendo llamado desde un thread diferente
	// al thread originador del trabajo base, entonces hay que avisarle al
	// thread originador del trabajo que no espere mas.
	if( sp and w!=this ) {
		sp->post();
		return false; // no se puede borrar el baseWork desde este thread
	}
	return true; // se puede joinear todo
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// UpWorkThread definition and methods

// Many UpWorkThread's are concurrently processing works
class UpWorkThread : public UpThread {
private:
	UpThreadWorkDistributor& m;
	UpWork* currentWork;
public:
	UpWorkThread(UpThreadWorkDistributor& _m);
	~UpWorkThread();
	void*	run();
};

UpWorkThread::UpWorkThread(UpThreadWorkDistributor& _m) : m(_m) { create(); }

UpWorkThread::~UpWorkThread() { } // UpThreads cancels and joins

void* UpWorkThread::run() {
	for(;;) { // endless loop
		currentWork = m.getConditionedWork();
		if( currentWork==0 ) break; // signalled by manager to finish thread
		if( currentWork->type=='w' ) { // work comun
			currentWork->init();
			currentWork->action(); // this is the method to be derived by the application
			currentWork->end();
			m.accountWorkEnd(currentWork);
			delete currentWork;
		} else { // work splitteable
			UpWorkSplittable* currentSplittable( (UpWorkSplittable*)currentWork );
			UpWorkSplittable* baseWork( currentSplittable->baseWork );
			if( currentSplittable==baseWork ) baseWork->init();
			currentSplittable->action(); // this is the method to be derived by the application
			if( baseWork->baseEndWork( currentSplittable ) ) {
				baseWork->end();
				m.accountWorkEnd( baseWork );
				delete baseWork;
			}
		}
	}
	m.endOfThread();
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
// UpThreadWorkDistributor methods

UpThreadWorkDistributor::UpThreadWorkDistributor( size_t _threadCount )
: threadCount(_threadCount),
  hwThreadCount( MIN( _threadCount , (size_t)sysconf(_SC_NPROCESSORS_ONLN) ) ),
  condition(0), // condition.var = number of pending works
  finishCondition(threadCount),
  baseWorkQueueSize(0), // number of pending base works
  maxQueueWait(MAXQUEUEWAIT_DEFAULT),
  waitCondAux(maxQueueWait*(float)threadCount),
  maxQueueLength(10*_threadCount),
  worksProcessing(0),
  lastArrival(0),
  workArrivalRate(0.0),
  workProcessRate(0.0),
  gamma(DEFAULT_GAMMA),
  oneMinusGamma(1.0-DEFAULT_GAMMA)
{
	for(size_t n=0; n<threadCount; ++n) {
		threadList.push_back(new UpWorkThread(*this) );
	}
	// VER http://stackoverflow.com/questions/150355/programmatically-find-the-number-of-cores-on-a-machine 
	// _SC_NPROCESSORS_CONF The number of processors configured.
	// _SC_NPROCESSORS_ONLN The number of processors currently online (available).
	// en windows se usaria lo siguiente en lugar de  sysconf(_SC_NPROCESSORS_ONLN)
	//     SYSTEM_INFO sysinfo;  GetSystemInfo( &sysinfo );
	//     rdwareThreads = sysinfo.dwNumberOfProcessors;
}


UpThreadWorkDistributor::~UpThreadWorkDistributor() {
	// first we signal threads to finish
	condition.lock(); // safe zone for c.var
		for( size_t i=0; i<threadCount ; ++i ) {
			++condition.var;
			workQueue.push_back(0); // UpWorkThreds will understand this and end
			condition.signal();
		}
	condition.unlock();
	// now we wait till everybody finishes
	UpTimeVal tlast,tnow,remaining;
	tlast.now(); tlast += maxFinishWait;
	finishCondition.lock();
		while( finishCondition.var>0 ) {
			tnow.now();
			remaining = tlast-tnow;
			if( remaining<(time_t)0 ) break;
			finishCondition.timedwait(tlast-tnow);
		}
		bool ok(finishCondition.var==0);
	finishCondition.unlock();
	if( ! ok ) UPLOG(LOG_ERR,"UpThreadWorkDistributor::~UpThreadWorkDistributor() "
					"not all threads finished before timeout");
	// now we delete them (this joins them!)
	std::list<UpWorkThread*>::iterator auxIt, it(threadList.begin());
	while( it!=threadList.end() ) {
		auxIt = it++;
		if( !ok and (*auxIt)->isRunning() ) {
			// if ok==true, then even if a thread is still running, it is going
			// to finish as soon it gets a cpu slice, so we only check
			// the running condition if finishCondition.var>0
			UPLOG(LOG_ERR,"UpThreadWorkDistributor::~UpThreadWorkDistributor() "
					"unterminated thread: %u",
					(unsigned int)(*auxIt)->getThreadId());
		}
		delete *auxIt; // this cancels and joins thread
	}
	condition.lock();
		while( ! workQueue.empty() ) {
			delete workQueue.back();
			workQueue.pop_back();
		}
	condition.unlock();
}


void UpThreadWorkDistributor::endOfThread() {
	finishCondition.lock();
		--(finishCondition.var);
	finishCondition.unlock();
}


UpWork* UpThreadWorkDistributor::getConditionedWork() {
	condition.lock(); // safe zone for condition.var
		// we check using while() instead of if() to prevent spurious wakeups
		while( condition.var==0 ) {
			// no works to process
			condition.wait(); // unlock c and wait for signal...
									// relock c on return
		}
		UpWork* e( workQueue.front() );
		workQueue.pop_front();
		--condition.var;
		if( e ) { // e may be zero!
			if( e->type=='w' or ((UpWorkSplittable*)e)->baseWork==e )
				--baseWorkQueueSize;
			++worksProcessing;
		}
	condition.unlock();
	return e;
}

void UpThreadWorkDistributor::accountWorkEnd(UpWork* e) {
	condition.lock();
		--worksProcessing;
		accountWorkDurationUnsafe(e->getDuration().getDouble());
	condition.unlock();
}

void UpThreadWorkDistributor::accountArrivalUnsafe() {
	UpTimeVal t; t.now();
	UpTimeVal dt = t;
	dt -= lastArrival;
	lastArrival = t;

	workArrivalRate *= gamma;
	workArrivalRate += oneMinusGamma / (float)dt.getDouble();
}

float UpThreadWorkDistributor::getWorkArrivalRateUpdated() {
	UpTimeVal dt;
	dt.now();
	condition.lock();
		dt -= lastArrival;
		float ret = workArrivalRate;
	condition.unlock();
	float aux = 1.0/(float)dt.getDouble();
	if( aux>ret ) return ret;
	ret *= gamma;
	ret += oneMinusGamma * aux;
	return ret;
}


void UpThreadWorkDistributor::accountWorkDurationUnsafe(float dt) {
	workProcessRate *= gamma;
	workProcessRate += oneMinusGamma / dt;
}


// after a number workMemoryCount of works, all of them separated with,
// an interval dt, the estimated interval will be (0.9*dt).
void UpThreadWorkDistributor::setWorkMemoryLength(unsigned int n) {
	condition.lock();
		gamma=expf(logf(0.1)/(float)n);
		oneMinusGamma=1.0-gamma;
	condition.unlock();
}

bool UpThreadWorkDistributor::checkQueueWaitUnsafe() {
	// Si hay threads disponibles se acepta siempre:
	if( worksProcessing<threadCount )
		return true;
	// Si llega hasta aqui es que todos los threads estan ocupados.

	// Si maxQueueWait es negativo o workProcessRate~0, no se chequea
	// la espera sino la cantidad de works encolados:
	if( maxQueueWait<0.0 or workProcessRate<1e-8 )
		return ( baseWorkQueueSize<maxQueueLength );
	// Si maxQueueWait es positivo se chequea el tiempo que tendria que esperar
	// el work para comenzar a ser procesado
	// condition.var     = workQueue.size() = cantidad de works encolados
	// baseWorkQueueSize = cant de base works encolados
  	// waitCondAux       = maxQueueWait*threadCount
  	// workProcessRate   = 1/workAvgDuration
	// Entonces el parentesis puede escribirse:
	// ( baseWorkQueueSize*workAvgDuration/threadCount <= maxQueueWait )
	// osea:
	// ( ExpectedWaitBeforeThisWorkStarts <= maxQueueWait )
	return ( baseWorkQueueSize <= (size_t)(waitCondAux*workProcessRate) );
}

// returns true if work is executed or queued, returns false if work
// queue is too long to accept a new work
bool UpThreadWorkDistributor::processWork( UpWork* w, bool waitCompletion )
{
	if( w->type=='w' ) {
		// UpWork
		condition.lock(); // safe zone for c.var
			accountArrivalUnsafe();
			bool retVal( checkQueueWaitUnsafe() );
			if( retVal ) {
				++condition.var;
				++baseWorkQueueSize;
				workQueue.push_back(w);
				condition.signal();
			}
		condition.unlock();
		return retVal;
	}
	// UpWorkSplittable
	UpWorkSplittable* e((UpWorkSplittable*)w);
	bool retVal(true);
	bool pushed(false);
	size_t availableThreads( e->exceedHwThreads() ? threadCount : hwThreadCount);
	char joinType( e->joinType() );
	UpWorkSplittable *currentWork( e ), *newWork( 0 );
	UpSemaphore* sp(0);
	e->baseWork = e->nextWork = e;
	e->mx = new UpMutex;
	// si no se encola el base work este thread debera esperar a las partes
	if( waitCompletion ) {
		sp = e->sp = new UpSemaphore(0);
		if( ! e->exceedHwThreads() ) --availableThreads;
	}
	// si se encola, entonces ya voy descontando un thread.
	else --availableThreads;

	// safe zone for condition.var
	condition.lock();
		accountArrivalUnsafe(); // se hace una sola vez (solo para el base Work)
		if( checkQueueWaitUnsafe() ) {
			while( availableThreads and ( newWork = currentWork->split() ) ) {
				--availableThreads;
				// set newWork
				newWork->baseWork = e;
				newWork->nextWork = currentWork->nextWork;
				newWork->prevWork = currentWork; // if( joinType!='n' ) newWork->prevWork=currentWork;
				// set old currentWork
				currentWork->nextWork = newWork;
				// set new currentWork
				currentWork = newWork->nextWork;
				if( joinType=='s' or currentWork==e ) currentWork->prevWork = newWork;
				// agendo el newWork para que sea procesado
				++condition.var;
				workQueue.push_back( newWork ); if( !pushed ) pushed = true;
				condition.signal();
			}
			if( ! waitCompletion ) {
				// agendo el base work para que sea procesado
				// (y lo agendo ultimo a proposito porque debe hacer join del resto)
				++condition.var;
				workQueue.push_back( e ); if( !pushed ) pushed = true;
				condition.signal();
			}
			if( pushed ) {
				 ++baseWorkQueueSize;
				if( e->prevWork ) {
					// para que el ultimo work no espere la terminacion de nadie al terminar
					e->prevWork->nextWork = 0;
					// para que el base work no avise a ningun work al terminar
					e->prevWork = 0;
				}
			}
  		} else {
			// Como hay que esperar mucho, no se encola.
			// Pero si se espera que termine, al menos se procesara en este thread.
			retVal = waitCompletion;
		}
	condition.unlock();

	// if waitCompletion the calling threads waits till the job is done
	if( waitCompletion ) {
      e->init();
      e->action();     // this is the method to be derived by the application
		if( ! e->baseEndWork(e) ) { // finish work
			if( sp->wait()<0 ) { // wait for all parts to finish
   	   	UPLOG(LOG_ERR,"UpWorkSplittable::processWork() semaphore wait "
					"error: %s", strerror(errno));
			}
		}
      e->end();
      accountWorkEnd(e);
		// si se agendo al menos un work debo descontar
		if( pushed ) --baseWorkQueueSize;
		delete w;
		delete sp;
	}
	return retVal;
}



