/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <UpTools/UpTimerManager.h>


UpTimerManager::UpTimerManager()
: tickInterval(TICKINTERVALSEC,TICKINTERVALUSEC),unnamedUpTimers(0) { nowTime.now(); }

// setea intervalo entre ejecuciones del catchup
void UpTimerManager::setTickInterval(const UpTimeVal& _tickInterval)
{ tickInterval = _tickInterval; }

// Next method returns time with a precision equal to tickInterval.
// If more precision is needed, then use another thing.
// (sizeof(UpTimeVal) is bigger than architecture max size so we need to lock)
UpTimeVal UpTimerManager::now() {
	mx.lock();
		UpTimeVal ret(nowTime);
	mx.unlock();
	return ret;
}

// This needs no locking because we assume time_t is equal to architecture max size
// and the compiler aligns this. Make sure of this!!
// In 32 bits architectures time_t is 4 bytes
// In 64 bits architecture time_t is 8 bytes
time_t UpTimerManager::now_t()
{ return nowTime.tv_sec; }

// add new timers to UpTimerManager
void UpTimerManager::add(UpTimer* _timer) {
	mx.lock();
		// if when is not set but interval is, then it means that first ocurrence
		// will be an interval from now.
		if( _timer->when==(time_t)0 ) _timer->when = nowTime + _timer->interval;

		tAddList.push_front(_timer);

		if( _timer->className==0 ) ++unnamedUpTimers;
		else ++(namedUpTimers[_timer->className]);
	mx.unlock();
}

// el codigo contenido en run se ejecua en un thread independiente cuando
// se hace startThread()
bool UpTimerManager::startThread()
{ return create(); }


void UpTimerManager::catchUp() {
	// primero se agregan los nuevos
	auxTime.now();
	mx.lock();
	while(!tAddList.empty()) {
		tSet.insert(tAddList.front());
		tAddList.pop_front();
	}
	nowTime = auxTime;
	mx.unlock();
	// catchup loop principal
	// cannot accept insertions in tSet during the catchup loop
	// it is also dangerous to accept them between each iteration of the loop
	UpTimerSet::iterator it,auxit;
	UpTimer* timer;
	it=tSet.begin();
	while( it!=tSet.end() ) {
		if( **it>nowTime ) break;
		timer = *it;
		auxit = it++;
		tSet.erase(auxit);
		if( timer->adminAction(nowTime) ) tSet.insert(timer);
		else {
			auxslist.push_front(timer->className);
			delete timer;
		}
    }
	if( !auxslist.empty() ) {
		mx.lock();
			do {
				const char* n( auxslist.front() );
				if( n==0 ) --unnamedUpTimers; else --(namedUpTimers[n]);
				auxslist.pop_front();
			} while( !auxslist.empty() );
		mx.unlock();
	}
}



#include <iostream>
void* UpTimerManager::run() {
	// usleep needs an unsigned int
 	unsigned int totalMicroSeconds=tickInterval.getTotalMicroSeconds();
	while(1){
    	catchUp();
    	usleep(totalMicroSeconds);
		testCancel(); 
    }
    exit((void*)"coming out of the timeLine thread");
	 // the following is a dummy instruction, just for the compiler warning
	 return (void*)"coming out of the timeLine thread";
}





UpTimerManager::~UpTimerManager() {
	cancel(); // cancel independent timers thread
	for(UpTimerSet::iterator it=tSet.begin(); it!=tSet.end(); it++)
		{ delete *it; }
	mx.lock();
	while(!tAddList.empty())
		{ delete tAddList.front(); tAddList.pop_front(); }
	mx.unlock();
}


unsigned int UpTimerManager::getCount() {
	mx.lock();
		unsigned int ret(unnamedUpTimers);
		CharPtrUIntMap::iterator it = namedUpTimers.begin();
		while( it!=namedUpTimers.end() ) { ret += it->second; ++it; }
	mx.unlock();
	return ret;
}

unsigned int UpTimerManager::getUnnamedCount() {
	return unnamedUpTimers;
}

unsigned int UpTimerManager::getNamedCount() {
	unsigned int ret(0);
	mx.lock();
		CharPtrUIntMap::iterator it = namedUpTimers.begin();
		while( it!=namedUpTimers.end() ) { ret += it->second; ++it; }
	mx.unlock();
	return ret;
}

unsigned int UpTimerManager::getCount(const char* _className) {
	mx.lock();
		unsigned int ret(namedUpTimers[_className]);
	mx.unlock();
	return ret;
}

void UpTimerManager::report(std::ostream& o) {
	mx.lock();
		o << "Unnamed UpTimer count:" << unnamedUpTimers<<'\n';
		CharPtrUIntMap::iterator it = namedUpTimers.begin();
		while( it!=namedUpTimers.end() ) {
			o << it->first << " count:" << it->second <<'\n';
			++it;
		}
	mx.unlock();
}
