/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef UPCONDITION_h
#define UPCONDITION_h

/* Template class UpCondition

This template class allows to use conditional variables for thread
	synchronization

One (or many) threads may block on a conditional variable waiting for
	it to meet certain condition in order to proceed execution.
If another thread signals the blocking thread(s) that a change has been made
	on the conditional variable, the blocked thread(s)	wake up to test the
	condition
The conditional variable is related to a mutex to protect its manipulation
	by many threads.

-----------------------------------------------------------------------------
Example usage:

#include <UpTools/UpCondition.h>
#include<iostream>
using namespace std;

class T : public UpThread {
	private:
		UpCondition<>& c;
	public:
		T(UpCondition<>&_c)	: c(_c) {
			 create(); // start thread
		}

		void*	run() {
			c.lock(); // safe zone for c.var
				while( c.var<5 ) { // test condition
					cout<<"thread "<<pthread_self()
						<<": c.var dos not meet condition, let's wait..."<<endl;
					c.wait(); // release c and wait for signal... relock c on return
					cout<<"thread "<<pthread_self()
						<<": signalled, lets check condition..."<<endl;
				}
				cout<<"thread "<<pthread_self()
					<<": condition met!... exiting thread"<<endl;
			c.unlock();
			return 0;
		}
};

int main() {
	UpCondition<int> c(0);
	T t(c);
	int aux;
	do {
		c.lock(); // safe zone for c.var
			aux = ++c.var;
			cout<<"thread "<<pthread_self()<<": c.var="<<c.var<<endl;
			c.signal();
		c.unlock();
		sleep(1);
	} while( aux<10 );
	
	t.join();
	return 0;

}

*/


#include <UpTools/UpThread.h>
#include <UpTools/UpTimeVal.h>



template<typename dataType=int>
class UpCondition {
private:
	pthread_mutex_t mutex;
	pthread_cond_t cond;
public:
	// constructor and destructor
	UpCondition();
	UpCondition(const dataType& _var);
	~UpCondition();
	// this is the variable that must be protected by the mutex
	// and the variable on which conditions must be tested
	dataType var;
	// get lock on variable
	int lock();
	// unlock variable
	int unlock();
	// test if it can lock the variable, without actually locking it
	int trylock();
	// signal only one thread waiting on this conditional variable
	int signal();
	// signal all threads waiting on this conditional variable
	int broadcast();
	// release the lock and stop until another thread signals the
	// conditional variable (both actions are done atomically)
	int wait();
	// release the lock and stop until another thread signals the
	// conditional variable (both actions are done atomically) or until timout
	int timedwait(time_t t);
	// release the lock and stop until another thread signals the
	// conditional variable (both actions are done atomically) or until timout
	int timedwait(const UpTimeVal& upt);

};

///////////////////////////////////////////////////////////////////////////////
// UpCondition inline Methods /////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
template<typename dataType>
inline UpCondition<dataType>::UpCondition()
	{ pthread_mutex_init(&mutex,0); pthread_cond_init(&cond,0); }

template<typename dataType>
inline UpCondition<dataType>::UpCondition(const dataType& _var)
	: var(_var)
	{ pthread_mutex_init(&mutex,0); pthread_cond_init(&cond,0); }


template<typename dataType>
inline UpCondition<dataType>::~UpCondition ()
	{ pthread_mutex_destroy(&mutex); pthread_cond_destroy(&cond); }

template<typename dataType>
inline int UpCondition<dataType>::lock()
	{ return pthread_mutex_lock(&mutex); }

template<typename dataType>
inline int UpCondition<dataType>::unlock()
	{ return pthread_mutex_unlock(&mutex); }

template<typename dataType>
inline int UpCondition<dataType>::trylock()
	{ return pthread_mutex_trylock(&mutex); }

template<typename dataType>
inline int UpCondition<dataType>::signal()
	{ return pthread_cond_signal(&cond); }

template<typename dataType>
inline int UpCondition<dataType>::broadcast()
	{ return pthread_cond_signal(&cond); }

template<typename dataType>
inline int UpCondition<dataType>::wait()
	{ return pthread_cond_wait(&cond,&mutex); }

template<typename dataType>
inline int UpCondition<dataType>::timedwait(time_t t) {
	timespec ts={ t,0 }; 
	return pthread_cond_timedwait(&cond,&mutex,&ts);
}

template<typename dataType>
inline int UpCondition<dataType>::timedwait(const UpTimeVal& upt) {
	timespec ts={upt.tv_sec,upt.tv_usec*1000};
	return pthread_cond_timedwait(&cond,&mutex,&ts);
}


#endif
