/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef UPCONF_H
#define UPCONF_H

#include <UpTools/UpHashMap.h>
#include <UpTools/UpRegex.h>
#include <UpTools/UpOp.h>

#define UPCONF_MAXLINE 1024

class UpConf {
public:
	UpConf();
	// add parameters from file (it is possible to parse several files)
	bool parseFile(const std::string& confFile);
	bool parseFile(const char* confFile);
	// get number of values stored for a specific parameter
	size_t paramCount(const std::string& label);
	// get param value, if called repetitively, it returns all values stored
	// for the parameter
	const std::string& getParam(const std::string& label);
	const std::string& operator()(const std::string& label);
	// add a value for a specific parameter
	void addParam(const std::string& label, const std::string& value);
	// delete all values for a specific parameter
	void delParam(const std::string& label);
	// delete all values for all parameters
	void clear();
	// number of different parameters
	size_t keyCount();
	// iterates over the different parameter names
	const std::string& getKey(bool first=false);
	// get status after parsing last file
	bool getStatus();
	const std::string& getStatusPhrase();
	// return loaded value or set value if unloaded;
	const std::string& getParamOrSetDefault(const std::string& label, const std::string& defVal);
	// show contents
	friend std::ostream& operator<<(std::ostream& o,UpConf& c);
private:
	StrStrMultiMap params;
	StrStrMultiMap::iterator paramIt,lastIt;
	StrSet keys;
	StrSet::iterator keyIt;
	bool status;
	std::string statusPhrase;
	std::string currentLabel;
	std::string confFile;
	UpRegex dataLineRegex;
	UpRegex eqLineRegex;
	UpRegex labelLineRegex;
	void statError(const char* err,size_t lnum);	
	static std::string emptyString;
};

/////////////////////////////////////////////////////////////////////////////
// inline methods

inline
const std::string& UpConf::operator()(const std::string& label) {
	return getParam(label);
}
inline
bool UpConf::getStatus() {
	return status;
}
inline
const std::string& UpConf::getStatusPhrase() {
	return statusPhrase;
}
inline
bool UpConf::parseFile(const std::string& _confFile) {
	return parseFile(_confFile.c_str());
}
inline
size_t  UpConf::paramCount(const std::string& label) {
	return params.count(label);
}
inline
size_t UpConf::keyCount() {
		return keys.size();
}
inline
void UpConf::delParam(const std::string& label) {
	params.erase(label);
	keys.erase(label);
}
inline
void UpConf::clear() {
	params.clear();
	keys.clear();
	status=true;
	statusPhrase="ok, no data";
}
inline
const std::string& UpConf::getParamOrSetDefault(
								const std::string& _label, const std::string& _defVal )
{
	if( params.count(_label)==0 ) { addParam(_label,_defVal); return _defVal; }
	return getParam(_label);
}




#endif
