/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef UPDNSBASE_H
#define UPDNSBASE_H

#include <UpTools/UpHashMap.h>
#include <UpTools/UpInet.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <string>
#include <ostream>

// forward declarations
class UpDnsQuery;
class UpDnsQuerySimple;
class UpResourceRecord;
class UpRR;
class DnsQueryType;
// globals
extern DnsQueryType dnsQueryType;

/////////////////////////////////////////////////////////////////////////////
// class DnsQueryType: querytype translator

class DnsQueryType {
public:
	DnsQueryType();
	ns_type operator()(const char* q) const;
	ns_type operator()(const std::string& q) const;
	const char* operator()(ns_type q) const;
private:
	typedef std::map<const char*,ns_type,UpLessThan> Qmap;
	Qmap qm;
	const char* qt[256];
};

/////////////////////////////////////////////////////////////////////////////
// class UpResourceRecord: a resource record contained in UpDnsQuery

class UpResourceRecord {
public:
	UpResourceRecord(const ns_rr &record, const ns_msg &msg);
	~UpResourceRecord();
	ns_type 					getType() const;		// all
	const std::string& 	getName() const;		// all
	const std::string&	getString() const;	// all except A and AAAA
	const UpAddress&		getAddr() const;		// A,AAAA
	const std::string& 	getServer() const;	// NS,MX
	const std::string&	getMailAddr() const;	// SOA
	const std::string&	getOrigin() const;	// SOA
	unsigned int			getSerial() const;	// SOA
	unsigned int			getRefresh() const;	// SOA
	unsigned int			getRetry() const;		// SOA
	unsigned int			getExpire() const;	// SOA
	unsigned int			getMinimum() const;	// SOA
	unsigned short			getOrder() const;		// NAPTR
	unsigned short			getPreference() const;	// NAPTR,MX,SRV
	const std::string&	getFlags() const;		// NAPTR
	const std::string&	getService() const;	// NAPTR
	const std::string&	getRegex() const;		// NAPTR
	const std::string&	getReplacement() const;	// NAPTR
	unsigned short			getWeight() const;	// SRV
	unsigned short			getPort() const;		// SRV
	const std::string&	getTarget() const;	// SRV
	friend std::ostream& operator<<(std::ostream& o, const UpResourceRecord& x);
	friend std::string& operator<<(std::string& o, const UpResourceRecord& x);
private:
	UpRR* rrPtr;
	UpResourceRecord* next;
	UpResourceRecord(bool); // private constructor for invalid RR
	UpResourceRecord(const UpResourceRecord &); // it is private so it is not used
	UpResourceRecord& operator=(const UpResourceRecord &); // it is private so it is not used
	friend class UpDnsQuery; // UpDnsQuery can use UpResourceRecord(bool)
	friend class UpDnsQuerySimple; // UpDnsQuerySimple can use UpResourceRecord(bool)
};

/////////////////////////////////////////////////////////////////////////////
// UpDnsQuery status: possible states for an UpDnsQuery or UpDnsQuerySimple
//
typedef enum {
	// status luego de construir el objeto
	QUERY_INITIAL,				// just created
	// status posibles luego de preparar el objeto
	QUERY_CREATION_ERROR,	// Error de creacion del query
	QUERY_CREATED,				// Se creo el query_msg
	// status posible durante la busqueda
	QUERY_SEARCHING,			// Buscando en el DNS 
	// status posibles luego de la busqueda
	QUERY_RECEIVED,			// Respuesta recibida, no parseado
	QUERY_READY,				// Respuesta recibida y parseado OK
	QUERY_TIME_OUT,			// Time out vencido
	QUERY_PARSE_ERROR,		// Respuesta recibida, error de parseo
	QUERY_SOCK_ERROR,			// Error de comunicacion
	// status posible luego de UpResolver::cancelQuery(...)
	QUERY_CANCELLED,			// Cancelada
	// status posible luego de UpResolver::terminateQuery(...)
	QUERY_TERMINATED			// Cancelada
} queryStatus;


extern std::string dnsQueryStatus[];

#endif

#include <UpTools/UpDnsBase.inl>
