/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UPHASHMAP_H
#define UPHASHMAP_H

#include	<string>
#include	<map>
#include	<set>

#if ( __GNUC__ < 4) || ( ( __GNUC__ == 4 ) && ( __GNUC_MINOR__ < 3 )) || ! defined(UNORDERED_MAP)
	#include	<ext/hash_map>
	#include	<ext/hash_set>
	#define UPHASH_MAP		__gnu_cxx::hash_map
	#define UPHASH_SET		__gnu_cxx::hash_set
	#define UPHASH_MULTIMAP	__gnu_cxx::hash_multimap
	#define UPHASH_MULTISET	__gnu_cxx::hash_multiset
#else
	#include	<unordered_map>
	#include	<unordered_set>
	#define UPHASH_MAP		std::unordered_map
	#define UPHASH_SET		std::unordered_set
	#define UPHASH_MULTIMAP	std::unordered_multimap
	#define UPHASH_MULTISET	std::unordered_multiset
#endif

#include <UpTools/UpInet.h>

// Este archivo define UpHash, UpEq, y UpLessThan, necesarios para la
// construccion de containers indexados por strings.

// Ademas se definen los containers StrStrMap y StrStrMap. Ambos son 
// indexados por string y continen strings. Pero el primero puede recorrerse
// en forma ordenada mientras que el segundo no asegura un orden
// definido (aunque es un poco mas rapido)

//////////////////////////////////////////////////////////////////////////
// Function Object de hash para indexar en los hash_maps
class UpHash {
public:
	// null terminated string hash
	inline size_t operator()(const char* _p) const {
		register size_t h(0);
		register const char*  p(_p);
		while( *p ) h = 5*h + *(p++);
		return h;
	}
	// std::string hash
	inline size_t operator()(const std::string& s) const {
		register size_t h(0);
		register const char* p(s.data());
		register const char* q(p+s.size());
		while( p<q ) h = 5*h + *(p++);
		return h;
	}

	inline size_t operator()(const UpSockAddr& usa) const {
		return usa.hash();
	}

	inline size_t operator()(const UpAddress& ua) const {
		return ua.hash();
	}

};


/////////////////////////////////////////////////////////////////
// funcion de igualdad para indexar en los hash_maps
class UpEq {
public:

	inline bool operator()(const char* x, const char* y) const {
		for(;;) {
			if(*x != *y) return false;
			if(! *x) return true;
			x++;y++;
		}
	}

	inline bool operator()(const std::string& x, const std::string& y) const {
		return x==y;
	}

	inline bool operator()(const UpSockAddr& x, const UpSockAddr& y) const {
		return x==y;
	}

	inline bool operator()(const UpAddress& x, const UpAddress& y) const {
		return x==y;
	}

};


/////////////////////////////////////////////////////////////////
// funcion de orden para indexar en los maps
class UpLessThan {
public:
	inline bool operator()(const char* x, const char* y) const {
		while( ( *x == *y ) && *x ) { x++; y++; }
		if(*x < *y) return true;
		return false;
	}

	inline bool operator()(const std::string& x, const std::string& y) const {
		return x<y;
	}

	inline bool operator()(const UpSockAddr& x, const UpSockAddr& y) const {
		return x<y;
	}

	inline bool operator()(const UpAddress& x, const UpAddress& y) const {
		return x<y;
	}
};


////////////////////////////////////////////////////////////////
// hash_map con std::string keys, y std::string elements
typedef UPHASH_MAP<std::string,std::string,UpHash,UpEq> StrStrHashMap;

// hash_multimap con std::string keys, y std::string elements
typedef UPHASH_MULTIMAP<std::string,std::string,UpHash,UpEq>
					 StrStrHashMultiMap;

// map con std::string keys, y std::string elements
typedef std::map<std::string,std::string,UpLessThan> StrStrMap;

// multimap con std::string keys, y std::string elements
typedef std::multimap<std::string,std::string,UpLessThan> StrStrMultiMap;

// hash_set con std::string elements
typedef  UPHASH_SET<std::string,UpHash,UpEq> StrHashSet;

// hash_multiset con std::string elements
typedef  UPHASH_MULTISET<std::string,UpHash,UpEq> StrHashMultiSet;

// set con std::string elements
typedef std::set<std::string,UpLessThan> StrSet;

// multiset con std::string elements
typedef std::multiset<std::string,UpLessThan> StrMultiSet;
#endif
