/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UPINET_H
#define UPINET_H

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/un.h>

#include <vector>
#include <ostream>

#include <UpTools/UpTimeVal.h>

//forward declarations
struct hostent;
class UpAddress;
class UpNet;
class UpRoute;
class UpSockAddr;
class UpHost;
class UpSocket;

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

#define MAX(a,b) ((a<b)?(b):(a))
#define MIN(a,b) ((a<=b)?(a):(b))
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
// class UpAdddress for socket addresses

#define IPBUFSIZE 40

class UpAddress {
public:
	UpAddress();
	~UpAddress();
	UpAddress(const in_addr&);	// network byte order
	UpAddress(const in6_addr&);	// network byte order
	UpAddress(const UpAddress&);
	UpAddress(const char*); //human readable null terminated string
	UpAddress(const char*,size_t len); // human readable not null terminated string
	UpAddress(const std::string&); //human readable string
	UpAddress(const UpSockAddr&);

	const char* c_str()	const;	// null terminated string in human readable form
	int			family() const;	// AF_INET, AF_INET6, AF_UNSPEC

	void clear(); // clear internal data

	UpAddress& operator=(const in_addr&);	// network byte order
	UpAddress& operator=(const in6_addr&);	// network byte order
	UpAddress& operator=(const UpAddress&);
	UpAddress& operator=(const UpSockAddr&);
	// these three guess family type
	UpAddress& operator=(const char*);        // human readable null terminated string
	UpAddress& operator=(const std::string&); // human readable string
	UpAddress& set(const char*);					// human readable null terminated string
	UpAddress& set(const std::string&);			// human readable string
	UpAddress& set(const char*,size_t len);   // human readable not null terminated string
	// these three are faster than the three above because they do not guess type
	UpAddress& setf(const char*,size_t len,int fmly); //human readable not null terminated
	UpAddress& setf(const char*,int fmly);            // human readable null terminated
	UpAddress& setf(const std::string&,int fmly);     // human readable string
	bool operator==(const UpAddress&) const;
	bool operator==(const UpSockAddr&) const;
	bool operator==(const UpHost&) const; // es igual a cualquier de UpHost?

	bool operator>(const UpAddress&) const;
	bool operator<(const UpAddress&) const ;
	bool operator>=(const UpAddress&) const;
	bool operator<=(const UpAddress&) const;
	bool operator!=(const UpAddress&) const;

	UpAddress& operator&=(const UpAddress&);
	UpAddress& operator+=(const UpAddress&);
	UpAddress& operator-=(const UpAddress&);
	UpAddress& operator++(); // preincrement operator
	UpAddress& operator--(); // predecrement operator
	UpAddress  operator++(int); // posincrement operator
	UpAddress  operator--(int); // preincrement operator
	UpAddress  operator~() const;
	UpAddress  operator+(const UpAddress&) const;
	UpAddress  operator-(const UpAddress&) const;
	UpAddress  operator&(const UpAddress&) const;

	void negate(); // negate bits:   (*this) = ~(*this)

	unsigned int maskBitsLen() const;
	void         setMask(unsigned int i,int addrType);


	size_t		hash()	const;	// useful for indexing

	// access to binary data in network byte order (big endian)
	size_t		size()	const;	// 4 for ipv4, 16 for ipv6
	const void*	ptr()   const;	// should be casted to in_addr* or in6_addr*
	void*		ptr();			// should be casted to in_addr* or in6_addr*

	// Writes binary address to mem. (external_addr_buf should have enough space!)
	void	write(void* external_addr_buf)	const;

	// Type casting operators should be used with care:
	// If casting to in_addr an AF_INET6 UpAddress, or casting
	// to in6_addr an AF_INET UpAddress, will result in a garbage address.
	// Again addresses are returned in network byte order (big endian)
	operator in_addr() const;
	operator in6_addr() const;
	// Operators to cast address to a numeric type. Very useful for some algorithms.
	// (return value is in host order)
	operator float();
	operator double();
	operator long double();
	operator unsigned int(); // for IPv6 returns the less significant 4 bytes

	// output to a stream or string
	friend std::ostream& operator<<(std::ostream& o, const UpAddress& a);
	friend std::string& operator<<(std::string& o, const UpAddress& a);

private:

	typedef union {
		in_addr a4;
		in6_addr a6;
		struct { // big endian 128 bit integer (net order is big endian)
			unsigned long long high; // 64 bits integer
			unsigned long long low;  // 64 bits integer
		} be128;
	} addrUnion;

	// UpAddress stores addresses in network byte order (big endian)
	int addrType; // AF_UNSPEC, AF_INET, AF_INET6
	addrUnion upAddr; // for ip4 and ipv6 addresses
	mutable char* buf; // for human readable null terminated string
	friend class UpNet;
	friend class UpRoute;

};
//////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
// class UpSockAddr for socket addresses


class UpSockAddr {
public:
	UpSockAddr();
	UpSockAddr(const UpSockAddr&);
	~UpSockAddr();
	UpSockAddr& operator=(const UpSockAddr&);

	void allocSA(int family=AF_INET);

	int family() const; // AF_UNIX, AF_INET, AF_INET6 or AF_UNSPEC=0
	unsigned short port() const; // if not AF_INET or AF_INET6 returns 0

	// operators
	bool operator==(const UpSockAddr&) const;
	bool operator==(const UpAddress&) const;
	bool operator==(const UpHost&) const; // test equality to any of UpHost's addresses
	bool operator>(const UpSockAddr&) const;
	bool operator<(const UpSockAddr&) const ;
	size_t hash() const ; // useful for indexing

	// methods to use with UpAddress
	UpSockAddr(const UpAddress&,unsigned short port);
	void set(const UpAddress&, unsigned short port);

	// Methods for in_addr and in6_addr types
	// the following methods assume network byte order in the ip address
	// but hardware byte order in the port
	UpSockAddr(const in_addr&,unsigned short port);
	UpSockAddr(const in6_addr&,unsigned short port);
	UpSockAddr& operator=(const in_addr&);  // port=0
	UpSockAddr& operator=(const in6_addr&); // port=0
	void set(const in_addr& in, unsigned short port);
	void set(const in6_addr& in6, unsigned short port);
	// _addr should be in_addr* or in6_addr* depending on fmly
	void set(const void* _addr, unsigned short port,int fmly=AF_INET);

	// Methods for struct sockaddr_xx types
	UpSockAddr(const sockaddr* saddr);
	UpSockAddr(const sockaddr_un* saddr);
	UpSockAddr(const sockaddr_in* saddr);
	UpSockAddr(const sockaddr_in6* saddr);
	UpSockAddr& operator=(const sockaddr* saddr);
	UpSockAddr& operator=(const sockaddr_un* saddr);
	UpSockAddr& operator=(const sockaddr_in* saddr);
	UpSockAddr& operator=(const sockaddr_in6* saddr);
	sockaddr* ptr(); // pointer to use in unix calls
	const sockaddr* ptr() const ; // pointer to use in unix calls
	void write(void*) const; // writes binary sockaddress to mem.
	size_t size() const; // space needed by the buffer to be used by write(void*)

	// methods to use with human readable strings
	// UpSockAddr accepts the following formats in these methods:
	//	"200.44.5.23" "200.44.5.23 80" "200.44.5.23:80"
	// "3ffe:1212::0003" "3ffe:1212::0003 443"
	// "[3ffe:1212::0003]" "[3ffe:1212::0003] 443" "[3ffe:1212::0003]:443"
	// "/tmp/usock"
	// port is in hardware byte order
	// set is faster than operator= because the family is not guessed from data
	UpSockAddr(const char*);
	UpSockAddr(const char*, unsigned short port);
	UpSockAddr& operator=(const char*);
	bool set(const char* addr, unsigned short port,int fmly=AF_INET);
	const char* c_str()       const;	// format "abcd::0001 443"
	const char* c_str_2dots() const;	// format "[abcd::0001]:443"
	const char* c_str_ip()    const;	// format "abcd::001"

	// the same as above but for std::string
	UpSockAddr(const std::string&);
	UpSockAddr(const std::string&, unsigned short port);
	UpSockAddr& operator=(const std::string&);
	bool set(const std::string& addr, unsigned short port,int fmly=AF_INET);
	bool set(const std::string& addr,int fmly=AF_INET); //if !AF_UNIX then port=0
	const std::string& string()       const;	// format "abcd::0001 443"
	const std::string& string_2dots() const; 	// format "[abcd::0001]:443"
	const std::string& string_ip()    const; 	// format "abcd::001"

	// hostPort Methods accept the same formats as before, and also domains,
	// but do not allow unix paths:
	//	"200.44.5.23" "200.44.5.23 80" "200.44.5.23:80"
	// "3ffe:1212::0003" "3ffe:1212::0003 443"
	// "[3ffe:1212::0003]" "[3ffe:1212::0003] 443" "[3ffe:1212::0003]:443"
	// "www.palermo.edu" "www.palermo.edu 80" "www.palermo.edu:80"
	bool hostPort(const char* hostAddr, unsigned short port);
	bool hostPort(const char* hostAddr);
	bool hostPort(const std::string& hostAddr, unsigned short port);
	bool hostPort(const std::string& hostAddr);

	// used to print UpSockAddr to ostreams or strings
	friend std::ostream& operator<<(std::ostream& o, const UpSockAddr& upa);
	friend std::string& operator<<(std::string& o, const UpSockAddr& upa);

private:

	typedef union {
		sockaddr*		a;
		sockaddr_in*	in;
		sockaddr_in6*  in6;
		sockaddr_un*	un;
	} sockaddrPtrUnion;
	typedef enum {
		SOCKADDR_2DOTS,
		SOCKADDR_SPACE,
		SOCKADDR_IP
	} sockaddrFormat;

	sockaddrPtrUnion sa;
	mutable std::string str;
	mutable sockaddrFormat format;

	void allocAux(int family);
	bool testAssign4(const char* p,unsigned short port);
	bool testAssign6(const char* p,unsigned short port);
	bool testAssignH(const char* p,unsigned short port);
	bool testAssignU(const char* p);
	void assignAux(const UpSockAddr&);
	void assignAux(const char*);
	void assignAux(const char*,unsigned short);

	friend class UpNet;
	friend class UpRoute;
};
//////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
// Class UpHost

#define HOSTBUFSIZE 256 // used for GNU GLIBC2 gethostbyname_r()

class UpHost {
public:
	UpHost();
	UpHost(const char*);
	UpHost(const std::string&);
	UpHost(const UpHost&);
	UpHost& operator=(const char*);
	UpHost& operator=(const std::string&);
	UpHost& operator=(const UpHost&);
	bool operator==(const UpHost&) const; // igual a cualquiera
	bool operator==(const UpAddress&) const; // igual a cualquiera
	bool operator==(const UpSockAddr&) const; // igual a cualquiera

	bool										status() const;
	int										errNumber() const;
	const std::string&					errPhrase() const;

	int 										family() const;
	size_t 									addrSize() const;
	const std::string&					hostName() const;
	const std::vector<std::string>&	aliases() const;
	const std::vector<UpAddress>&		addresses() const;
	const UpAddress&						address() const;

	friend std::ostream& operator<<(std::ostream& o, const UpHost& h);
	friend std::string& operator<<(std::string& o, const UpHost& h);

private:
	std::string              _hostName;		// host name
	std::vector<std::string> _aliases;		// container for the aliases
	std::vector<UpAddress>   _addresses;	// container for the addresses
	int addrType;									// AF_UNIX, AF_INET, AF_INET6
	size_t addrLen;								// depends on the address type
	bool lookupOk;									// did the host lookup succeed?
	int lookupErrno;							// error number in case the lookup fail

	bool init(const char* host);
	void init(const hostent* h);

};


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
// Class UpSocket


#define DEFAULTQUEUESIZE	5  // old BSD's could listen up to 5 incoming max

class UpSocket {
public:
	UpSocket();
	~UpSocket(); // closes socket before destroying

	// set flags
	void setNonBlocking(bool);	// sets non-blocking operations
	void setDirect(bool);	// sets direct operations (flushes immedialtely)
	void setAsync(bool);		// sets async operations (SIGIO when data available)
	// set maximum wait time for connect(...), write(...) or send(...)
	void setSendTimeout(const UpTimeVal& tv);
	// set maximum wait time for accept(...), read(...) or recv(...)
	void setRecvTimeout(const UpTimeVal& tv);


	// set options
	// after calling fd(), bind(...) or connect(...) these methods cannot change
	// sockType and return false if sockType differs from what was requested
	bool setDatagram();		// sets sockType = SOCK_DGRAM   (udp for ip sockets)
	bool setStream();			// sets sockType = SOCK_STREAM  (tcp for ip sockets)

	// administrative
	void showFcntlFlags();
	int  fd();
	const UpSockAddr& getLocalSockAddr();
	const UpSockAddr& getRemoteSockAddr();

	// for clients or servers (not necessary for clients)
	bool bind(const UpSockAddr& localSA);
	bool bind(const sockaddr* localSA);
	bool bind(const std::string& localAddr);
	bool bind(const std::string& localAddr,unsigned short localPort);
	bool bind(const char*);
	bool bind(const char*,unsigned short port);
	// these two acept only AF_INET and AF_INET6
	bool bindInterface(const std::string& interf,unsigned short localPort,int _family=AF_INET);
	bool bindInterface(const char* interf, unsigned short localPort,int _family=AF_INET);

	// for server sockets
	bool listen();
	bool listen(int queueSize);

	// connected socket is initialized with the new accepted connection
	bool accept(UpSocket& connectedSocket);

	// for client sockets
	// generally used by stream sockets
	bool connect(const sockaddr* remoteSA);
	bool connect(const UpSockAddr& remoteSA);
	bool connect(const std::string& remoteAddr);
	bool connect(const std::string& remoteAddr, unsigned short remoteP);
	bool connect(const char* remoteAddr);
	bool connect(const char* remoteAddr, unsigned short remoteP);

	// for all sockets
	void close();

	// for all sockets
	// generally used by stream sockets
	int write(const std::string&);
	int write(const char*buf ,size_t len);
	int writeUrg(const std::string&);
	int writeUrg(const char* buf, size_t len);

	// for ipv4 or ipv6 sockets
	// generally used by datagram sockets
	int sendto(const std::string&, const UpSockAddr& remoteSA);
	int sendto(const char*,size_t len, const UpSockAddr& remoteSA);
	int sendtoUrg(const std::string& data, const UpSockAddr& remoteSA);
	int sendtoUrg(const char* data, size_t len, const UpSockAddr& remoteSA);

	// generally used by stream sockets
	int read(char*,size_t len);
	int readUrg(char*,size_t len);
	// int read(std::string&);
	// int readUrg(std::string&);

	// remoteSA is rewritten with de data of the incoming packet
	// generally used in datagram sockets
	// if data does not fit in the buf, the extra bytes are discarded
   int recvFrom(char* buf, size_t len, UpSockAddr& remoteSA);
   int recvFromUrg(char* buf, size_t len, UpSockAddr& remoteSA);
	// int recvFrom(std::string&, UpSockAddr& remoteSA);
   // int recvFromUrg(std::string& data, UpSockAddr& remoteSA);

private:
	int sd;
	UpSockAddr localSockAddr,remoteSockAddr;
	int family;
	//flags
	bool nonBlocking;	// default false
	bool direct;		// default false
	bool async;			// default false
	UpTimeVal sendTimeout; // default inifinite
	UpTimeVal recvTimeout; // default inifinite
	// administrative data
	int  sockType;		// default SOCK_STREAM;
	// internal methods
	bool bind();		// internal bind method
	bool connect();	// internal connect method
	bool setFlags();	// internal setFlags method

};


////////////////////////////////////////////////////////////////////////
// UpNet is a class for manipulating networks.
// (for fast route table lookups see UpRoute)

typedef enum {
	UPNR_BEFORE,
	UPNR_AFTER,
	UPNR_SUPERNET,
	UPNR_SUBNET,
	UPNR_SUBNET_COLLIDES,
	UPNR_EQUAL,
	UPNR_INVALID_RELATION
} UpNetRelation;



class UpNet {
public:

	UpNet();
	UpNet(const UpNet&);
	UpNet(const UpAddress& baseAddr,const UpAddress& mask);
	UpNet(const UpAddress& baseAddr,unsigned int maskLen);
	UpNet(in_addr baseAddr,in_addr mask);   // net order
	UpNet(const in6_addr& baseAddr,const in6_addr& mask);   // net order
	UpNet(const std::string& net);
	virtual ~UpNet();

	int family() const; // AF_INET, AF_INET6, AF_UNSPEC

	virtual void clear(); // clear internal data

	virtual void set(const UpNet&);
	virtual void set(const UpAddress& baseAddr,const UpAddress& mask);
	virtual void set(const UpAddress& baseAddr,unsigned int maskLen);
	virtual void set(in_addr baseAddr,in_addr mask);   // net order
	virtual void set(const in6_addr& baseAddr,const in6_addr& mask);   // net order
	virtual bool set(const std::string& net);

	UpNet&         operator=(const UpNet&);
	virtual UpNet& operator=(const std::string&); // same as set(...)
	// examples: "200.64.0.0/10" or "200.64.0.0/255.192.0.0"
	// examples: "2001::/16"     or "2001::/ffff::"

	// next methods check the relation between addr and this net
	UpNetRelation check(const UpAddress& addr) const;
	UpNetRelation check(const UpSockAddr& addr) const;
	UpNetRelation check(in_addr addr) const;   // net order
	UpNetRelation check(const in6_addr& addr) const;   // net order
	UpNetRelation check(const sockaddr* sa) const;   // net order
	UpNetRelation check(const sockaddr_in* sa) const;   // net order
	UpNetRelation check(const sockaddr_in6* sa) const;   // net order
	UpNetRelation check(const UpNet& net) const;

	// next methods check only (true or false) if addr is included in this net
	bool includes(const UpAddress& addr) const;
	bool includes(const UpSockAddr& addr) const;
	bool includes(in_addr addr) const;   // net order
	bool includes(const in6_addr& addr) const;   // net order
	bool includes(const sockaddr* addr) const;   // net order
	bool includes(const sockaddr_in* addr) const;   // net order
	bool includes(const sockaddr_in6* addr) const;   // net order
	bool includes(const UpNet& net) const;

	const UpAddress& getBase() const; // network base address
	const UpAddress& getMask() const;
	unsigned int getMaskLen() const;
	UpAddress getLastIp() const; // higher address in net
	UpAddress getSize() const; // address count (an integer up to 128 bit long)
	// To get size as a simple number, cast it to a float, double or long double:
	// UpNet n("1234:5000::/20"); double siz = (double)n.getSize();

	
	virtual void append(std::ostream& o) const;
	virtual void append(std::string& o) const;
	friend std::ostream& operator<<(std::ostream& o,const UpNet& net);
	friend std::string&  operator<<(std::string& o,const UpNet& net);

	static const std::string& relationString(UpNetRelation);

protected:
	UpAddress ip;
	UpAddress m;

private:
	void checkPrivateData();
	static std::string relString[];
	friend class UpRoute;
};


#endif

#include <UpTools/UpInet.inl>
