/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Ignacio Kaufmann for his work in this class

/////////////////////////////////////////////////////////////////////////////
// class UpResolver: executes synchronous and asynchronous DNS queries
// class UpDnsQuery: DNS query and resource records, and completion methods
// class UpResourceRecord: a resource record contained in UpDnsQuery
// class DnsQueryType: query-type translator
// class UpResolverSimple: executes only blocking DNS queries
// class UpDnsQuerySimple: blocking DNS queries
/////////////////////////////////////////////////////////////////////////////

// Class UpResolver executes synchronous (blocking) and asynchronous
// (non-blocking) DNS queries. It is thread safe.

// It helps applications to utilize DNS services without degrading
// performance by waiting for DNS server answers.

// A thread can execute an UpDnsQuery and block until
// it completes successfully or it timouts. Or it can post
// an UpDnsQuery and continue its own processing, while
// UpResolver will execute the UpDnsQuery::complete() or
// UpDnsQuery::timeout() method when it receives the corresponding
// answer from the DNS server or until the UpDnsQuery timeouts.

// UpResolver can handle lots of pending UpDnsQuery's posted by any
// of the threads of the application. Each thread must use its
// own UpDnsQuery's.

// Applications can derive classes from UpDnsQuery in order to taylor
// UpDnsQuery::complete() or UpDnsQuery::timeout() to its own needs.
// These methods are called from the UpResolver own thread when needed.

// UpResolver does not own the UpDnsQuery's, the user owns them.
// When an UpDnsQuery is destroyed it is automatically cancelled and
// retired from UpResolver.

// WARNING: if a class derived from UpDnsQuery reimplements methods
// UpDnsQuery::complete() and UpDnsQuery::timeout(), so that they use
// data outside the UpDnsQuery base class data, it must call
// UpDnsQuery::cancel() before destroying that data.

/////////////////////////////////////////////////////////////////////////////
// query types and UpResourceRecord methods

// SOA:   getOrigin(), getMailaddr() , getSerial(),
//        getRefresh(), getRetry(), getExpire(), getMinimum()
// NAPTR: getOrder(), getPreference(), getFlags(), getService(), getRegex(),
//        getReplacement()
// SRV:   getPreference(), getWeight(), getPort(), getTarget()
// MX:    getPreference(), getServer()
// A:     getAddr()
// AAAA:  getAddr()
// PTR:   getString()
// CNAME: getString()
// TXT:   getString()
// NS:    getString()
		
/////////////////////////////////////////////////////////////////////////////

#ifndef UPRESOLVER_H
#define UPRESOLVER_H

#include <UpTools/UpDnsBase.h>
#include <UpTools/UpThread.h>
#include <UpTools/UpCondition.h>

// forward declarations
class	UpResolver;

/////////////////////////////////////////////////////////////////////////////
// UpDnsQuery

// status in resolver
typedef enum {
	QR_IDLE,
	QR_MARKED_FOR_COMPLETION,
	QR_REPOSTED,
	QR_MARKED_FOR_DELETION,
	QR_AWAITING_RESPONSE
} queryResolverStatus;

typedef enum {
	QUERY_BLOCKING,    // wait for completion when calling resolver.query(q)
	QUERY_NONBLOCKING  // do not wait when calling resolver.query(q)
} dnsQueryBlockingType;

typedef enum {
	QUERY_USE_CALLBACK,		// execute query.completed() on completion
	QUERY_DONT_USE_CALLBACK // do nothing on completion
} dnsQueryCallbackType;


class UpDnsQuery {
	private:
		unsigned int	id;
		queryStatus 	status;
		queryResolverStatus rstatus;
		std::string 	domain;
		ns_type 			type;
		unsigned char	query_buff[NS_PACKETSZ];
		unsigned int 	querySize;
		unsigned char	answer_buff[NS_PACKETSZ];
		unsigned int	answerSize;
		ns_msg 			msg;
		time_t			transmitTime;
		unsigned int	retra;
		unsigned int 	count;  		// total elements in answer
		mutable unsigned int	parsed; 		// elements already parsed in answer
		mutable UpResourceRecord* first;	// pointer to linked list of resource records
		mutable UpResourceRecord* current;  // pointer to linked list of resource records
		dnsQueryCallbackType callbackType;  // call completed on completion?
		dnsQueryBlockingType blockingType; // block on resolver.query(q)?
		// resolverPtr is only handled inside blockingCondition safe zone
		UpResolver* 		resolverPtr; // UpResolver in charge of this query
		// conditional variable
		// -1 : query is not proceeding
		//  0 : query is proceeding, no threads are blocked on it.
		//      Will execute UpDnsQuery::complete() at the end of proceeding.
		//  N : query is proceeding, N threads are blocked on it.
		//		  Will not execute UpDnsQuery::complete() at the end of proceeding.
		UpCondition<int>	blockingCondition;
		// next two methods are called by UpResolver
		void allowWaiting(); // allows threads to wait for query response
		void releaseWaitingThreads(); // reactivates waiting threads

		static UpResourceRecord invalidResourceRecord;

		bool initialResponseParse();
		bool prepare();
		void clear();
		void deleteRecords();
		friend class UpResolver;

	public:
		// constructor and destructor
		UpDnsQuery( const std::string& _domain, ns_type _type,
						dnsQueryCallbackType _callbackType=QUERY_DONT_USE_CALLBACK,
					   dnsQueryBlockingType _blockingType=QUERY_BLOCKING );
		UpDnsQuery( const UpAddress& _upa,          // reverse query
						dnsQueryCallbackType _callbackType=QUERY_DONT_USE_CALLBACK,
					   dnsQueryBlockingType _blockingType=QUERY_BLOCKING );
		UpDnsQuery();
		virtual ~UpDnsQuery();
		// reset query data
		void set( const std::string& _domain, ns_type _type,
					 dnsQueryCallbackType _callbackType=QUERY_DONT_USE_CALLBACK,
					 dnsQueryBlockingType _blockingType=QUERY_BLOCKING );
		void set( const UpAddress& _upa,          // reverse query
						dnsQueryCallbackType _callbackType=QUERY_DONT_USE_CALLBACK,
					   dnsQueryBlockingType _blockingType=QUERY_BLOCKING );
		// cancel query execution
		void cancel();
		// If query is proceeding, it stops the calling thread execution until
		// successful or unsuccessful completion.
		// if query is not proceeding, it does nothing
		// This method may allow for more than one thread to wait for completion.
		bool waitCompletion(); // returns true only on successful completion
		// Method to be redefined in derived classes for asynchronus queries
		// in order to inform the thread or object originating the query.
		// The method will only be called if no thread is blocked waiting for
		//	the end of query processing.
		virtual void completed();
		// getFirst and getNext() return the resource records obtained from server.
		// They return an invalid record if query is not incomplete or failed.
		///  getFirst() returns the first record
		const UpResourceRecord&	getFirst() const;
		/// getNext() returns the next record, except in the following cases:
		/// _ if getFirst() and getNext() were never called, it returns the first record
		/// _ after calling resetGetNext() it returns the first record
		/// _ if there are no more records it returns an invalid record
		const UpResourceRecord&	getNext() const;
		/// resetGetNext() ensures that the following time that
		/// getNext() is called it returns the first record
		void resetGetNext();

		// regular methods
		unsigned int 				getCount() const; // number of records
		queryStatus 				getStatus() const;// query status
		unsigned int				getId() const;		// query id
		UpResolver*					getResolverPtr() const; //Pointer to resolver in use
		const std::string&		getDomain() const;// query domain
		ns_type						getType() const;	// query type
		friend std::ostream& operator<<(std::ostream& o,const UpDnsQuery& q);
		friend std::string& operator<<(std::string& o,const UpDnsQuery& q);
};


///////////////////////////////////////////////////////////////////////////////
//Class UpResolver

class UpResolver : public UpThread {
	private: // data
		time_t lastTick;
		time_t tickInterval;

		UpSockAddr dnsServerAddr;
		UpSocket sock;
		unsigned char   buff[NS_PACKETSZ];
		UpMutex mx;
		std::map<unsigned int,UpDnsQuery*> queryMap;

		time_t retrans;
		unsigned int retry;

	private: // methods
		bool init();		// Initializes data
		void* run(); 		// contains listener thread code
		bool selfQuery(UpDnsQuery& q);
		bool retransmitQuery(UpDnsQuery&, time_t);

	public:
		UpResolver(); // gets data from libresolv
		UpResolver(const std::string& _nameserver,time_t _retrans,
						unsigned int _retry);
		~UpResolver();
		// next method cancels a query and releases waiting threads
		void cancelQuery(UpDnsQuery& q);
		// next method is called by UpDnsQuery::~UpDnsQuery()
		void terminateQuery(UpDnsQuery& q);
		// Next method sends query to server and schedules a timeout.
		// It waits for completion or immediately returns control to the caller
		//  depending on q.blockingType.
		// It will execute q.completed() on completion depending on q.callbackType
		bool query(UpDnsQuery& q);
};

#endif

/////////////////////////////////////////////////////////////////////////////
// inline methods

#include <UpTools/UpResolver.inl>

