/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Ignacio Kaufmann for his work in this class

/////////////////////////////////////////////////////////////////////////////
// class UpResolverSimple: executes only blocking DNS queries
// class UpDnsQuerySimple: blocking DNS queries
// class UpResourceRecord: a resource record contained in UpDnsQuerySimple
//                         (see UpResolver.h). UpDnsQuerySimple may contain
//                         many of these.
// class DnsQueryType: query-type translator (see UpResolver.h)
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// query types and UpResourceRecord methods

// SOA:   getOrigin(), getMailaddr() , getSerial(),
//        getRefresh(), getRetry(), getExpire(), getMinimum()
// NAPTR: getOrder(), getPreference(), getFlags(), getService(), getRegex(),
//        getReplacement()
// SRV:   getPreference(), getWeight(), getPort(), getTarget()
// MX:    getPreference(), getServer()
// A:     getAddr()
// AAAA:  getAddr()
// PTR:   getString()
// CNAME: getString()
// TXT:   getString()
// NS:    getString()
		
/////////////////////////////////////////////////////////////////////////////

#ifndef UPRESOLVER_SIMPLE_H
#define UPRESOLVER_SIMPLE_H

#include <UpTools/UpDnsBase.h>

/////////////////////////////////////////////////////////////////////////////

class UpDnsQuerySimple {
	private:
		unsigned int	id;
		queryStatus 	status;
		std::string 	domain;
		ns_type 			type;
		unsigned char	query_buff[NS_PACKETSZ];
		unsigned int 	querySize;
		unsigned char	answer_buff[NS_PACKETSZ];
		unsigned int	answerSize;
		ns_msg 			msg;
		unsigned int 	count;  		// total elements in answer
		mutable unsigned int	parsed; 		// elements already parsed in answer
		mutable UpResourceRecord* first;	// pointer to linked list of resource records
		mutable UpResourceRecord* current;  // pointer to linked list of resource records

		static UpResourceRecord invalidResourceRecord;

		bool initialResponseParse();
		bool prepare();
		void clear();
		void deleteRecords();
		friend class UpResolverSimple;

	public:
		// constructor and destructor
		UpDnsQuerySimple( const std::string& _domain, ns_type _type);
		UpDnsQuerySimple( const UpAddress& upa); // reverse query
		UpDnsQuerySimple();
		virtual ~UpDnsQuerySimple();
		// reset query data
		void set( const std::string& _domain, ns_type _type);
		void set( const UpAddress& upa); // reverse query
		// getFirst and getNext() return the resource records obtained from server.
		// They return an invalid record if query is incomplete or failed.
		///  getFirst() returns the first record
		const UpResourceRecord&	getFirst() const;
		/// getNext() returns the next record, except in the following cases:
		/// _ if getFirst() and getNext() were never called, it returns the first record
		/// _ after calling resetGetNext() it returns the first record
		/// _ if there are no more records it returns an invalid record
		const UpResourceRecord&	getNext() const;
		/// resetGetNext() ensures that the following time that 
		/// getNext() is called it returns the first record
		void resetGetNext();
		// regular methods
		unsigned int 				getCount() const; // number of records
		queryStatus 				getStatus() const;// query status
		unsigned int				getId() const;		// query id
		const std::string&		getDomain() const;// query domain
		ns_type						getType() const;	// query type
		friend std::ostream& operator<<(std::ostream& o,const UpDnsQuerySimple& q);
		friend std::string& operator<<(std::string& o,const UpDnsQuerySimple& q);
};

///////////////////////////////////////////////////////////////////////////////
//Class UpResolver
class UpResolverSimple {
	private: // data

		UpSockAddr dnsServerAddr;
		UpSocket sock;

		time_t retrans;
		unsigned int retry;

	private: // methods
		bool init();		// Initializes data

	public:
		UpResolverSimple(); // gets data from libresolv
		UpResolverSimple(const std::string& _nameserver,time_t _retrans,
						unsigned int _retry);
		~UpResolverSimple();
		bool query(UpDnsQuerySimple& q);
};

#endif

/////////////////////////////////////////////////////////////////////////////
// inline methods

#include <UpTools/UpResolverSimple.inl>

