/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
// Special thanks to Horacio J. Peña for his work in this class

#ifndef UPROUTE_H
#define UPROUTE_H

#include <UpTools/UpInet.h>
#include <list>

// UpRoute is a class for ip routing tables, and ultra fast table lookup even
// when tables are very big.

class UpRoute : public UpNet {
public:

	UpRoute();
	UpRoute(const UpRoute&);
	UpRoute(const UpNet&,unsigned int _metric=0);
	UpRoute(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric=0);
	UpRoute(const UpAddress& baseAddr,unsigned int maskLen,unsigned int _metric=0);
	UpRoute(in_addr baseAddr,in_addr mask,unsigned int _metric=0);   // net order
	UpRoute(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric=0);   // net order
	UpRoute(const std::string& net);
	UpRoute(const std::string& net,unsigned int _metric);
	virtual ~UpRoute();

	// This class inherits getBase(), getMask(), getMaskLen(),
	// getLastIp(), getSize(), all check(...) and all includes(...) methods

	// clear internal data
	virtual void clear(); // clear internal data
	// set internal data
	virtual void set(const UpRoute&);
	virtual void set(const UpNet&,unsigned int _metric=0);
	virtual void set(const UpAddress& baseAddr,const UpAddress& mask,unsigned int _metric=0);
	virtual void set(const UpAddress& baseAddr,unsigned int maskLen,unsigned int _metric=0);
	virtual void set(in_addr baseAddr,in_addr mask,unsigned int _metric=0);   // net order
	virtual void set(const in6_addr& baseAddr,const in6_addr& mask,unsigned int _metric=0);   // net order
	virtual bool set(const std::string& net,unsigned int _metric);
	virtual bool set(const std::string& net);
	void setMetric(unsigned int _metric);

	UpRoute&         operator=(const UpRoute&);
	virtual UpRoute& operator=(const std::string&); // same as set(...)
	// examples: "200.64.0.0/10 5" or "200.64.0.0/255.192.0.0 8"
	// examples: "2001::/16 20"     or "2001::/ffff:: 0"

	// Add an UpRoute to internal network tree. The route being added must be a
	// subnet of (*this) in order to be accepted.
	// If it is accepted then it returns UPNR_SUBNET, and the acceptor stays in
	// charge of deleting the added UpRoute.
	UpNetRelation addCheck(UpRoute* r);
	// Remove an UpRoute from internal network tree, and optionally its subnets.
	// Returns true if any UpRoutes are deleted.
	bool remove(const UpAddress& _ip,const UpAddress& _m,bool includeSubnets=false);
	bool remove(const UpAddress& _ip,unsigned int _m,bool includeSubnets=false);
	// Next method optimizes internal network tree in order to speed up lookups
	// (strongly recommended for big tables)
	void optimize();
	// Remove all auxiliary nets added during optimization
	void unoptimize();
	// Return a pointer to the most specific UpRoute including addr,
	// or 0 if no UpRoute is found that includes addr.
	// If useMetric is true, then the UpRoute with the smaller metric including
	// addr is returned instead.
	const UpRoute* lookup(const UpAddress& addr,bool useMetric=false) const;
	const UpRoute* lookup(const UpSockAddr& sa ,bool useMetric=false) const;
	const UpRoute* lookup(const UpNet& net     ,bool useMetric=false) const;
	const UpRoute* lookup(in_addr addr         ,bool useMetric=false) const;
	const UpRoute* lookup(const in6_addr& addr ,bool useMetric=false) const;
	// Return all the UpRoutes in the tree that contain addr
	// (the user may later choose the route he likes)
	std::list<const UpRoute*> lookupAll(const UpAddress& addr) const;
	// show internal network tree
	void tree(std::ostream& o, int indent=0) const;

	virtual void append(std::ostream& o) const;
	virtual void append(std::string& o) const;
	friend std::ostream& operator<<(std::ostream& o,const UpRoute& net);
	friend std::string&  operator<<(std::string& o,const UpRoute& net);

protected:
	//	UpAddress ip and m are inherited from UpNet
	unsigned int metric;

private:
	UpRoute *next,*previous;
	UpRoute *firstSon,*lastSon;
	bool auxNet;
	void insertHead(UpRoute* newRoute);
	void insertBefore(UpRoute* currentNet, UpRoute* newRoute);
	void insertTail(UpRoute *newRoute);
	void replace(UpRoute* currentNet, UpRoute* newRoute);
	void unlist(UpRoute* oldRoute);
	void becomeAncestor(UpRoute*,UpRoute*);
};



#endif

#include <UpTools/UpRoute.inl>
