/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef SMARTLINKPTR_H
#define SMARTLINKPTR_H

// template<typename T> class UpSmartLinkPtr<T>
// Smart pointer that deletes the object it points to when the object is
// not being pointed by itself nor any other smart pointer.
// It is efficient for handling objects residing in STL containers, and
// objects that must be accessed from many entities and from many threads.
// It must point to objects derived from class UpLinkable.
// UpLinkable objects mantain a count of the number of smart pointers that
// point them.

// IMPORTANT:

// Objects pointed by smart pointers must not be automatic data nor arrays!
// That is, they must not reside in the stack segment, nor be declared
// as automatic data in any class, nor be be allocated in arrays.

// In order to monitor the number of objects from classes derived from
// UpLinkable, and to be able to make reports at any time using the
// global object upLinkableMonitor, compile with -DUPLINKABLE_MONITOR
// This is a very useful option.

// DEBUG OPTIONS:

// UpLinkable objects can detect if they are being created in the stack
// and stop execution in such case. To enable this check compile with
// -DUPLINKABLE_HEAPCHECK

// UpLinkable objects are not yet detecting if they are created in arrays
// or as automatic data in a specific class. In order to detect if they are
// created in the heap but as automatic data or in an array, we should
// overload new and new[].

// In order to log the creation and destruction of objects derived from
// class UpLinkable, compile with -DUPLINKABLE_LOG_EVENTS

// In order to log the linking and unlinking of objects derived from
// class UpLinkable, compile with -DUPLINKABLE_LOG_LINKS

// PERFORMANCE:
// All debug options slow down processing time. But when they are not
// activated, they disappear from the object code. So, when no debug
// options are activated, UpSmartLinkPtr and UpLinkable are very fast!
// Anyway, -DUPLINKABLE_MONITOR is not very demanding, and very useful.


///////////////////////////////////////////////////////////////////////////
// Good examples:
//
//	class MyClass : public UpLinkable {
// private:
//    MyClass() : UpLinkable("MyClass") {  }
//    MyClass(int a,int b) : UpLinkable("MyClass") { ... }
//    ...
//	};
//
// UpSmartLinkPtr<MyClass> MyClassPtr1 = new MyClass;
//
// UpSmartLinkPtr<MyClass> MyClassPtr2 = new MyClass(arg1,arg2);
//
// UpSmartLinkPtr<MyClass> MyClassPtr3 = new MyClass(*MyClassPtr1);
//
// UpSmartLinkPtr<MyClass> MyClassPtr4 = MyClassPtr2;
//
// UpSmartLinkPtr<MyClass> MyClassPtr5;
//
//	MyClassPtr3 = MyClassPtr2;
//
//	MyClassPtr1 = 0;
//
// MyClassPtr1->exec();
//
//	void someFunc(UpSmartLinkPtr<MyClass> ptrArg) {
//		if( ptrArg!=0 ) {
//			... some code
//		}
// }
//
//	UpSmartLinkPtr<MyClass> someFunc() {
//		UpSmartLinkPtr<MyClass> p = new MyClass;
//		... // some code
//		return p;
//	}
//
///////////////////////////////////////////////////////////////////
// Bad examples (must NOT be done!)
//
// class SomeClass {
// private:
//		MyClass mc;	// ugh!, automatic data (may not be detected by heapcheck)
//		int someInt;
//	public:
//		SomeClass();
//		~SomeClass();
//		UpSmartLinkPtr<MyClass> getMC() {
//			UpSmartLinkPtr<MyClass> p = &mc; // ugh! pointer to automatic data!
//			return p;
//		}
//	};
//		
// int main() {
//		MyClass mc;							// ugh! mc automatic residing in stack
												// (detected if heapcheck is activated)
//		MyClassPtr mcptr = &mc;			// ugh! pointer to automatic data in stack
//		SomeClass* sc=new SomeClass;	// sc resides in heap
//												// sc.mc is automatic data residing in heap
//		mcptr = sc->getMC();				// ugh! pointer to automatic data in heap
//		mcptr = new MyClass[24];		// ugh! pointer to array in heap
//		... // extra code
//		// delete sc;
//	}


/////////////////////////////////////////////////////////////////////////////
class UpLinkable {
private:
	mutable int __count;
	const char *className;
public:
	UpLinkable();
	// _className MUST be a const string literal (see example above)
	UpLinkable(const char* _className); // --> useful for debug
	UpLinkable(const UpLinkable& );
	virtual ~UpLinkable(); // will be inherited: must have a virtual destructor
	void link() const;
	int  unlink() const;
	int  unlinkAndReturnPriorValue() const;
	int  count() const;
	const char* name() const;
};


//////////////////////////////////////////////////////////////////////////////
// UpSmartLinkPtr points to objects derived from class UpLinkable
// residing in data segment, (it must never point to an object in
// the stack segment!)

template<typename T>
class UpSmartLinkPtr {
public:
	// constructores
	UpSmartLinkPtr();
	UpSmartLinkPtr(const T* ptr);
	UpSmartLinkPtr(const UpSmartLinkPtr<T>& sptr);
	// destructor
	~UpSmartLinkPtr();

	// operadores de asignacion
	UpSmartLinkPtr<T>& operator=(const T* ptr);
	UpSmartLinkPtr<T>& operator=(const UpSmartLinkPtr<T>& sptr);

	// C++ considera un error la conversion implicita de  un void* a cualquier
	// otro tipo de puntero.
	// Pero C++ convierte en forma automatica cualquier puntero en un void* .
	// Para que pueda hacer esto con nuestra esta necesitamos el operador
	// de conversion.
	operator	void*();

	// operadores utiles
	T&				operator*();
	const T&		operator*() const;
	T*				operator->();
	const T*		operator->() const;


	// operadores de comparacion
	bool	operator!() const;
	bool	operator==(const T* ptr) const;
	bool	operator!=(const T* ptr) const;
	bool	operator==(const UpSmartLinkPtr<T>& sptr) const;
	bool	operator!=(const UpSmartLinkPtr<T>& sptr) const;
	bool	operator<(const UpSmartLinkPtr<T>& sptr) const;

private:
	T* sp;
	void hold();
	void release();
};


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// UpLinkable global monitor

// There is a global object upLinkableMonitor of the class UpLinkableMonitor.
// Objects derived from UpLinkable report to upLinkableMonitor only if compiled
// with option -DUPLINKABLE_MONITOR

#include <UpTools/UpHashMap.h>
#include <UpTools/UpThread.h>
#include<ostream>

class UpLinkableMonitor {
public:
	UpLinkableMonitor();
	~UpLinkableMonitor();
	// get the UpLinkable count
	unsigned int getCount();                      // all UpLinkables
	unsigned int getUnnamedCount();               // unnamed UpLinkables
	unsigned int getNamedCount();                 // named UpLinkables, all types
	unsigned int getCount(const char* classname); // classname UpLinkables
	// reports of UpLinkable counts to ostream
	void report(std::ostream&);
	void reportUnnamed(std::ostream&);
	void reportNamed(std::ostream&);
	void report(std::ostream&,const char* classname);
private:
	typedef std::map<const char*,unsigned int,UpLessThan>
                  CharPtrUIntMap;
	_Atomic_word   unnamedUpLinkables;
	CharPtrUIntMap namedUpLinkables;
	UpMutex mx;
	friend class UpLinkable;
	// method used by UpLinkable when -DUPLINKABLE_MONITOR
	_Atomic_word constructorEvent();
	_Atomic_word constructorEvent(const char* _classname);
	_Atomic_word destructorEvent(const char* _classname);
};

//////////////////////////////////////////////////////////////////////////////
// UplinkableMonitor global object
extern UpLinkableMonitor upLinkableMonitor;

/////////////////////////////////////////////////////////////////////////////

#endif

#include <UpTools/UpSmartLinkPtr.inl>

