/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Query and connection classes for various DB engines.
// Current supported engines:
//		Mysql
//		Posgresql

#ifndef UPCORE_SQL_H
#define UPCORE_SQL_H

// includes
#include <string>
#include <list>
// forward declarations
class UpSqlQuery;
class UpSqlConn;
class UpSqlConnPool;
class QueryElem;
class AuxThreadConn;
// definitions
#define NOT_CONNECTED "not connected"
#define OK_CONNECTED  "ok"
// types
typedef enum {
	DBTYPE_UNDEF=0,
	DBTYPE_MYSQL,
	DBTYPE_PGSQL,
	DBTYPE_MAX
} upDbType;



//////////////////////////////////////////////////////////////////////////////

class UpSqlConn {
public:
	UpSqlConn(upDbType t=DBTYPE_UNDEF);
	UpSqlConn(const UpSqlConn& c);
	virtual ~UpSqlConn();

	bool copyParams(const UpSqlConn& c);

	bool        status();
	const char* errorPhrase();

	// Next method tries to connect to all configured servers in
	// order until success
	bool connect();
	// Close connection to DB. It is automatically done in destructor.
	virtual void disconnect()=0;

	// If current connection fails, wait retryTime seconds before trying to
	// reconnect or connect to the other servers. May be needed before a
	// backup server synchronizes or comes up.
	void setRetryTime(unsigned int t);

	// set connection parameters (this host will be tried first):
	//    "dbtype://user@host/dbname"
	// or "dbtype://user:passwd@host:port/dbname"
	// or "dbtype://user:passwd@host:port/dbname;param1=val1;param2=val2"
	// dbtype: postgresql or mysql
	bool set(const std::string& sqlUri);
	// set and connect in one pass
	bool connect(const std::string& sqlUri);
	// set connection parameters (this host will be tried first):
	bool set(const std::string& _host, const std::string& _user,
				const std::string& _pass, const std::string& _dbname,
				unsigned short _port , const std::string& _params="");
	// set and connect in one pass
	bool connect(	const std::string& _host, const std::string& _user,
						const std::string& _pass, const std::string& _dbname,
						unsigned short _port , const std::string& _params="");
	// Add a host to the list of possible DB servers
	// When a connection to a specific server fails, all other servers will
	// be tried in order.
	void addHost(const std::string& host,bool pushFront=false);

	// execute query and store results
	// (there may be many results if there are more than one sql
	// sentence in query)
	virtual bool exec(UpSqlQuery& q)=0;
	virtual bool exec(UpSqlQuery& q,const char* sqlText)=0;
	virtual bool exec(UpSqlQuery& q,const char* sqlText,size_t sqlTextLen)=0;

protected:
	upDbType		type; 
	bool			need_close;
	bool        statusOk;
	const char* statusPhrase;
	std::string user;
	std::string pass;
	std::string dbname;
	std::string connParams; // connection params
	unsigned short port;
	std::list<std::string> hosts;
	std::list<std::string>::iterator currentHostIt;
	unsigned int retryTime;
	static const std::string dbTypeStr[];
	void scrPass();
	virtual bool privateConnect(const std::string& host)=0;
	virtual bool errorIsConnectionError(unsigned int err)=0;
	friend class UpSqlConnPool;
	friend class AuxThreadConn;
	virtual UpSqlConn* newCopyConn() const =0;
};


//////////////////////////////////////////////////////////////////////////////

class UpSqlQuery {
public:
	std::string sql;
	UpSqlQuery();
	~UpSqlQuery();
	unsigned long long	lastInsertId(); // not compatible with all engines!!
	unsigned long long	rowCount();
	unsigned int			fieldCount();
	bool						fetchRow();
	void						clear();
	const char*				col(unsigned int c);
	int						colInt(unsigned int c);
	long long 				colLongLong(unsigned int c);
	double					colDouble(unsigned int c);
	bool						colBool(unsigned int c);
	unsigned int			errorNumber();
	const char*				errorPhrase();
	bool						nextResult(); // get next result if there is one
												  // (for multiple statemants)
private:
	QueryElem		*elem; // SQL result
	unsigned int	fCount;
	unsigned int	errNumber;
	char				errPhrase[256];
	static QueryElem emptyQueryElem;

	void clearResultList();
	void addQueryElem(QueryElem* e);
	void setError( unsigned int _errorNumber, const char* _errorPhrase );
	friend class UpSqlConn;
	friend class UpSqlConnMysql;
	friend class UpSqlConnPgsql;
};

//////////////////////////////////////////////////////////////////////////////

#include <UpTools/UpSql.inl>

#endif 

/*
EJEMPLO DE USO:
	UpSqlConnMysql db; // UpSqlConnMysql inherits from UpSqlConn
	db.set("mysql://dbmaster:masterpass67@dbserver.data.com:789/sess_db");
	db.addHost("dbbackup.data.com");
	db.connect();
	if( !db.status() ) raiseError();

	UpSqlQuery r;
	r.sql << "SELECT id,nombre FROM test WHERE id>" << 9 << " ORDER BY id";
	if(db.exec(r)) process(r);
	else queryFailure();
*/
