/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// This class implements a pool of db connections, one for each thread
// so each thread can access the DB engine without blocking the
// other threads (as is the case with SqlConnPool)

// Each thread can ask for itr own connection or it can use the pool methods:
// Example 1:
//		UpSqlConn& myPrivateConn(sqlPool.getConnectedConn());
//		...
//		UpSqlQuery q; q << "SELECT * FROM " << table;
//		if( myPrivateConn.exec(q) ) {
//			/* process q here */
//		}
//
// Example 2 (better):
//		UpSqlQuery q; q << "SELECT * FROM " << table;
//		if( sqlPool.exec(q) ) { // this use tthread's own conn
//			/* process q here */
//		}

#ifndef UPCORE_SQLCONNPOOL_H_
#define UPCORE_SQLCONNPOOL_H_

#include <UpTools/UpThread.h>
#include <UpTools/UpSql.h>
class UpSqlQuery;
class AuxThreadConn;


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class UpSqlConnPool {
   public:
		// Connection Model must be set to a class derived from
		//   UpSqlConn in order to set db type (DBTYPE_MYSQL,DBTYPE_PGSQL).
		//   If connection model has hosts and DB parameters set, these will be
		//   set for all connections, in addition to any db parameters set from
		//   this class).
      UpSqlConnPool();
      UpSqlConnPool(const UpSqlConn& modelConn);
		void setConnectionModel(const UpSqlConn& modelConn);
      ~UpSqlConnPool();

		UpSqlConn&		getConn();			 // get thread's own UpSqlConn
		UpSqlConn&		getConnectedConn();// get thread's own connected UpSqlConn
		unsigned int	getConnCount();	 // get number of thread UpSqlConn's	

		bool status();
		const char* errorPhrase();

		// Next method tries to connect to all configured servers in
		// order until success
		bool connect();
		// Close connection to DB. It is automatically done in destructor.
		// (closes only this thread connection, not all of them)
		void disconnect();

		// set connection parameters (this host will be tried first):
		//    "dbtype://user@host/dbname"
		// or "dbtype://user:passwd@host:port/dbname"
		// or "dbtype://user:passwd@host:port/dbname;param1=val1;param2=val2"
		// dbtype: postgresql or mysql
		bool set(const std::string& sqlUri);
		// set and connect in one pass
		bool connect(const std::string& sqlUri);
		// set connection parameters (this host will be tried first):
		bool set(const std::string& _host, const std::string& _user,
					const std::string& _pass, const std::string& _dbname,
					unsigned short _port , const std::string& _params="");
		// set and connect in one pass
		bool connect(	const std::string& _host, const std::string& _user,
							const std::string& _pass, const std::string& _dbname,
							unsigned short _port , const std::string& _params="");
		// Add a host to the list of possible DB servers
		// When a connection to a specific server fails, all other servers will
		// be tried in order.
		bool addHost(const std::string& host,bool pushFront=false);

		bool exec(UpSqlQuery& q);
		bool exec(UpSqlQuery& q,const char* sqlText);
		bool exec(UpSqlQuery& q,const char* sqlText,int sqlTextLen);
   private:
	  UpMutex			mx;
     AuxThreadConn	*firstConn;
	  unsigned int		connCount;
     UpSqlConn*		basePtr; // only to hold param data, will not connect
     void distribParamsUnsafe();
};

#include <UpTools/UpSqlConnPool.inl>

#endif 

