/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Wrapper around openssl or yassl (compatible openssl api for ssl)
//   use yassl if a GPL-license or a commercial-license is needed
//   use openssl if an openssl-license is needed

// Openssl and yassl need library initialization in a way,
// that make us separate ssl related classes in separate library
// to be statically linked only

#ifndef UPSSL_H
#define UPSSL_H

#ifndef USE_YASSL
	#include <openssl/ssl.h>
	#include <openssl/x509.h>
	#include <openssl/rsa.h>
	#include <openssl/engine.h>
   
#else
	#include <yassl/ssl.h>
	#include <yassl/x509.h>
	#include <yassl/rsa.h>
	#include <yassl/engine.h>
#endif

#include <UpTools/UpInet.h>
#include <string>

#ifdef UP_NONBLOCKING_RANDOM
	#define RND_SOURCE "/dev/urandom" // does not block waiting for entropy
#else
	#define RND_SOURCE "/dev/random" // blocks waiting for entropy
#endif

   
#define UP_KEY_ENC EVP_aes_256_cfb()

// to debug lib callbacks define UPSSL_CALLBACK_DEBUG before compiling library

// forward declarations
class UpCryptoSslLib;
class UpCryptoSslLibMultithreading;
class UpSslContext;
class UpX509;
class UpRsa;
class UpSsl;

extern const std::string ssl_empty_string;

//////////////////////////////////////////////////////////////////////////////
// Classes for library initialization.
// In order to use any of the other classes in this header file, you need to
// one of the following two classes. Only one of these can be instantiated.
//
// If used in a multithreaded application UpCryptoSslLibMultithreading must
// me instantiated, and linking with -lcrypto -lssl -lpthread is needed.
//
// If used in a signle threaded application UpCryptoSslLib must be
// instantiated, and linking with -lcrypto -lssl is needed


class UpCryptoSslLib {
public:
	UpCryptoSslLib();
	~UpCryptoSslLib();
};


class UpCryptoSslLibMultithreading : public UpCryptoSslLib {
public:
	UpCryptoSslLibMultithreading();
	~UpCryptoSslLibMultithreading();
};


//////////////////////////////////////////////////////////////////////////////
// x509 certificate class

class UpX509 {
public:
	UpX509();
	~UpX509();
	const std::string& getStatus();
	// We could read or write encrypted certs, but it makes no sense.
	// For the moment these are onle PEM certs, not DER (see man d2i_X509).
	bool readCert(BIO* bk);									// read from BIO
	bool readCert(const char* pemKeyBuf,size_t pemKeyLen);	// read from buffer
	bool readCert(const std::string& pemKey);				// read from string
	bool readCertFile(const std::string& filename);		 	// read from file
	bool saveCert(BIO* bk);									// save to BIO
	bool saveCert(std::string& pemKey);						// save to string
	bool saveCertFile(const std::string& filename); 		// save to file
	bool showCert(BIO* bk);									// export to BIO
	bool showCert(std::string& text);						// export to string
	bool showCertFile(const std::string& textCertFilename); // export to file
	std::string getSubject(); // use this to check coincidence with domain name
	std::string getIssuer();
	bool checkTimeValidity(); // check if today  falls within validity period
private:
	X509* x;
	std::string status;
	void freeX();
	void statusPadError();
	friend class UpSsl;
	friend class UpSslContext;
};
//////////////////////////////////////////////////////////////////////////////
// rsa keys and algorithm class
   
class UpRsa {
public: 
   UpRsa();
   ~UpRsa();
   const std::string& getStatus();
   bool havePrivKey();
   // Autogenerar un par de claves publica-privada.
   // ver "man RSA_generate_key":
   // The modulus size will be num bits, and the public exponent will be e.
   // Key sizes with num < 1024 should be considered insecure.  The exponent
   // is an odd number, typically 3, 17 or 65537.
   // Comment: esta funcion de generacion no es segura
   bool generateKey(int bits=2048,unsigned long exponent=65537);
   // Alternativas para leer la clave publica o privada:
   // El passphrase puede ser nulo, en ese caso se le pregunta al
   // usuario por stdin/stdout si es que hace falta.
   // El archivo o buffer puede contener la clave publica o el par
   // publico-privado. En caso de contener ambas claves,los metodos
   // encrypt(...) y decrypt(...) pueden optar por usar una u otra clave.
   bool readKey(BIO* bk,                                // read from BIO
               const std::string& passphrase=ssl_empty_string);
   bool readKey(const char* pemKeyBuf,size_t pemKeyLen, // read from buffer
               const std::string& passphrase=ssl_empty_string);
   bool readKey(const std::string& pemKey,              // read from string
               const std::string& passphrase=ssl_empty_string);
   bool readKeyFile(const std::string& filename,        // read from file
               const std::string& passphrase=ssl_empty_string);
   // Alternativas para guardar la clave publica y/o privada
   // Si se tiene en memoria el par privado-publico se puede optar por
   // guardar ambas o solo la publica. Si se tiene la publica solamente,
   // se ignoran los parametros savePublicOnly, encryptKey y passphrase.
   bool saveKey(BIO* bk,                          // save to BIO
                bool savePublicOnly=false, bool encryptKey=false,
                const std::string& passphrase=ssl_empty_string);
   bool saveKey(std::string& pemKey,              // save to string
                bool savePublicOnly=false, bool encryptKey=false,
                const std::string& passphrase=ssl_empty_string);
   bool saveKeyFile(const std::string& filename,  // save to file
                bool savePublicOnly=false, bool encryptKey=false,
                const std::string& passphrase=ssl_empty_string);
   // Alternativas para encriptar:
   // La primera devuelve un string encriptado, o vacio en caso de error
   // La segunda devuelve la longitud del cipher recientemente calculado o -1
   // en caso de error.
   std::string privEncrypt(const std::string& cleartext);
   std::string pubEncrypt(const std::string& cleartext);
   int         privEncrypt(const std::string& cleartext, std::string& cipher);
   int         pubEncrypt(const std::string& cleartext, std::string& cipher);
   // Alternativas para cifrar:
   // La primera devuelve un string descifrado.
   // la segunda devuelve la longitud del texto recientemente descifrado o -1
   // en caso de error.
   std::string privDecrypt(const std::string& cipher);
   std::string pubDecrypt(const std::string& cipher);
   int         privDecrypt(const std::string& cipher, std::string& cleartext);
   int         pubDecrypt(const std::string& cipher, std::string& cleartext);
private:
   RSA* x;
   bool privKey;
   std::string status;
   void freeX();
   void statusPadError();
   friend class UpSsl;
   friend class UpSslContext;
};


//////////////////////////////////////////////////////////////////////////////
// Ssl context
// It is used to reuse certificates and sessions between successive ssl connections

typedef enum {
	UPCTXMT_ALL,
	UPCTXMT_CLIENTONLY,
	UPCTXMT_SERVERONLY
} upCtxMethodtype;


class UpSslContext {
public:
	UpSslContext(upCtxMethodtype mt=UPCTXMT_ALL);
	~UpSslContext();
	int getErr();
	const char* getErrBuf();
	// Set own certificate and key to be used on authentication for all
	// connections to be created using this context. It is also possible to
	// set each connection individually (see UpSsl::useCertAndRsaPrivateKey)
	// IMPORTANT: this method must be called after calling any of the
	//            addCAChainXXX methods, because they also change
	//            the default certificate beside setting the accepted CA's)
	bool useCertAndRsaPrivateKey(UpX509& x509, UpRsa& rsa);
	// Add one certificate to the CA (certificate authorities) chain
	bool addCAChainCert(UpX509& x509);
	// Add a PEM file containing many certificates authorities to the CA chain
	bool addCAChainFile(const char* CAChainFile);
	// Add a PEM file containing many certificates authorities,
	//   or add a directory containing hashedNames CA files
	//   (see man SSL_CTX_load_verify_locations or cacertdir_rehash)
	bool addCAChainFilePath(const char* CAChainFile,
							const char* hashedCANamesDir);
	
private:
	SSL_CTX* ctx;
	char errBuf[128]; // library needs only 120
	int sslErr;
	void setErr(const char* errTxt);
	bool checkErr(int sslRet);
	friend class UpSsl;
};




//////////////////////////////////////////////////////////////////////////////
// ssl connection class
class UpSsl {
public:
	UpSsl(UpSslContext& ctx);
	~UpSsl();
	bool		getLastOpStatus();
	int         getErr();
	const char* getErrBuf();
	// inidicar al socket a utilizar (UpSocket o comun)
	bool setChannel(UpSocket& ups);
	bool setChannel(int fileDescriptor);
	// indica si esperar o no el saludo del otro lado al final
	void setBidirectionalShutdown(bool); // default: no se espera
	// these can be set for all connections using a context or for each ssl
	// connection individually
	bool useCertAndRsaPrivateKey(UpX509& x509, UpRsa& rsa);
	// initiate TLS/SSL handshake with a server
	int connect();
	// initiate TLS/SSL handshake with a client
	int accept();
	// El siguiente metodo sirve para hacer el salido final protocolar
	// return values: 1(ok), 0(falta recibir saludo del otro lado), -1(error).
	// Para que devuelva 1(ok) hay que llamar a la funcion 2 veces o
	// setear setbidirectionalShutdown. La segunda vez (o la primera
	// en caso de setear bidirectionalShutdown) se bloquea hasta que recibe
	// el close notify. Terminar correctamente el shutdown puede ser
	// necesario para reutilizar una sesion. Ver man SSL_shutdown. 
	int shutdown();
	// el siguiente metodo cierra el socket y saluda si es que hace falta
	void close();
	// closes and reinits object, erase all data and setup (including socket)
	int reInit();
	// read and write
	int write(const char* buf, size_t len);
	int read(char* buf, size_t len);
	// number of already buffered bytes ready to be read by calling read(...)
	int readPending();
	// This class is set to automatically renegotiate, but if it is desired
	// to force a renegotiation:
	int forceRenegotiation();
	
private:
	int			fd;
	UpSocket*	upsock;
	SSL*		ssl;
	bool		bidirectionalShutdown;
	char		errBuf[128];
	int 		sslRet;
	int 		sslErr;
	void ctxErr();
	bool checkErr( int sslRet );

};

#include <UpTools/UpSsl.inl>

#endif
