/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// classes in this file: UpThread, UpMutex, UpSemaphore, UpScopedLock,UpRWLock

#ifndef UPTHREAD_H
#define UPTHREAD_H

#include <signal.h>
#include <pthread.h>
#include <semaphore.h>

///////////////////////////////////////////////////////////////////////////////
// class UpThread
///////////////////////////////////////////////////////////////////////////////


class UpThread {
public:
	/////////////////////////////////////////////////////////////////////////
	// Constructor and destructor
	UpThread();
	virtual ~UpThread(); // destructor must be virtual if there is at
								// least one virtual method
	/////////////////////////////////////////////////////////////////////////
	// methods for C++ classes derived from UpThread
	virtual void* run();	// should contain the running code for the thread
								// this method should be replaced in derived classes
	bool create();				// thread creation
	/////////////////////////////////////////////////////////////////////////
	// methods to use when starting a C-like function as a thread instead
	// of using derived classes
	// We recommend using derived classes instead of this approach
	bool create(void* (*threadFunc)(void*), void* arg);
	// the function void* threadFunc(void*) must contain the running code
	// for the thread and can not use testCancel(), setCancelEnable(),
	// setCancelDisable(), setCancelTypeAsync(), setCancelTypeDeferred(),
	// but must call instead: pthread_exit(void*), pthread_testcancel(),
	// pthread_setcancelstate(int,int*), pthread_setcanceltype(int,int*)
	/////////////////////////////////////////////////////////////////////////
	// operations on running threads
	bool kill(int signum); // send signal to thread
	bool detach();	// thread will cleanup whithout waiting a join
	void *join();	// wait for thread calling UpThreadExit(void*)
	bool cancel();	// thread will exit at next cancellation point
	// cancellation points are: pthread_join,  pthread_cond_wait,
	// 	pthread_cond_timedwait,  pthread_testcancel,  sem_wait,  sigwait.
	/////////////////////////////////////////////////////////////////////////
	// check if a thread is running() (this method works only if the default
	// action for SIGCONT is IGNORE, i.e. the default action)
	bool isRunning();
	// get thread id (valid only if thread has already been started
	pthread_t getThreadId();
	// checks if the thread who executes this method is the
	// same thread that was started by this object
	bool checkSelf();
	/////////////////////////////////////////////////////////////////////////
	// Sheduling policy and priority on running threads
	bool setSchedOther();
	bool testSchedOther();
	bool setSchedRoundRobin(int priority);
	bool testSchedRoundRobin(int& priority);
	bool setSchedFifo(int priority);
	bool testSchedFifo(int& priority);
	/////////////////////////////////////////////////////////////////////////
	// Attribute setting and testing before running threads
	bool setAttrDetached();	// detach trhead when create
	bool testAttrDetached();

	bool setAttrJoinable(); // do not detach thread when create
	bool testAttrJoinable();

	bool setAttrSchedOther();	// regular schedule when create 
	bool testAttrSchedOther();

	bool setAttrSchedRoundRobin();	// round robin schedule when create
	bool testAttrSchedRoundRobin();	//   only when uid=root

	bool setAttrSchedFifo();		// fifo schedule when create
	bool testAttrSchedFifo();		//   only when uid=root

	bool setAttrExplicitSched();	// threads created from here will not
	bool testAttrExplicitSched();	//   scheduling policy

	bool setAttrInheritSched();		// threads created from here will
	bool testAttrInheritSched();	//   scheduling policy

	bool setAttrScopeSystem();		// scheduling respect to all system
	bool testAttrScopeSystem();		//   processes

	bool setAttrScopeProcess();		// scheduling respect to all this
	bool testAttrScopeProcess();	//   process's theads

	bool setAttrSchedPriority(int priority); // set schedule priority
	int  getAttrSchedPriority();	//   only for round robin or fifo schedule

private:
	pthread_attr_t	attr;
	pthread_t thread;
	bool started;
protected:
	/////////////////////////////////////////////////////////////////////////
	// methods callable only from methods in the derived classes 
	void exit(void* retval);	// exits immediately if detached,
								// waits for join if not detached
	void testCancel(); 			// test for pending cancel requests
	bool setCancelEnable();		// enable honoring of cancel requests
	bool setCancelDisable();	// ignore cancel requests
	bool setCancelTypeAsync();	// set immediate honoring of cancel requests
	bool setCancelTypeDeferred(); // set deferred honoring of cancel requests

};


/* cleanup handlers used in void* run() method defined in derived class:
	ROUTINE will be called with arguments ARG when the thread is cancelled
	or calls UpThreadExit.  ROUTINE will also be called with arguments ARG
	when the matching UpThreadcleanup_pop is executed with non-zero EXECUTE
	argument.
   UpThreadCleanupPush and UpThreadCleanupPop are macros and must always
   be used in matching pairs at the same nesting level of braces.
*/
#define UpThreadCleanupPush(ROUTINE,ARG) pthread_cleanup_push(ROUTINE,ARG)
#define UpThreadCleanupPop(EXECUTE)      pthread_cleanup_pop(EXECUTE)



///////////////////////////////////////////////////////////////////////////////
// class UpMutex
///////////////////////////////////////////////////////////////////////////////


class UpMutex {
public:
	UpMutex();
	~UpMutex();
	UpMutex(int kind); // non portable constructor (should only be used in Linux)
						// kind can be PTHREAD_MUTEX_FAST_NP,
						// PTHREAD_MUTEX_RECURSIVE_NP or PTHREAD_MUTEX_ERRORCHECK_NP
	int lock();
	int unlock();
	int trylock();
private:
	pthread_mutex_t mutex;

};

///////////////////////////////////////////////////////////////////////////////
// class UpScopedLock
///////////////////////////////////////////////////////////////////////////////
// This class is used to lock an UpMutex inside a scope and be
// sure that it will be unlocked when the scope is left. For example
// from a method with several return points.

class UpScopedLock {
public:
	UpScopedLock(UpMutex& _mx);
	~UpScopedLock();
private:
	UpMutex& mx;

};
/////////////////////////////////////////////////////////////////////////////
// class UpSemaphore
///////////////////////////////////////////////////////////////////////////////

class UpSemaphore {
public:
	// Creates semaphore, initializing its value
	UpSemaphore(unsigned int initialValue);
	// destroys semaphore
	~UpSemaphore();

	// Atomically tests if semaphore is greater than 0 and then decrements  its
	// value. If it is zero, it waits until it is raised by other thread that
	// calls post()
	int wait();
	// Raises the semaphore value by one.
	int post();
	// Returns semaphore value.
	int getValue();
	
private:
	sem_t s;
};

///////////////////////////////////////////////////////////////////////////////
// class UpRWLock
///////////////////////////////////////////////////////////////////////////////


class UpRWLock {
public:
	UpRWLock();
	~UpRWLock();
	int    RDLock();
	int tryRDLock();
	int    WRLock();
	int tryWRLock();
	int unlock();
private:
	pthread_rwlock_t rwl;

};

/****************************************************************************
// pthread functions that must be used raw

// handlers management
pthread_atfork - register handlers to be called at fork(2) time
pthread_cleanup_pop_restore_np [pthread_cleanup_push] - install and remove cleanup handlers
pthread_cleanup_push_defer_np [pthread_cleanup_push] - install and remove cleanup handlers

// miscellaneous
pthread_self - return identifier of current thread
pthread_equal - compare two thread identifiers
pthread_kill_other_threads_np - terminate all threads in program except calling thread
pthread_once - once-only initialization

pthread_sigmask - handling of signals in threads


*****************************************************************************/

#include <UpTools/UpThread.inl>

//////////////////////////////////////////////////////////////////////////////
// log thread id definitions
//////////////////////////////////////////////////////////////////////////////

#ifdef upOpenLogFile
	#undef upOpenLogFile
#endif

#define upOpenLogFile(ident,options,filename,logLevel) \
				upOpenLogFileInternal(ident,options,filename,logLevel,upLogHdrThreads)

extern "C" {
int upLogHdrThreads(char* buf);
}

#endif
