/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef UPEVENTTHREAD_H
#define UPEVENTTHREAD_H

#include <UpTools/UpCondition.h>
#include <list>
#include <iostream> // debug this!

class UpWork;
class UpWorkSplittable;
class UpThreadWorkDistributor;
class UpWorkThread;

//////////////////////////////////////////////////////////////////////////////

// This class must be derived to process works specific to the application.
// Objects of this class are deleted by the UpThreadWorkDistributor so they
// must point to data saved elsewhere.
class UpWork {
	protected:
		UpTimeVal startTime; // may be accessed by derived classes
		UpTimeVal duration;
		char type; // w:work, s:splittable, b:splittable base
		friend class UpThreadWorkDistributor;
	private:
		void init();
		void end();
		// this method must be derived by the application develper in order to do
		// any useful work
		virtual void action();
		friend class UpWorkThread;
	public:
		UpWork();
		virtual ~UpWork();
		const UpTimeVal& getDuration();
		const UpTimeVal& getStartTime();
};

//////////////////////////////////////////////////////////////////////////////
// This class must be derived to process algorithms specific to the application
// UpWorkDistributor splits the UpWorkSplittable as many times as is convenient
// depending on the number of hardware threads or the nature of the work.
// The calling thread can tell UpThreadWorkDistributor to wait for the completion
// of the work if it needs to.
class UpWorkSplittable : public UpWork {
	private:
		typedef enum {
			WS_NOTFINISHED=0,
			WS_OWNFINISHED=1,  // termino el trabajo propio
			WS_JOINEDCHILDS=2, // idem anterior y ademas joineo todos sus childs
			WS_DONE=3,         // idem anterior y ademas fue joineado
			WS_DONEBASE=4
		} workSplittableState;

		static std::string stateStr[];
		UpWorkSplittable* prevWork; // el work que hara join a este
		UpWorkSplittable* nextWork; // el work a joinear (excepto si es arbol)
		UpWorkSplittable* baseWork; // el work base
		pthread_t waitingThread; // se manda senial de terminacion a un thread?
		workSplittableState state;
		UpMutex* mx; // to provide mutual exclusion between subworks (owned)
		UpSemaphore* sp; // to wakeup base work after all has ended (not owned)
		bool baseEndWork(UpWorkSplittable* w);
		friend class UpWorkThread;
		friend class UpThreadWorkDistributor;
	private: // methods to be derived by users of this class
		// returns zero if not splittable
		// returns a new instance charged with a portion of its own
		//   work if splittable
		virtual UpWorkSplittable* split();
		// this method consolidates the job of a previously splitted work
		virtual void join(UpWorkSplittable* workToJoin);
		// this method indicates 3 possible join types: serial(s), tree(t),
		//   nojoin(n)
		// these types are best explained in an example:
		//  (the tree represents how each work is splitted)
		//  a
		//  a               a0              (a is splitted in two)
		//  a      a1       a0     a00      (a and a0 are splitted in two)
		//  a  a2  a1  a10  a0 a01 a00 a000 (a, a1, a0, a00 are splitted in two)
		//  depending on the splitType, the join process is:
		//  nojoin: no need to explain
		//  serial: a00.join(a000), a01.join(a00), a0.join(a01), a10.join(a0),
		//     a1,join(10) , a2.join(a1), a.join(a2)
		//  tree: a.join(a2), a1.join(a10), a0.join(a01), a00.join(a000),
		//  	 a.join(a1), a0.join(a00), a,join(a0)		
		virtual char joinType();
		// Some works are optimized when splitted in a number of works that equal
		// the number of hardware threads. These have generally non-blocking
		// algorithms.
		// Other works are optimized even when splitted in a number
		// jobs greater than the number of threads. These are generally works
		// that do blocking operations like I/O. UpThreadWorkDistributor keeps
		// spliting them until they return 0 the the method split() or until
		// it surpasses its max thread count.
		virtual bool exceedHwThreads();
public:
		UpWorkSplittable();
		virtual ~UpWorkSplittable();
		virtual void show(std::ostream& o ) const {};
		friend std::ostream& operator<<(std::ostream& o, const UpWorkSplittable& w)
			{ w.show(o); return o;}
};

//////////////////////////////////////////////////////////////////////////////
// this class maintains a pool of threads and processes works as soon as there
// is an available thread
class UpThreadWorkDistributor {
public:
	// constructor, set the number of work execution threads
	// In linux it finds the number of hardware threads automatically
	// in ohter OS's 
	UpThreadWorkDistributor(size_t _threadCount);
	~UpThreadWorkDistributor();

	// Set the memory for rate and time calcluations. It means:
	// After a number n of works, all of them separated with an interval dt,
	// the estimated interval will be (0.9*dt).
	void setWorkMemoryLength(unsigned int n);

	// get work statistics
	float getWorkExecutionAvgTime();    // work "average" execution time 
	float getWorkArrivalRate();			// rate of arrival of new works
	// Rate of arrival of new works, but also takes into account the time
	// elapsed since last work arrival
	float getWorkArrivalRateUpdated();

	// When the expected wait time for a work exeeds this, then processWork(...)
	// will reject the work.
	// Set to negative value if processWork(...) should reject works when the
	// number of works in the queue exceeds maxQueueLength.
	// Set to a finite positive number to set the maximum "expected" wait time.
	// It defaults to 0.5 seconds
	void setMaxQueueWait(float seconds);

	// Set the maximum number of works allowed to wait in the queue at one time.
	// maxQueueLength is used only if maxQueueWait is negative.
	// It defaults to 10*threadCount
	void setMaxQueueLength(size_t maxLength);

	// Next method adds a new work to be processed as soon as there are
	// available threads. This method maybe called from any thread in the
	// application. Returns true if work is immediately executed or queued,
	// and returns false if work queue is too long to accept new works.
	// If the work is accepted, it will be deleted after completion
	// by UpThreadWorkDistributor, so it must save internal data elsewhere.
	// The work is splitted as many times as is convenient depending on the
	// number of threads, or the nature of the work.
	// If waitCompletion is true the method will return only after the work
	// is finished, if it is false it will return immediately (while the work
	// continues to be precessed in other threads).
	bool processWork( UpWork* e, bool waitCompletion=false );

		
private:
	std::list<UpWorkThread*> threadList;
	size_t                   threadCount;
	size_t						 hwThreadCount;

	std::list<UpWork*> workQueue;
	UpCondition<>		 condition; // condition.var will hold workQueue.size()
	UpCondition<>		 finishCondition; // condition used to synchronize destructor
	size_t				 baseWorkQueueSize; // base works in the queue
	float					 maxQueueWait; // in seconds
	float					 waitCondAux;
	size_t				 maxQueueLength; // to be used only when maxQueueWait<0

	size_t worksProcessing; // works procesandose actualmente

	static time_t maxFinishWait;

	// estimacion de rates y tiempos
	UpTimeVal lastArrival;
	float workArrivalRate; // teoria de colas: lambda
	float	workProcessRate; // teoria de colas: mu
	float	gamma,oneMinusGamma; // parametro para la estimacion de rates
	// M/M/1: una cola servida por un server (thread)
	//  probabilidad de estar ocupado:           1-p0 = rho = lambda / mu
	//  numero promedio de works en el sistema:   N = lambda / ( mu - lambda )
	//  timepo promedio de un work en el sistma:  T =    1   / ( mu - lambda )
	// M/M/n: una cola servida por multiples servers (threads)
	//  rho = lambda / ( n * mu )
	//  los otros datos son mas dificiles de calcular en este caso

	bool checkQueueWaitUnsafe();
	void accountArrivalUnsafe();
	void accountWorkDurationUnsafe(float dt);
	void accountWorkEnd(UpWork* e);
	UpWork* getConditionedWork();
	void endOfThread();
	friend class UpWorkThread;
		
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


#define MAXQUEUEWAIT_DEFAULT 0.5      /* 0.5 seconds */
#define DEFAULT_GAMMA        0.79433 /* memory detph = 10 */

#include <UpTools/UpThreadWorkDistributor.inl>

#endif
