/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TIMEVAL_HPP
#define TIMEVAL_HPP

#include <sys/time.h>
#include <cmath>
#include <ostream>
#include <cstring>
#include <UpTools/UpOp.h>

extern const char* UPTVdateFormat;
extern const char* UPTVtimeFormat;
extern const char* UPTVdateTimeFormat;

class UpTimeVal {
public:
	time_t		tv_sec;
	suseconds_t	tv_usec;

	UpTimeVal();	// inicilaiza en cero
	UpTimeVal(double d);
	UpTimeVal(time_t,time_t);
	UpTimeVal(const UpTimeVal&);
	UpTimeVal(const timeval&);
	~UpTimeVal() {};	// en este destructor no hacemos nada

	void clear();	// hace tv_sec y tv_usec iguales a cero
	void now();		// carga la hora actual en tv_sec y tv_usec

	const UpTimeVal& tic();	// carga la hora actual en tv_sec y tv_usec
	const UpTimeVal& toc();	// carga la diferencia entre la hora actual
									// y la que estaba cargada previamente

	bool isset() const;	// devuelve true si tv_sec o tv_usec difieren de cero

	// The following methods use the formats defined in global variables:
	//   UPTVdateFormat="%d-%b-%Y", UPTVtimeFormat="%H:%M:%S", and
	//   UPTVdateTimeFormat="%d-%b-%Y %H:%M:%S"
	// (these are char pointers, and may be modified in order to point
	// to different formats, be careful only if linking to the shared version
	// of the library, do not be careful if linking to the static version)
	// These methods use the time zone set by the system.
	
	// copies in buf the date, returns bytes written (not including '\0')
	size_t charDate(char* buf,size_t bufsize) const;
	// copies in buf the time of day, returns bytes written (not including '\0')
	size_t charTime(char* buf,size_t bufsize) const;
	// copies in buf the date and time, rets bytes written (not including '\0')
	size_t charDateTime(char* buf,size_t bufsize) const;
	// same as before but writes into string
	void charDate(std::string& s) const;
	void charTime(std::string& s) const;
	void charDateTime(std::string& s) const;
	// set internal data from human readable date and time
	bool readDate(const char* buf);
	bool readDateTime(const char* buf);
	bool readDate(const std::string& s);
	bool readDateTime(const std::string& s);

	UpTimeVal operator+(const UpTimeVal&) const;
	UpTimeVal operator+(time_t) const;
	UpTimeVal operator+(const timeval&) const;
	UpTimeVal operator+(double) const;

	UpTimeVal operator-(const UpTimeVal&) const;
	UpTimeVal operator-(time_t) const;
	UpTimeVal operator-(const timeval&) const;
	UpTimeVal operator-(double) const;

	double operator/(const UpTimeVal&) const;
	double operator/(const timeval&) const;
	double operator/(time_t) const;
	double operator/(double) const;

	bool operator>(const UpTimeVal&) const;
	bool operator>(time_t) const;
	bool operator>(const timeval&) const;
	bool operator>(double) const;

	bool operator<(const UpTimeVal&) const;
	bool operator<(time_t) const;
	bool operator<(const timeval&) const;
	bool operator<(double) const;

	bool operator!=(const UpTimeVal&) const;
	bool operator!=(time_t) const;
	bool operator!=(const timeval&) const;
	bool operator!=(double) const;

	bool operator==(const UpTimeVal&) const;
	bool operator==(time_t) const;
	bool operator==(const timeval&) const;
	bool operator==(double) const;


	UpTimeVal& operator+=(const UpTimeVal&);
	UpTimeVal& operator+=(const timeval&);
	UpTimeVal& operator+=(time_t);
	UpTimeVal& operator+=(double);

	UpTimeVal& operator-=(const UpTimeVal&);
	UpTimeVal& operator-=(const timeval&);
	UpTimeVal& operator-=(time_t);
	UpTimeVal& operator-=(double);

	UpTimeVal& operator*=(unsigned int); // faster then operator*=(double)
	UpTimeVal& operator*=(unsigned long long); // faster then operator*=(double)
	UpTimeVal& operator*=(double);

	UpTimeVal& operator=(time_t);
	UpTimeVal& operator=(const timeval&);
	UpTimeVal& operator=(const UpTimeVal&);
	UpTimeVal& operator=(double);
	UpTimeVal& operator=(const std::string& s);
	UpTimeVal& operator=(const char* p);

	void	 		copyTo(timeval& tv) const; //copies itself to tv

	timeval 		getStruct() const;
	double  		getDouble() const;
	suseconds_t	getTotalMicroSeconds();	

	friend std::ostream& operator<<(std::ostream& o,const UpTimeVal& t);
	friend std::string&  operator<<(std::string& s, const UpTimeVal& t);

private:
	static void basicCout(std::ostream& o,const UpTimeVal& t);
	static void basicSout(std::string& s, const UpTimeVal& t);
};

///////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// global operators
inline
time_t& operator<<(time_t& t, const std::string& s) {
	#if __WORDSIZE==64
	*((long long*)(&t)) << s; return t;
	#else
	*((int*)(&t)) << s; return t;
	#endif
}

inline
std::string& operator<<(std::string& s,time_t t) {
	#if __WORDSIZE==64
	s << (long long)t; return s;
	#else
	s << (int)t; return s;
	#endif
}

#include <UpTools/UpTimeVal.inl>

#endif
