/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Class UpTimer

// Class UpTimer is used to schedule the execution of code at specific
// moments, with or without repetition. The desired code must be included in a
// class derived from UpTimer, in the method bool DerivedClass::action(_now).
// All UpTimers must be created with operator 'new', and must be added to an
// UpTimerManager in order to be executed at its determined time(s). After
// executing an UpTimer, UpTimerManager handles its deletion.
// Example:
//
// class MyTimer : public UpTimer {
// public:
//	MyTimer(const UpTimeVal& _interval,unsigned int _count)
//		: UpTimer(0,_interval,_count,"MyTimer")
//		{}
//	bool action(const UpTimeVal& _now) {
//		/* specific timer code goes here */
//		return true;
//	}
// };
//
//
// UpTimerManager tm;
// tm.startThread();
//
// MyTimer q = new MyTimer(2.5,10);
// tm.add(q);

#ifndef UPTIMER_H
#define UPTIMER_H
#include <UpTools/UpLog.h>
#include <UpTools/UpTimeVal.h>


class UpTimerManager;

class UpTimer {
public:
	friend class UpTimerManager;
	// Constructor one shot.
	UpTimer(const UpTimeVal& _when);
	// Constructor with repetition. Timer will be executed at time '_when',
	// and will be rescheduled for execution after an '_interval', only
	// if action(_when) returns true. This will be reapeated up to a maximum
	// '_repeatCount' noumber of executions.
	// Note: if '_when' is 0 then the first execution is scheduled at
	//       now()+'_interval'
	UpTimer(const UpTimeVal& _when,const UpTimeVal& _interval,
			unsigned int _repeatCount);
	// If there is a need to inquire about the number of pending UpTimers of
	// each type, then use the following constructors instead.
	// _className MUST be a const string literal (see example above)
	UpTimer(const UpTimeVal& _when,const char* _className);
	UpTimer(const UpTimeVal& _when, const UpTimeVal& _interval,
			unsigned int _repeatCount,const char* _className);
	// Destructor
	virtual ~UpTimer();
	// Next method should be replaced in classes derived from this class.
	virtual bool action(const UpTimeVal& _now);
		// _now is included as a parameter to action(...) because it may be
		// executed slightly after _when
	const char* name();
	// Operators (UpTimer A is less than UpTimer B if scheduled time for A is
	// less than the scheduled time for B, in case both times are equal, then
	// their pointers are compared)
	bool operator<(const UpTimer& t) const;
	bool operator<(const UpTimeVal& t) const;
	bool operator>(const UpTimeVal& t) const;
protected:
	UpTimeVal when;
	UpTimeVal interval;
	unsigned int count;
	// to be used inside the method action(...)
	void repeat(const UpTimeVal& _interval, unsigned int _repeatCount);
private:
	bool adminAction(const UpTimeVal& _now);
	const char* className;
};

#include <UpTools/UpTimer.inl>

#endif
