/* UpTools v8.6
 *
 * Copyright (c) 2005-2012 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Class UpTimerManager

// Class UpTimerManager manages the timely execution of UpTimers.
// UpTimers are run on the UpTimerManager own thread, outside of any
// locks (UpTimerManager's only lock is used when managing its
// internal data structures, but not used when executing UpTimers).
// UpTimerManager deletes UpTimers after executing them
// A tick interval (defaults to 0.1sec) can be set to tune how often
// UpTimerManager checks what UpTimers to execute.

#ifndef UP_MICRO_TIMER_MANAGER_H
#define UP_MICRO_TIMER_MANAGER_H
#include <UpTools/UpHashMap.h>
#include <UpTools/UpThread.h>
#include <UpTools/UpTimer.h>
#include <ext/slist>
#include <set>

////////////////////////////////////////////////////////////////////7
// default values (TICKINTERVAL=0.1 seconds)
////////////////////////////////////////////////////////////////////7
#define TICKINTERVALSEC  ((time_t)0)
#define TICKINTERVALUSEC ((suseconds_t)100000)

////////////////////////////////////////////////////////////////////7
// Class UpTimerManager
////////////////////////////////////////////////////////////////////7
class UpTimerManager : public UpThread {
public:
	// constructor
	UpTimerManager();
	// destructor
	~UpTimerManager();
	// sets interval between catchUp() executions
	void setTickInterval(const UpTimeVal& _tickInterval);
	// Next two methods return time with a precision equal to tickInterval.
	// They return time without using system calls to time() or gettimeofday().
	UpTimeVal now();
	time_t    now_t();
	// add new timers to UpTimerManager
	void add(UpTimer* _timer);
	// UpTimerManager runs in an independent thread. Call this method to start
	// this thread.
	bool startThread();
	// methods to get the current number of pending UpTimers
	unsigned int getCount();
	unsigned int getUnnamedCount();
	unsigned int getNamedCount();
	unsigned int getCount(const char* _className);
	// report of UpTimers to ostream
	void report(std::ostream&);
private:
	// Order operator considering time and memory address between UpTimer*
	class UpTimerPtrLessThan {
	public: bool operator()(const UpTimer* a,const UpTimer* b)
		{ return (*a) < (*b); }
	};
	// single linked list
	typedef __gnu_cxx::slist<UpTimer*>				UpTimerSList;
	// set sorted by timer action time and memory address
	typedef std::set<UpTimer*,UpTimerPtrLessThan>	UpTimerSet;

	UpMutex      mx;
	UpTimerSList tAddList;
	UpTimerSet   tSet;
	UpTimeVal    tickInterval;
	UpTimeVal    nowTime;
	void         catchUp();
	void* run();

	typedef std::map<const char*,unsigned int,UpLessThan> CharPtrUIntMap;
	typedef __gnu_cxx::slist<const char*> CharPtrSList;
	unsigned int unnamedUpTimers;
	CharPtrUIntMap namedUpTimers;
	CharPtrSList auxslist;
	UpTimeVal auxTime;

};
#endif 
